
/**************************************************************************
 *
 *  $Id: mbgclock_main.c 1.2 2017/07/06 09:24:03 martin REL_M $
 *
 *  Description:
 *    Main file for for mbgclock driver to support Meinberg bus level
 *    devices under FreeBSD.
 *
 *    The binary is a loadable module called mbgclock which implements
 *    /dev/mbgclock* devices.
 *
 *    Based on FreeBSD's mypci.c sample program by Murray Stokely.
 *
 * -----------------------------------------------------------------------
 *  $Log: mbgclock_main.c $
 *  Revision 1.2  2017/07/06 09:24:03  martin
 *  Implemented IOCTL handler which basically can also
 *  supports privilege checking for IOCTL.
 *  Improved resource handling.
 *  Use 'struct thread' instead of obsolete 'd_thread_t' type.
 *  Reported by John Baldwin and George Neville-Neil (FreeBSD bug #196692).
 *  Unified handling of program version information.
 *  Support DEBUG messages.
 *  Code cleanup.
 *  Revision 1.1  2011/01/26 13:56:32  martin
 *  Initial skeleton based on FreeBSD's mypci.c sample program by Murray Stokely.
 *
 **************************************************************************/

#include <pcpsdrvr.h>
#include <mbgddmsg.h>
#include <mbgversion.h>
#include <mbgioctl.h>

#include <sys/param.h>        /* defines used in kernel.h */
#include <sys/module.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/kernel.h>       /* types used in module initialization */
#include <sys/conf.h>         /* cdevsw struct */
#include <sys/uio.h>          /* uio struct */
#include <sys/malloc.h>
#include <sys/bus.h>          /* structs, prototypes for pci bus stuff */

#include <sys/rman.h>

#include <dev/pci/pcivar.h>   /* For pci_get macros! */
#include <dev/pci/pcireg.h>



#define MBG_MICRO_VERSION_CODE   0

#define MBG_COPYRIGHT    "(c) Meinberg 2010-" MBG_CURRENT_COPYRIGHT_YEAR_STR


#define MBG_DRVR_NAME    "mbgclock"
const char pcps_driver_name[] = MBG_DRVR_NAME;

#if USE_DEBUG_PORT
  static MBG_DBG_DATA mbg_dbg_data;
  static MBG_DBG_PORT mbg_dbg_port = 0x378 + 0;  //##++
  static PCPS_IO_ADDR_MAPPED mbg_dbg_port_mapped;  //##++
#endif  // defined( DEBUG )

static PCPS_DRVR_INFO drvr_info = { MBG_VERSION_CODE( MBG_MICRO_VERSION_CODE ),
                                    0, MBG_DRVR_NAME " radio clock driver" };

MALLOC_DEFINE( M_MBGCLOCK, "short desc", "long desc" );

#include <macioctl.h>



/* The softc holds our per-instance data. */
struct mbgclock_softc
{
  device_t device;
  struct cdev *cdev;
  PCPS_DDEV *pddev;
};


/* Function prototypes */
static d_open_t mbgclock_open;
static d_close_t mbgclock_close;
static d_read_t mbgclock_read;
static d_write_t mbgclock_write;
static d_ioctl_t mbgclock_ioctl;


/* Character device entry points */

static struct cdevsw mbgclock_cdevsw =
{
  .d_version = D_VERSION,
  .d_open = mbgclock_open,
  .d_close = mbgclock_close,
  .d_read = mbgclock_read,
  .d_write = mbgclock_write,
  .d_ioctl = mbgclock_ioctl,
  .d_name = "mbgclock"
};



static __mbg_inline
void set_dev_connected( PCPS_DDEV *pddev, int state )
{
  _mbgddmsg_4( MBG_DBG_DETAIL, "%s: setting dev %s_%s connected state to %i",
               pcps_driver_name, _pcps_ddev_type_name( pddev ), _pcps_ddev_sernum( pddev ),
               state );
  atomic_store_rel_int( &pddev->connected, state );

}  // set_dev_connected



/*
 * In the cdevsw routines, we find our softc by using the si_drv1 member
 * of struct cdev.  We set this variable to point to our softc in our
 * attach routine when we create the /dev entry.
 */
int
mbgclock_open( struct cdev *dev, int oflags, int devtype, struct thread *td )
{
  struct mbgclock_softc *psc = dev->si_drv1;
  PCPS_DDEV *pddev = psc->pddev;

  _mbgddmsg_3( MBG_DBG_INFO, "%s: dev %s_%s opened successfully.", pcps_driver_name,
               _pcps_ddev_type_name( pddev ), _pcps_ddev_sernum( pddev ) );

  atomic_add_int( &pddev->open_count, 1 );

  return 0;

}  // mbgclock_open



int
mbgclock_close( struct cdev *dev, int fflag, int devtype, struct thread *td )
{
  struct mbgclock_softc *psc = dev->si_drv1;
  PCPS_DDEV *pddev = psc->pddev;

  atomic_subtract_int( &pddev->open_count, 1 );

  _mbgddmsg_3( MBG_DBG_INFO, "%s: dev %s_%s closed.", pcps_driver_name,
               _pcps_ddev_type_name( pddev ), _pcps_ddev_sernum( pddev ) );

  return 0;

}  // mbgclock_close



int
mbgclock_read( struct cdev *dev, struct uio *uio, int ioflag )
{
  #if defined( DEBUG )
    struct mbgclock_softc *psc = dev->si_drv1;
    PCPS_DDEV *pddev = psc->pddev;

    _mbgddmsg_4( MBG_DBG_INFO, "%s: dev %s_%s asked to read %li bytes",
                 pcps_driver_name, _pcps_ddev_type_name( pddev ),
                 _pcps_ddev_sernum( pddev ), (long) uio->uio_resid );
  #endif

  return 0;

}  // mbgclock_read



int
mbgclock_write( struct cdev *dev, struct uio *uio, int ioflag )
{
  #if defined( DEBUG )
    struct mbgclock_softc *psc = dev->si_drv1;
    PCPS_DDEV *pddev = psc->pddev;

    _mbgddmsg_4( MBG_DBG_INFO, "%s: dev %s_%s asked to write %li bytes",
                 pcps_driver_name, _pcps_ddev_type_name( pddev ),
                 _pcps_ddev_sernum( pddev ), (long) uio->uio_resid );
  #endif

  return 0;

}  // mbgclock_write



/*
 * ATTENTION: This function returns one of the MBG_RETURN_CODES.
 * However, POSITIVE numbers need to be returned in case of an error,
 * so mbg_errno_to_os() needs to be called to convert the return code.
 */
int
mbgclock_ioctl( struct cdev *dev, u_long cmd, caddr_t data,
                int32_t flag, struct thread *td )  // credentials in thread
{
  struct mbgclock_softc *psc = dev->si_drv1;
  PCPS_DDEV *pddev = psc->pddev;
  int rc;
  int priv_lvl;

  // Find out which privilege level is required
  // to execute this IOCTL command.
  priv_lvl = ioctl_get_required_privilege( cmd );

  // Check if the calling process has the required privilege.
  switch ( priv_lvl )
  {
    case MBG_REQ_PRIVL_NONE:
      // Always allow.
      break;

    case MBG_REQ_PRIVL_EXT_STATUS:
    case MBG_REQ_PRIVL_CFG_READ:
      // This may require some privilege for the calling process.
      // Anyway, always allow for now.
      break;

    case MBG_REQ_PRIVL_CFG_WRITE:
    case MBG_REQ_PRIVL_SYSTEM:
      #if 0    // TODO Check if required privileges are available, e.g.:
        if ( !capable( CAP_SYS_ADMIN ) )  // don't allow if no root privileges
        {
          _mbgddmsg_5( MBG_DBG_INFO, "%s: %p IOCTL 0x%02lX: permission denied, dev %s_%s",
                       pcps_driver_name, dev, cmd, _pcps_ddev_type_name( pddev ), _pcps_ddev_sernum( pddev ) );
          return EPERM;
        }
        //##+++++  !capable(CAP_SYS_ADMIN) is also used in FreeBSD's xfs_ioctl.c
      #endif
      break;

    default:
      _mbgddmsg_5( MBG_DBG_INFO, "%s: %p IOCTL 0x%02lX: unknown, permission denied, dev %s_%s",
                   pcps_driver_name, dev, cmd, _pcps_ddev_type_name( pddev ), _pcps_ddev_sernum( pddev ) );
      return mbg_errno_to_os( MBG_ERR_PERM );
  }

  rc = ioctl_switch( pddev, cmd, (void *) data, (void *) data );

  return mbg_ret_val_to_os( rc );

}  // mbgclock_ioctl



/* PCI Support Functions */

static void
mbg_deallocate_resource( device_t device, BSD_RSRC_INFO *p_ri, int type )
{
  if ( p_ri->res )
  {
    bus_deactivate_resource( device, type,
        p_ri->rid, p_ri->res );
    bus_release_resource( device, type,
        p_ri->rid, p_ri->res );
    p_ri->res = NULL;
  }

}  // mbg_deallocate_resource



/*
 * deallocate resources
 */
static void
mbg_dealloc_rsrcs( device_t device )
{
  struct mbgclock_softc *psc = device_get_softc( device );
  PCPS_DDEV *pddev = psc->pddev;
  PCPS_RSRC_INFO *prsrci = &pddev->rsrc_info;
  int i;

//  mbg_deallocate_resource( device, &prsrci->irq.bsd, SYS_RES_IRQ );

  for ( i = 0; i < N_PCPS_MEM_RSRC; i++ )
    mbg_deallocate_resource( device, &prsrci->mem[i].bsd, SYS_RES_MEMORY );

  for ( i = 0; i < N_PCPS_PORT_RSRC; i++ )
    mbg_deallocate_resource( device, &prsrci->port[i].bsd, SYS_RES_IOPORT );

}  // mbg_dealloc_rsrcs



static void
mbg_alloc_rsrc( device_t device, int rid, BSD_RSRC_INFO *p_ri, int type, int flags )
{
  p_ri->rid = rid;

  p_ri->res = bus_alloc_resource_any( device, type, &p_ri->rid, flags );

  if ( p_ri->res )
  {
    p_ri->bst = rman_get_bustag( p_ri->res );
    p_ri->bsh = rman_get_bushandle( p_ri->res );
  }

}  // mbg_alloc_rsrc



static void
mbg_alloc_rsrcs( device_t device )
{
  struct mbgclock_softc *psc = device_get_softc( device );
  PCPS_DDEV *pddev = psc->pddev;
  PCPS_RSRC_INFO *prsrci = &pddev->rsrc_info;
  BSD_RSRC_INFO ri;
  int bar;

//    _mbgddmsg_3( MBG_DBG_INIT_DEV, "%s: alloc I/O range %i: PCI device 0x%04X:0x%04X supported",
//                 pcps_driver_name, vend_id, dev_id );

  for ( bar = 0; bar < 5; bar ++ )
  {
    int rid = PCIR_BAR( bar );

    if ( prsrci->num_rsrc_io < N_PCPS_PORT_RSRC )
    {
      mbg_alloc_rsrc( device, rid, &ri, SYS_RES_IOPORT, RF_ACTIVE );

      if ( ri.res )
      {
        prsrci->port[prsrci->num_rsrc_io].bsd = ri;
        pcps_add_rsrc_io( pddev, rman_get_start( ri.res ), rman_get_size( ri.res ) );
        continue;
      }
    }

    if ( prsrci->num_rsrc_mem < N_PCPS_MEM_RSRC )
    {
      mbg_alloc_rsrc( device, rid, &ri, SYS_RES_MEMORY, RF_ACTIVE );

      if ( ri.res )
      {
        prsrci->mem[prsrci->num_rsrc_mem].bsd = ri;
        pcps_add_rsrc_mem( pddev, rman_get_start( ri.res ), rman_get_size( ri.res ) );
        continue;
      }
    }
  }


  // single IRQ resource
#if 0  // currently not used / required
  mbg_alloc_rsrc( device, 0, &ri, SYS_RES_IRQ, RF_SHAREABLE | RF_ACTIVE );

  if ( ri.res )
  {
    prsrci->port[prsrci->num_rsrc_irq].bsd = ri;
    pcps_add_rsrc_irq( pddev, rman_get_start( ri.res ) );
  }
#endif

}  // mbg_alloc_rsrcs



/*
 * Probe: compare the device ID of this device against the IDs that this driver
 * supports.  If there is a match, set the description and return success.
 */
static int
mbgclock_probe( device_t device )
{
  uint16_t vend_id = pci_get_vendor( device );
  uint16_t dev_id = pci_get_device( device );
  PCPS_DEV_TYPE *pdt;


  if ( vend_id != PCI_VENDOR_MEINBERG )
    goto fail;

  pdt = pcps_get_dev_type_table_entry( PCPS_BUS_PCI, dev_id );

  if ( pdt == NULL )
    goto fail;

  device_set_desc( device, pdt->name );

  _mbgddmsg_3( MBG_DBG_INIT_DEV, "%s: probe: PCI device 0x%04X:0x%04X supported",
               pcps_driver_name, vend_id, dev_id );

  return BUS_PROBE_DEFAULT;


fail:
  _mbgddmsg_3( MBG_DBG_INIT_DEV, "%s: probe: PCI device 0x%04X:0x%04X not supported",
                 pcps_driver_name, vend_id, dev_id );
  return ENXIO;

}  // mbgclock_probe



/*
 * Attach function is only called if the probe is successful.
 */
static int
mbgclock_attach( device_t device )
{
  uint16_t dev_id = pci_get_device( device );
  struct mbgclock_softc *psc = device_get_softc( device );
  int rc;

  _mbgddmsg_2( MBG_DBG_INIT_DEV, "%s: attach for device 0x%04X",
               pcps_driver_name, dev_id );

  psc->pddev = pcps_alloc_ddev();

  if ( psc->pddev == NULL )
  {
    _mbgddmsg_2( MBG_DBG_INIT_DEV, "%s: attach device 0x%04X: pcps_alloc_ddev() failed",
                 pcps_driver_name, dev_id );
    rc = ENOMEM;
    goto fail;
  }


  rc = pcps_init_ddev( psc->pddev, PCPS_BUS_PCI, dev_id );

  if ( rc != PCPS_SUCCESS )
  {
    _mbgddmsg_3( MBG_DBG_INIT_DEV, "%s: attach device 0x%04X: pcps_init_ddev() failed, rc: %i",
                 pcps_driver_name, dev_id, rc );
    rc = ENXIO;
    goto fail;
  }


  //##++++++  rc = pci_enable_device( dev );


  mbg_alloc_rsrcs( device );

  //##++++++  rc = pcps_probe_device( pddev, device->bus->number, device->devfn );
  rc = pcps_probe_device( psc->pddev, 0, 0 );

  if ( rc != PCPS_SUCCESS )
  {
    _mbgddmsg_3( MBG_DBG_INIT_DEV, "%s: attach device 0x%04X: pcps_probe_device() failed, rc: %i",
                 pcps_driver_name, dev_id, rc );
    rc = ENXIO;
    goto fail;
  }


  set_dev_connected( psc->pddev, 1 );

  /* Initialize our softc. */
  psc->device = device;
  /*
   * Create a /dev entry for this device.  The kernel will assign us
   * a major number automatically.  We use the unit number of this
   * device as the minor number and name the character device
   * "mbgclock<unit>".
   */
  psc->cdev = make_dev( &mbgclock_cdevsw, device_get_unit( device ),
    UID_ROOT, GID_WHEEL, 0600, "mbgclock%u", device_get_unit( device ) );

  psc->cdev->si_drv1 = psc;

  _mbgddmsg_2( MBG_DBG_INIT_DEV, "%s: device 0x%04X attached successfully",
               pcps_driver_name, dev_id );

  return 0;


fail:
  mbg_dealloc_rsrcs( device );

  if ( psc->pddev )
    pcps_free_ddev( psc->pddev );

  _mbgddmsg_2( MBG_DBG_INIT_DEV, "%s: failed to attach device 0x%04X",
               pcps_driver_name, dev_id );
  return rc;

}  // mbgclock_attach



/*
 * Detach device.
 */
static int
mbgclock_detach( device_t device )
{
  #if defined( DEBUG )
    uint16_t dev_id = pci_get_device( device );
  #endif
  struct mbgclock_softc *psc;
  PCPS_DDEV *pddev;

  _mbgddmsg_2( MBG_DBG_INIT_DEV, "%s: detach for device 0x%04X",
               pcps_driver_name, dev_id );

  /* Teardown the state in our softc created in our attach routine. */
  psc = device_get_softc( device );
  pddev = psc->pddev;

  set_dev_connected( pddev, 0 );

//  pcps_free_ddev( pddev );  //##++++++  should wait for outstanding requests
  mbg_dealloc_rsrcs( device );

  if ( psc->pddev )
    pcps_free_ddev( psc->pddev );

  psc->pddev = NULL;

  destroy_dev( psc->cdev );

  _mbgddmsg_2( MBG_DBG_INIT_DEV, "%s: device 0x%04X detached",
               pcps_driver_name, dev_id );

  return 0;

}  // mbgclock_detach



/*
 * Called during system shutdown after sync.
 */
static int
mbgclock_shutdown( device_t device )
{
  printf( "mbgclock shutdown!\n" );

  return 0;

}  // mbgclock_shutdown



/*
 * Device suspend routine.
 */
static int
mbgclock_suspend( device_t device )
{
  printf( "mbgclock suspend!\n" );

  return 0;

}  // mbgclock_suspend



/*
 * Device resume routine.
 */
static int
mbgclock_resume( device_t device )
{
  printf( "mbgclock resume!\n" );

  return 0;

}  // mbgclock_resume



static device_method_t mbgclock_methods[] =
{
  /* Device interface */
  DEVMETHOD( device_probe,     mbgclock_probe ),
  DEVMETHOD( device_attach,    mbgclock_attach ),
  DEVMETHOD( device_detach,    mbgclock_detach ),
  DEVMETHOD( device_shutdown,  mbgclock_shutdown ),
  DEVMETHOD( device_suspend,   mbgclock_suspend ),
  DEVMETHOD( device_resume,    mbgclock_resume ),
  { 0, 0 }
};



static devclass_t mbgclock_devclass;

DEFINE_CLASS_0( mbgclock, mbgclock_driver, mbgclock_methods, sizeof( struct mbgclock_softc ) );
DRIVER_MODULE( mbgclock, pci, mbgclock_driver, mbgclock_devclass, 0, 0 );
