
/**************************************************************************
 *
 *  $Id: pcpsdrvr.c 1.52 2017/07/04 16:45:36 martin REL_M $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Driver functions that detect Meinberg PC plug-in devices and set up
 *    the software environment (port base address, clock features, etc.).
 *
 *    These functions should be used with programs which have direct
 *    access to the hardware (e.g. device drivers).
 *
 *    Programs which access the devices via device drivers should
 *    use the functions provided by the mbgdevio module.
 *
 *    There are several preprocessor symbols defined at the top of
 *    pcpsdrvr.h which control the default support of some features
 *    under the different operating systems. If required, each of
 *    those symbols can be overridden by compiler arguments.
 *
 *    Basically the following devices are supported:
 *      USB v2:          DCF600USB, TCR600USB, MSF600USB, WVB600USB
 *      USB v1:          USB5131, TCR51USB, MSF51USB, WWVB51USB
 *      PCI express:     PEX511, TCR511PEX, GPS170PEX, PTP270PEX,
 *                       FRC511PEX, TCR170PEX, GPS180PEX, TCR180PEX
 *                       PZF180PEX, GLN180PEX, GPS180AMC, GNS181PEX
 *      PCI bus 5V/3.3V: PCI510, PCI511, GPS169PCI, GPS170PCI,
 *                       TCR510PCI, TCR167PCI, TCR511PCI
 *      PCI bus 5V:      PCI32, GPS167PCI, PCI509, GPS168PCI
 *      MCA bus:         PS31
 *      ISA bus:         PC31, PC32, GPS167PC
 *
 *    USB is not supported for all target environments, eventually
 *    because an operating systems doesn't provide full USB support,
 *    or USB support hasn't yet been implemented.
 *
 *    PCI support is possible in two different ways. The preferred
 *    functions are compiled in if one of the symbols _PCPS_USE_PCI_PNP
 *    or _PCPS_USE_PCI_BIOS is defined != 0.
 *
 *    If _PCPS_USE_PCI_PNP is != 0 it is assumed that the operating
 *    system's PCI layer detects a new PCI device and calls a driver's
 *    add_device()/start_device() function to initialize the device.
 *    This technique is supported with PNP operating systems
 *    (e.g. Windows versions after NT, Linux, *BSD).
 *
 *    If _PCPS_USE_PCI_BIOS is != 0 the program scans the PCI bus
 *    during startup to detect and initialize supported PCI devices.
 *    This techique is used with old non-PNP operating systems.
 *
 *    The symbol _PCPS_USE_RSRCMGR must be defined != 0 to include
 *    support of resource managers, if necessary.
 *
 *    If the symbol _PCPS_USE_MCA is defined != 0 then Micro Channel
 *    detection (and therefore auto-detection of a MCA clock) is
 *    supported.
 *
 *    MCA devices are accessed using the same low level functions as
 *    ISA devices, so if autodetection of MCA clocks is not supported
 *    then a MCA device's known port number can be passed to
 *    pcps_detect_devices() to let it be treated like an ISA device.
 *
 * -----------------------------------------------------------------------
 *  $Log: pcpsdrvr.c $
 *  Revision 1.52  2017/07/04 16:45:36  martin
 *  Support GPS180AMC and GNS181PEX.
 *  Renamed some functions: Use _device instead of _clock,
 *  pcps_start_device() is now called pcps_probe_device(), etc.
 *  Runtime support for forcing I/O rather than MM access.
 *  Increase MAX_BOOT_TIME_PTP270PEX from 27 to 40 seconds
 *  to be safe in case a firmware update is applied at startup.
 *  Fixed type of a register address.
 *  Cleaned up I/O port usage.
 *  Older defines N_SUPP_DEV, PCPS_MAX_DDEVS, and MBG_MAX_DEVICES
 *  have been obsoleted by new defines N_SUPP_DEV_BUS, N_SUPP_DEV_EXT,
 *  and N_SUPP_DEV_TOTAL.
 *  Fixed DEBUG build under *BSD.
 *  Added DEBUG code dumping RECEIVER_INFO.
 *  Moved mbg_delta_sys_time_ms() to new module mbgsystm.c.
 *  Check for MBG_TGT_POSIX instead of MBG_TGT_UNIX.
 *  Provided a driver name string for debug build on direct-access targets.
 *  Avoid 'redundant redeclaration' warning under FreeBSD 8.2.
 *  Fixed macro definition syntax to avoid clang compiler warnings.
 *  Fixed some other warnings from clang.
 *  Attribute always_inline is now in __mbg_inline.
 *  Conditional USB debug code.
 *  Account for renamed symbols.
 *  Fixed typos, wording, and doxygen comments.
 *  Removed trailing white space.
 *  Revision 1.51  2013/10/01 14:19:03  martin
 *  Support GLN180PEX.
 *  Revision 1.50  2013/03/15 10:01:58  martin
 *  Modified/added some debug messages.
 *  Revision 1.49  2013/03/15 08:35:08  martin
 *  Account for PTP270PEX HW v2 cards which can indicate
 *  when they are ready to be accessed.
 *  Account for unified, renamed PLX symbols.
 *  Revision 1.48  2012/11/05 16:32:02Z  martin
 *  Fixed and enhanced some timing DEBUG code.
 *  Rewrote report_uptime().
 *  Revision 1.47  2012/10/15 14:12:08  martin
 *  Support GPS180PEX, TCR180PEX, and PZF180PEX.
 *  Support DCF600USB, TCR600USB, MSF600USB, and WVB600USB.
 *  Support FreeBSD and NetBSD.
 *  Support on-board event log.
 *  If required, wait until PTP270PEX has finished booting.
 *  Use USB micro frame timing conditionally only, yet disabled by default.
 *  Conditional debug code to test PCI access time and/or execution time
 *  of the low level functions.
 *  Always read the serial number directly from the device.
 *  Unified low level functions and use 16 bit types for buffer sizes.
 *  Added warnings in case a device is not handled by chip setup
 *  or device feature check.
 *  Modified low level AMCC read functions for SPARC to fix unaligned access.
 *  Added debug messages for generic I/O.
 *  Initialize device semaphores only early at device initializition
 *  if required, otherwise later, since early initialization can lead
 *  to a trap e.g. under Windows.
 *  Conditional code to test MM I/O for new PCI cards.
 *  Use common mutex primitives from mbgmutex.h.
 *  Optionally support timespec for sys time (USE_TIMESPEC).
 *  Set up basic default receiver info for devices which don't
 *  support this structure.
 *  Moved MBG_TGT_SUPP_MEM_ACC definition to pcpsdev.h.
 *  Introduced and use new function check_ri_feature().
 *  Also detect support for raw IRIG data from RECEIVER_INFO.
 *  Revision 1.46  2009/12/15 14:45:33  daniel
 *  Account for feature to read the raw IRIG bits.
 *  Revision 1.45  2009/09/29 07:24:50Z  martin
 *  Use standard feature flag to check if fast HR time is supported.
 *  Revision 1.44  2009/06/23 07:10:47  martin
 *  Fixed/modified some debug messages.
 *  Revision 1.43  2009/06/19 12:13:59  martin
 *  Check if TCR cards support raw IRIG time.
 *  Revision 1.42  2009/06/09 10:15:33  daniel
 *  Check if card has LAN interface and supports PTP.
 *  Revision 1.41  2009/04/08 08:33:20  daniel
 *  Check whether the TCR511PCI or devices with
 *  RECEIVER_INFO support IRIG control function bits.
 *  Revision 1.40  2009/03/27 09:55:13Z  martin
 *  Added some debug messages.
 *  Account for renamed library symbols.
 *  Revision 1.39  2009/03/19 12:04:31Z  martin
 *  Adjust endianess of ASIC version and ASIC features after having read.
 *  Revision 1.38  2009/03/17 15:33:53  martin
 *  Support reading IRIG control function bits.
 *  Revision 1.37  2009/03/13 09:17:00Z  martin
 *  Bug fix: Hadn't checked whether TCR170PEX card provides the
 *  programmable synthesizer.
 *  As a fix moved the code from the body of check_opt_features()
 *  into pcps_start_device() so that the check is done for every
 *  type of card.
 *  Swap receiver_info to make this work on non-x86 architectures.
 *  Support configurable time scales, and reading/writing GPS UTC
 *  parameters via the PC bus.
 *  Use mbg_get_pc_cycles() instead of _pcps_get_cycles().
 *  Revision 1.36  2009/01/13 12:03:57Z  martin
 *  Generate a separate warning message if the firmware could not
 *  be read from an ISA card.
 *  Care about "long long" in debug msg.
 *  Revision 1.35  2008/12/16 14:38:49Z  martin
 *  Account for new devices PTP270PEX, FRC270PEX, TCR170PEX, and WWVB51USB.
 *  Check the firmware / ASIC version of PEX cards and flag the device
 *  unsafe for IRQs if the versions are older than required.
 *  Check whether PEX511 and PCI511 support HR time.
 *  Moved initialization of common spinlocks and mutexes to pcps_start_device().
 *  Take access cycles count in the low level routines, with interrupts disabled.
 *  Cleanup for pcps_read_usb() which is now possible since access cycles count
 *  is now taken inside the low evel routines.
 *  Support mapped I/O resources, unaligned access and endianess conversion.
 *  Account for ASIC_FEATURES being coded as flags, and account for
 *  new symbol PCI_ASIC_HAS_MM_IO.
 *  Account for new MBG_PC_CYCLES type.
 *  Account for signed irq_num.
 *  Renamed MBG_VIRT_ADDR to MBG_MEM_ADDR.
 *  Use MBG_MEM_ADDR type for memory rather than split high/low types.
 *  Distinguish device port variables for IRQ handling.
 *  Preliminarily support USB latency compensation under Win32 PNP targets
 *  and account for USB EHCI microframe timing which requires a different
 *  latency compensation approach. This is useful if a USB 2.0 hub is connected
 *  between device and host.
 *  Also read the ASIC version at device initialization.
 *  pcps_alloc_ddev() does not take a parameter anymore.
 *  Cleaned up comments.
 *  Revision 1.34  2008/02/27 10:03:02  martin
 *  Support TCR51USB and MSF51USB.
 *  Preliminary support for mapped memory access under Windows and Linux.
 *  Enabled PCPS_IRQ_1_SEC for USB within WIN32 targets
 *  in pcps_start_device().
 *  Fixed a bug in pcps_write() where the error code
 *  that was returned from a USB device was misinterpreted
 *  due to a signed/unsigned mismatch (added typecast).
 *  Removed obsolete function pcps_cleanup_all_devices().
 *  Code cleanup.
 *  Revision 1.33  2008/01/31 08:51:30Z  martin
 *  Picked up changes from 1.31.2.1:
 *  Changed default definition of PCI_DWORD to uint32_t.
 *  Removed erraneous brace from debug code.
 *  Revision 1.32  2007/09/26 11:05:57Z  martin
 *  Added support for USB in general and new USB device USB5131.
 *  Renamed ..._USE_PCIMGR symbols to ..._USE_PCI_PNP.
 *  Renamed ..._USE_PCIBIOS symbols to ..._USE_PCI_BIOS.
 *  Added new symbol _USE_ISA_PNP to exclude non-PNP stuff.
 *  from build if ISA devices are also handled by the PNP manager.
 *  Use new MBG_... codes defined in mbgerror.h.
 *  Unified timeout handling in low level functions by using an inline function.
 *  Renamed pcps_pnp_start_device() to pcps_start_device().
 *  Renamed pcps_setup_pci_dev() to pcps_setup_and_startpci_dev().
 *  Merged code from init_ddev_cfg() and finish_ddev_cfg() into pcps_start_device().
 *  Improved and unified handling of ISA devices.
 *  Removed calling register_pnp_devices() from pcps_detect_clocks(),
 *  this is now called directly.
 *  Added missing IRIG support to pcps_rsrc_register_device().
 *  Revision 1.31  2007/07/17 08:22:47Z  martin
 *  Added support for TCR511PEX and GPS170PEX.
 *  Revision 1.30  2007/07/16 12:56:01Z  martin
 *  Added support for PEX511.
 *  Rewrote common resource handling code in order to simplify
 *  OS specific code.
 *  Revision 1.29  2007/03/02 09:40:33Z  martin
 *  Use generic port I/O macros.
 *  Pass PCPS_DDEV structure to the low level read functions.
 *  Use new _pcps_..._timeout_clk() macros.
 *  Added init code qualifier.
 *  Preliminary support for *BSD.
 *  Preliminary support for USB.
 *  Revision 1.28  2006/07/11 10:24:20  martin
 *  Use _fmemcpy() in pcps_generic_io() to support environments which
 *  require far data pointers.
 *  Revision 1.27  2006/07/07 09:41:15  martin
 *  Renamed pci_..() function calls to _mbg_pci_..() calls which are defined according to the
 *  OS requirements, in order to avoid naming conflicts.
 *  Revision 1.26  2006/06/19 15:28:52  martin
 *  Added support for TCR511PCI.
 *  Modified parameters required to detect ISA cards.
 *  The array of port addresses does no more require a 0 address
 *  as last value.
 *  Revision 1.25  2006/03/10 11:01:27  martin
 *  Added support for PCI511.
 *  Revision 1.24  2005/11/03 15:50:45Z  martin
 *  Added support for GPS170PCI.
 *  Revision 1.23  2005/09/16 08:21:08Z  martin
 *  Also flag PCI cards which have base_addr set to 0 as uninitialized.
 *  Revision 1.22  2005/06/02 10:32:07Z  martin
 *  Changed more types to C99 fixed size types.
 *  New function pcps_generic_io().
 *  Revision 1.21  2004/12/13 14:19:38Z  martin
 *  Support configuration of on-board frequency synthesizer.
 *  Revision 1.20  2004/11/09 13:02:48Z  martin
 *  Redefined fixed width data types using standard C99 types.
 *  Fixed warnings about lvalue casts.
 *  Revision 1.19  2004/10/14 15:01:24  martin
 *  Added support for TCR167PCI.
 *  Revision 1.18  2004/09/06 15:16:57Z  martin
 *  Support a GPS_DATA interface where sizes are specified
 *  by 16 instead of the original 8 bit quantities, thus allowing
 *  to transfer data blocks which exceed 255 bytes.
 *  Conditionally skip assertions under Linux.
 *  Revision 1.17  2004/04/22 14:47:54  martin
 *  Fixed conversion of firmware rev. number.
 *  Revision 1.16  2004/04/07 09:45:04Z  martin
 *  Support new feature PCPS_HAS_IRIG_TX for GPS169PCI.
 *  Revision 1.15  2003/12/22 16:15:21Z  martin
 *  Support PCPS_HR_TIME for TCR510PCI.
 *  Revision 1.14  2003/07/30 07:28:23Z  martin
 *  Moved prototype for register_pci_devices() outside to top of file.
 *  Revision 1.13  2003/07/08 15:11:55  martin
 *  Support PCI PNP interface under Linux.
 *  New function pcps_rsrc_release().
 *  Made some functions public.
 *  Renamed some public functions to start with pcps_...
 *  Revision 1.12  2003/06/19 10:08:43  MARTIN
 *  Renamed some functions to follow common naming conventions.
 *  Made a function's parameter pointer const.
 *  Changes due to renamed symbols in pcpsdev.h.
 *  Check devices for _pcps_has_ucap() support.
 *  Revision 1.11  2003/05/16 09:28:06  MARTIN
 *  Moved inclusion of some headers to pcpsdrvr.h.
 *  Revision 1.10  2003/04/09 16:35:57  martin
 *  Supports PCI510, GPS169PCI, and TCR510PCI,
 *  and new PCI_ASIC used by those devices.
 *  Revision 1.9  2003/03/20 11:42:37  martin
 *  Fixed syntax for QNX.
 *  Revision 1.8  2002/08/09 08:25:50  MARTIN
 *  Support feature PCPS_CAN_CLR_CAP_BUFF.
 *  Fixed a bug resulting in an unterminated string
 *  if SERNUM was being read.
 *  Revision 1.7  2002/02/26 09:31:57  MARTIN
 *  New function pcps_read_sernum().
 *  Revision 1.6  2002/02/19 09:46:26  MARTIN
 *  Use new header mbg_tgt.h to check the target environment.
 *  Removed function pcps_sn_str_from_ident(), use new
 *  function mbg_gps_ident_decode() from identdec.c now.
 *  If a PCI clock's interface is not properly configured don't
 *  enable the device and set the read function to the new
 *  dummy function pcps_read_null() to prevent driver from
 *  accessing random ports.
 *  Revision 1.5  2002/02/01 12:06:12  MARTIN
 *  Added support for GPS168PCI.
 *  Removed obsolete code.
 *  Revision 1.4  2001/11/30 09:52:48  martin
 *  Added support for event_time which, however, requires
 *  a custom GPS firmware.
 *  Revision 1.3  2001/09/18 06:59:18  MARTIN
 *  Account for new preprocessor symbols in the header file.
 *  Added some type casts to avoid compiler warnings under Win32.
 *  Added some debug messages to clock detection functions.
 *  Revision 1.2  2001/03/16 14:45:33  MARTIN
 *  New functions and definitions to support PNP drivers.
 *  Revision 1.1  2001/03/01 16:26:41  MARTIN
 *  Initial revision for the new library.
 *
 **************************************************************************/

#define _PCPSDRVR
  #include <pcpsdrvr.h>
#undef _PCPSDRVR

#include <identdec.h>
#include <mbgddmsg.h>
#include <plxdefs.h>
#include <pci_asic.h>
#include <amccdefs.h>
#include <pcidefs.h>

#if defined( MBG_TGT_WIN32_PNP )
  #include <usbdrv.h>
  #include <pcpsdefs.h>
  #include <ntddk.h>
  #include <stdio.h>
#elif defined( MBG_TGT_WIN32 )
  #include <pcps_ioc.h>
  #include <stdio.h>
#elif defined( MBG_TGT_DOS )
  #include <mbgplx.h>
#endif

#if !defined( MBG_TGT_LINUX ) && !defined( MBG_TGT_BSD )
  #include <assert.h>
#endif

#if defined( MBG_TGT_FREEBSD )
  #include <sys/rman.h>
#endif

#if _PCPS_USE_MCA
  #include <mca.h>
#endif

#if _PCPS_USE_PCI
  #include <pci.h>
#endif

#if _PCPS_USE_USB
  #define MBGUSB_MIN_ENDPOINTS_REQUIRED  3
#endif


// In NetBSD defining DEBUG to build a debug version of this driver
// would interfere with DEBUG being defined for the kernel sources,
// so we need to define MBG_DEBUG instead. For simplicity, we define
// DEBUG locally in this module if MBG_DEBUG is defined.
#if defined( MBG_TGT_NETBSD ) && defined( MBG_DEBUG )
  #undef DEBUG
  #define DEBUG MBG_DEBUG
#endif


#if !defined( USE_CMD_PTR )
  #define USE_CMD_PTR  1
#endif

#if !defined( DEBUG_ACCESS_TIMING )
  // test how much cycles it takes to read/write a register on the board
  #if ( defined( DEBUG ) && ( DEBUG >= 10 ) )
    #define DEBUG_ACCESS_TIMING  1
  #else
    #define DEBUG_ACCESS_TIMING  0
  #endif
#endif

#if !defined( DEBUG_IO_TIMING )
  // test how much cycles it takes for a low level function to execute
  #if ( defined( DEBUG ) && ( DEBUG >= 10 ) )
    #define DEBUG_IO_TIMING  1
  #else
    #define DEBUG_IO_TIMING  0
  #endif
#endif

#if !defined( DEBUG_IO )
  // debug which bytes are written to or read from the board
  #if ( defined( DEBUG ) && ( DEBUG >= DEBUG_LVL_IO ) )
    #define DEBUG_IO  1
  #else
    #define DEBUG_IO  0
  #endif
#endif

#if !defined( DEBUG_PORTS )
  #if ( defined( DEBUG ) && ( DEBUG >= DEBUG_LVL_PORTS ) )
    #define DEBUG_PORTS  1
  #else
    #define DEBUG_PORTS  0
  #endif
#endif

#if !defined( DEBUG_SERNUM )
  #if ( defined( DEBUG ) && ( DEBUG >= DEBUG_LVL_SERNUM ) )
    #define DEBUG_SERNUM  1
  #else
    #define DEBUG_SERNUM  0
  #endif
#endif

#if !defined( USE_USB_MICRO_FRAMES )
  #define USE_USB_MICRO_FRAMES  0
#endif


/**
 * @brief Max. time required for PTP270PEX v1 card to be ready after power-up.
 *
 * The PTP270PEX cards have an on-board Linux system which must have
 * finished booting before the card can be accessed via the PCI bus.
 * If the card is accessed earlier then the PCI bus and thus the computer
 * may hang.
 *
 * There are 2 different hardware versions of the PTP270PEX card.
 * PTP270PEX v2 hardware and firmware can flag via a bit in the PCI
 * configuration space when the card is ready, so the driver can just
 * wait until the card has flagged that it's ready.
 *
 * However, the PTP270PEX v1 hardware doesn't support this, so the driver
 * checks the system uptime and waits until it is greater or equal the
 * maximum time required for a PTP270PEX v1 card to finish booting.
 *
 * The nominal time is 27 s, but if a firmware update has been submitted
 * then the update is applied first, which takes another few seconds.
 *
 * @see ::wait_ptp270pex_ready
 * @see ::report_uptime
 */
#define MAX_BOOT_TIME_PTP270PEX    40  // [s]


#if defined( MBG_TGT_BSD )
  // Avoid compiler warnings "redundant redeclaration of ..."
  // e.g. under FreeBSD 8.2 with gcc 4.2.1.
  #define AVOID_REDUNDANT_REDECLARATION  1
#endif

#if defined( MBG_TGT_KERNEL )
  // Driver name defined by the skeleton driver for the target OS.
  extern const char pcps_driver_name[];
#else
  // No skeleton driver, but we need a string to pass to debug messages.
  #if DEBUG
    const char pcps_driver_name[] = "pcpsdrvr";
  #endif
#endif


// In some environments special far functions are are neither
// required nor supported, so redefine calls to those functions
// to appropriate standard function calls.
#if defined( MBG_TGT_NETWARE ) || defined( MBG_TGT_WIN32 ) || \
    defined( MBG_TGT_LINUX ) || defined( MBG_TGT_BSD ) || \
    defined( MBG_TGT_QNX )
  #define _fmemcpy( _d, _s, _n )      memcpy( _d, _s, _n )
  #define _fstrlen( _s )              strlen( _s )
  #define _fstrncmp( _s1, _s2, _n )   strncmp( (_s1), (_s2), (_n) )
#elif defined( MBG_TGT_OS2 )
  #define _fstrncmp( _s1, _s2, _n )   _fmemcmp( (_s1), (_s2), (_n) )
#endif

#if defined( MBG_TGT_OS2 )
  // Watcom C Compiler options for the OS/2 device driver result in
  // warnings if automatic stack addresses are passed to functions.
  #define static_wc static
  #define FMT_03X "%X"
  #define FMT_08X "%X"
#else
  #define static_wc
  #define FMT_03X "%03X"
  #define FMT_08X "%08lX"
#endif

#if defined( MBG_TGT_LINUX )
  typedef unsigned int PCI_DWORD;
#else
  typedef uint32_t PCI_DWORD;
#endif


#if defined( MBG_TGT_LINUX )

  #define _pcps_irq_flags \
    unsigned long irq_flags;

  #define _pcps_disb_local_irq_save() \
    local_irq_save( irq_flags )

  #define _pcps_local_irq_restore() \
    local_irq_restore( irq_flags )

#elif defined( MBG_TGT_WIN32 )

  #define _pcps_irq_flags \
    KIRQL old_irq_lvl;

  #define _pcps_disb_local_irq_save() \
     KeRaiseIrql( HIGH_LEVEL, &old_irq_lvl )

  #define _pcps_local_irq_restore() \
     KeLowerIrql( old_irq_lvl )

#else

  // Nothing to define here.

#endif

#if !defined( _pcps_irq_flags ) && \
    !defined( _pcps_disb_local_irq_save ) && \
    !defined( _pcps_local_irq_restore)
  #define _pcps_irq_flags
  #define _pcps_disb_local_irq_save();
  #define _pcps_local_irq_restore();
#endif


#if defined( MBG_TGT_LINUX ) && defined( time_after )
  #define _pcps_time_after( _curr, _tmo ) \
          time_after( (unsigned long) _curr, (unsigned long) _tmo )
#else
  #define _pcps_time_after( _curr, _tmo )   ( _curr >= _tmo )
#endif


#if defined( MBG_TGT_DOS ) || \
    defined( MBG_TGT_QNX )
  #define MBG_TGT_HAS_UPTIME    0
#elif defined( MBG_TGT_FREEBSD )
  // Under FreeBSD (at least 8.2) the kernel calls to read uptime always
  // return 1 when this driver is loaded automatically, so the system
  // locks up if we wait util uptime has reached a certain value.
  #define MBG_TGT_HAS_UPTIME    0
#else
  #define MBG_TGT_HAS_UPTIME    1
#endif



/**
 * @brief A table used to map ::RECEIVER_INFO::features to ::PCPS_DEV_CFG::features
 *
 * The enumerated ::GPS_FEATURE_BITS can be used as index
 * to this table, and the table entry contains a combination
 * of associated @ref PCPS_FEATURE_MASKS, if there are any.
 *
 * @note Devices which support a configurable time scale do also
 * support reading/writing the GPS %UTC parameters via the PC bus.
 * This is not explicitly coded in the ::RECEIVER_INFO::features
 * since the the ::RECEIVER_INFO structure can also be read via
 * the serial port, and reading/writing the GPS %UTC parameters
 * via the serial port is supported by all GPS devices anyway.
 *
 * @note Devices which support reading ::MBG_RAW_IRIG_DATA via the PC bus
 * interface also support reading ::PCPS_IRIG_TIME. However, there is
 * no associated flag in ::RECEIVER_INFO::features since this call
 * is not supported via the serial interface. Thus we use the
 * ::GPS_HAS_RAW_IRIG_DATA flag to check both features.
 *
 * @see ::GPS_FEATURE_BITS
 * @see @ref PCPS_FEATURE_MASKS
 */
static uint32_t ri_feat_tbl[N_GPS_FEATURE] =
{
  0,                                            ///< see ::GPS_FEAT_PPS
  0,                                            ///< see ::GPS_FEAT_PPM
  PCPS_HAS_SYNTH,                               ///< see ::GPS_FEAT_SYNTH
  0,                                            ///< see ::GPS_FEAT_DCFMARKS
  PCPS_HAS_IRIG_TX,                             ///< see ::GPS_FEAT_IRIG_TX
  0,                                            ///< see ::GPS_FEAT_IRIG_RX
  PCPS_HAS_LAN_INTF,                            ///< see ::GPS_FEAT_LAN_IP4
  0,                                            ///< see ::GPS_FEAT_MULTI_REF

  0,                                            ///< see ::GPS_FEAT_RCV_TIMEOUT
  0,                                            ///< see ::GPS_FEAT_IGNORE_LOCK
  0,                                            ///< see ::GPS_FEAT_5_MHZ
  0,                                            ///< see ::GPS_FEAT_XMULTI_REF
  0,                                            ///< see ::GPS_FEAT_OPT_SETTINGS
  PCPS_HAS_TIME_SCALE | PCPS_HAS_UTC_PARM,      ///< see ::GPS_FEAT_TIME_SCALE
  PCPS_HAS_IRIG_CTRL_BITS,                      ///< see ::GPS_FEAT_IRIG_CTRL_BITS
  PCPS_HAS_PTP,                                 ///< see ::GPS_FEAT_PTP

  0,                                            ///< see ::GPS_FEAT_NAV_ENGINE_SETTINGS
  PCPS_HAS_IRIG_TIME | PCPS_HAS_RAW_IRIG_DATA,  ///< see ::GPS_FEAT_RAW_IRIG_DATA
  0,                                            ///< see ::GPS_FEAT_RAW_IRIG_TIME
  0,                                            ///< see ::GPS_FEAT_PTP_UNICAST
  0,                                            ///< see ::GPS_FEAT_GPIO
  0,                                            ///< see ::GPS_FEAT_XMRS_MULT_INSTC
  0,                                            ///< see ::GPS_FEAT_10MHZ_DISBD
  PCPS_HAS_EVT_LOG,                             ///< see ::GPS_FEAT_EVT_LOG

  0,                                            ///< see ::GPS_FEAT_IMS
  0,                                            ///< see ::GPS_FEAT_HAVEQUICK
  0,                                            ///< see ::GPS_FEAT_NTP
  0,                                            ///< see ::GPS_FEAT_NET_CFG
  0,                                            ///< see ::GPS_FEAT_VST
  0,                                            ///< see ::GPS_FEAT_SHS
  0                                             ///< see ::GPS_FEAT_XBP
};



#if defined( DEBUG )

/**
 * @brief A table of name strings associated with ::GPS_FEATURE_BITS
 */
static const char *gps_ri_feature_names[N_GPS_FEATURE] = DEFAULT_GPS_FEATURE_NAMES;

/**
 * @brief A table of name strings associated with ::PCPS_FEATURE_BITS
 */
static const char *pcps_feature_names[N_PCPS_FEATURE_BITS] = PCPS_FEATURE_NAMES;

#endif  // defined( DEBUG )



static __mbg_inline /*HDR*/
/**
 * @brief Check if a device is a PTP270PEX card
 */
int pcps_ddev_is_ptp270pex( const PCPS_DDEV *pddev )
{
  return _pcps_ddev_is_pci( pddev ) &&
       ( _pcps_ddev_dev_id( pddev ) == PCI_DEV_PTP270PEX );

}  // pcps_ddev_is_ptp270pex




#if defined( MBG_TGT_LINUX )

static /*HDR*/
/**
 * @brief Read a dword from a PLX PECS register
 *
 * @param[in]  pNode  Device structure
 * @param[in]  reg    Number of the register to read
 * @param[out] pval   Pointer to a variable to take the value read from the register
 *
 * @return ::PCI_SUCCESS on success, else one of the PCI BIOS error codes
 */
int mbg_plx_read_pecs_reg( struct pci_dev *pNode,
                           uint16_t reg, uint32_t *pval )
{
  int rc;

  if ( reg & 0x03 )
    return PCI_BAD_REGISTER_NUMB;

  if ( reg < 0x1000 )
  {
    rc = pci_read_config_dword( pNode, reg, pval );
  }
  else
  {
    uint32_t mcr_idx_sav = (uint32_t) -1;

    reg -= 0x1000;

    rc = pci_read_config_dword( pNode, PLX_PECS_MAININDEX, &mcr_idx_sav );

    if ( rc != PCI_SUCCESS )
      goto done;

    rc = pci_write_config_dword( pNode, PLX_PECS_MAININDEX, reg );

    if ( rc != PCI_SUCCESS )
      goto done;

    rc = pci_read_config_dword( pNode, PLX_PECS_MAINDATA, pval );

    if ( rc != PCI_SUCCESS )
      goto done;

    rc = pci_write_config_dword( pNode, PLX_PECS_MAININDEX, mcr_idx_sav );
  }

done:
  return rc;

}  // mbg_plx_read_pecs_reg

#endif



static /*HDR*/
/**
 * @brief Check if a PTP270PEX card can indicate when it's ready
 *
 * A PTP270PEX card must be accessed by the driver only
 * after it has finished booting. Otherwise the host system
 * may be locked up.
 *
 * On HW v2 cards the firmware can indicate when the card is
 * ready to be accessed, so this can be checked by the driver.
 * On these cards the on-board GPIO3 pin is hardwired to 0
 * to indicate this is supported, whereas on older v1 cards
 * the GPIO3 pin is pulled up to 1.
 *
 * @param[in]  pddev   Pointer to the device structure
 *
 * @return true if the card can flag "ready"
 *
 * @see ::ptp270pex_has_flagged_ready
 */
bool ptp270pex_can_flag_ready( const PCPS_DDEV *pddev )
{
  // The GPIO3 input level can only be read via a register
  // which is located inside the PCI configuration space
  // of the built-in PLX8111 PCIe-to-PCI bridge.
  //
  // So we must first locate the PCI bridge associated with our device,
  // read the PECS_GPIOCTL register and return 1 if the GPIO3 bit
  // has been pulled down to 0.

  uint32_t reg_val = (uint32_t) -1;
  int rc = -1;

  #if defined( MBG_TGT_LINUX )

    struct pci_dev *bridge = NULL;

    // search for any PLX PCI bridge device
    while ( ( bridge = pci_get_device( PCI_VENDOR_ID_PLX,
                         PCI_DEVICE_ID_PLX_8111, bridge ) ) != NULL )
    {
      uint8_t uc;

      #if defined( DEBUG )
        printk( KERN_INFO "%s: Found bridge: %s\n",
                pcps_driver_name, pci_name( bridge ) );
      #endif

      // Read the secondary bus number.
      rc = pci_read_config_byte( bridge, 0x19, &uc );

      if ( ( rc == PCI_SUCCESS ) && ( uc == pddev->dev.cfg.bus_num ) )
      {
        #if defined( DEBUG )
          printk( KERN_INFO "%s: Found bridge associated with device %s\n",
                  pcps_driver_name, _pcps_ddev_type_name( pddev ) );
        #endif
        break;
      }
    }

    if ( bridge )  // associated PCI bridge has been found
    {
      rc = mbg_plx_read_pecs_reg( bridge, PLX_PECS_GPIOCTL, &reg_val );

      #if defined( DEBUG )
        printk( KERN_INFO "%s: Read cfg dword %04lX: %08lX, rc: %i\n",
                pcps_driver_name, (ulong) PLX_PECS_GPIOCTL, (ulong) reg_val, rc );
      #endif

      // The pci_get_device() calls above has increased the use count
      // for the last device which has been found, so we need to decrease
      // the use count if we don't need to access the device anymore.
      // This is done by calling the complementary function.
      pci_dev_put( bridge );
    }

  #elif defined( MBG_TGT_WIN32 )

    // We don't scan the PCI bus to find the associated bridge device
    // but just assume the card can flag when it's ready. If the card
    // actually can't then the driver just waits until the maximum
    // required uptime has been reached or exceeded, so there's no danger
    // that the system might lock up.
    rc = PCI_SUCCESS;
    reg_val = 0;

  #elif defined( MBG_TGT_DOS )

    PLX_DEVICE_NODE dn_bridge = { 0 };

    rc = mbg_find_plx8311_bridge( pddev->dev.cfg.bus_num, &dn_bridge );

    if ( rc == PCI_SUCCESS )
      rc = mbg_plx_read_pecs_reg( &dn_bridge, PLX_PECS_GPIOCTL, &reg_val );

  #else

  #endif

  // The device can indicate when it's ready if reg_val
  // has been read successfully, and the GPIO3 bit is 0.
  return ( rc == PCI_SUCCESS ) && ( ( reg_val & PLX_PECS_GPIOCTL_GPIO3_DATA ) == 0 );

}  // ptp270pex_can_flag_ready



static /*HDR*/
/**
 * @brief Check if a PTP270PEX card indicates it is ready
 *
 * A PTP270PEX card must be accessed by the driver only
 * after it has finished booting. Otherwise the host system
 * may be locked up.
 *
 * On HW v2 cards the firmware can indicate when the card is
 * ready to be accessed, so the driver can call this function
 * to check this.
 *
 * @note The function ::ptp270pex_can_flag_ready should
 * have been called before to check if the card actually
 * supports this.
 *
 * @param[in]  pddev   Pointer to the device structure
 *
 * @return true if the card has flagged "ready"
 *
 * @see ::ptp270pex_can_flag_ready
 */
bool ptp270pex_has_flagged_ready( const PCPS_DDEV *pddev )
{
  // GPIO pin USERI is pulled down to 0 by the firmware
  // as soon as the card is ready to be accessed.
  // With PTP270PEX HW v1 cards this pin is always 1.

  // Read the LCS_CNTRL register and return 1 if the USERI
  // bit is set.

  PCPS_IO_ADDR_MAPPED cntrl_reg = _pcps_ddev_io_base_mapped( pddev, 1 )
                                + PLX_LCS_CNTRL;
  uint32_t reg_val = _mbg_inp32_to_cpu( pddev, 0, cntrl_reg );

  return ( reg_val & PLX_LCS_CNTRL_USERI ) == 0;

}  // ptp270pex_has_flagged_ready



#if MBG_TGT_HAS_UPTIME

static /*HDR*/
/**
 * @brief Report the system uptime
 *
 * This is mainly used for debugging, and informational.
 *
 * @param[in]  p_uptime  The system uptime
 *
 * @see ::wait_ptp270pex_ready
 * @see ::MAX_BOOT_TIME_PTP270PEX
 */
void report_uptime( const MBG_SYS_UPTIME *p_uptime )
{
  #if defined( MBG_TGT_WIN32 )
    // Simplified code for Windows since Windows doesn't provide
    // snprintf() or an equivalent function in kernel mode.
    WCHAR wcs_msg[120];

    swprintf( wcs_msg, L"system uptime: %I64u s, required %u s",
              (int64_t) *p_uptime, MAX_BOOT_TIME_PTP270PEX );
    _evt_msg_w( GlbDriverObject, wcs_msg );

  #else

    char ws[120];
    int l = sizeof( ws );
    int n = 0;

    n += snprintf( &ws[n], l - n, "%s: system uptime ", pcps_driver_name );

    #if defined( MBG_TGT_LINUX )
      n += snprintf( &ws[n], l - n, "%llu jiffies -> ",
                     (unsigned long long) ( get_jiffies_64() - INITIAL_JIFFIES ) );
    #endif

    n += snprintf( &ws[n], l - n, "%llu", (unsigned long long) *p_uptime );
    n += snprintf( &ws[n], l - n, " s, required %u s", MAX_BOOT_TIME_PTP270PEX );

    if ( *p_uptime < MAX_BOOT_TIME_PTP270PEX )
    {
      int waiting = (int) ( MAX_BOOT_TIME_PTP270PEX - *p_uptime );
      n += snprintf( &ws[n], l - n, ", waiting %i s", waiting );
    }
    else
      n += snprintf( &ws[n], l - n, ", OK." );

    #if defined( MBG_TGT_LINUX )
      printk( KERN_INFO "%s\n", ws );
    #elif defined( MBG_TGT_BSD )
      printf( "%s\n", ws );
    #endif

  #endif

}  // report_uptime

#endif  // MBG_TGT_HAS_UPTIME



static /*HDR*/
/**
 * @brief Wait until a PTP270PEX card is ready after power-up
 *
 * A PTP270PEX card must be accessed by the driver only
 * after it has finished booting. Otherwise the host system
 * may be locked up.
 *
 * @param[in]  pddev   Pointer to the device structure
 *
 * @see ::ptp270pex_can_flag_ready
 * @see ::ptp270pex_has_flagged_ready
 * @see ::report_uptime
 * @see ::MAX_BOOT_TIME_PTP270PEX
 */
void wait_ptp270pex_ready( const PCPS_DDEV *pddev )
{
  MBG_SYS_TIME t1;
  MBG_SYS_TIME t2;
  MBG_SYS_UPTIME uptime;
  int delayed = 0;
  bool can_flag_ready;
  int i = 0;

  // A newer PTP270PEX card (HW v2.0) can indicate if it has
  // finished booting and thus is ready to be accessed, but
  // older PTP270PEX cards (HW v1.0) don't support this.
  can_flag_ready = ptp270pex_can_flag_ready( pddev );

  mbg_get_sys_time( &t1 );

  for (;;)
  {
    mbg_get_sys_uptime( &uptime );

    #if MBG_TGT_HAS_UPTIME
      #if !defined( DEBUG )
        if ( delayed )
      #endif
          report_uptime( &uptime );
    #endif

    if ( can_flag_ready )
    {
      if ( ptp270pex_has_flagged_ready( pddev ) )
        break;
    }
    else
    {
      // If the card is unable to indicate when it is ready
      // and the target system doesn't support uptime
      // then it makes no sense to keep looping and waiting,
      // so just stop waiting and hope the card is ready.
      if ( uptime == 0 || uptime == -1 )
        break;  // assume uptime not supported
    }

    // If the target system supports uptime then
    // keep looping until the system uptime exceeds
    // the max. boot time required by the card.

    if ( uptime >= MAX_BOOT_TIME_PTP270PEX )
      break;

    mbg_sleep_sec( 1 );
    delayed = 1;

    if ( ++i >= MAX_BOOT_TIME_PTP270PEX )
      break;
  }

  if ( delayed )
  {
    long dt;

    mbg_get_sys_time( &t2 );

    dt = mbg_delta_sys_time_ms( &t2, &t1 );

    #if defined( MBG_TGT_LINUX )
      printk( KERN_INFO "PTP270PEX startup delay: %li.%03li s\n",
              dt / 1000, ( ( dt < 0 ) ? -dt : dt ) % 1000 );
    #elif defined( MBG_TGT_BSD )
      printf( "PTP270PEX startup delay: %li.%03li s\n",
              dt / 1000, ( ( dt < 0 ) ? -dt : dt ) % 1000 );
    #elif defined( MBG_TGT_WIN32 )
    {
      WCHAR wcs_msg[128];
      swprintf( wcs_msg, L"PTP270PEX startup delay: %li.%03li s",
                dt / 1000, ( ( dt < 0 ) ? -dt : dt ) % 1000 );
      _evt_msg_w( GlbDriverObject, wcs_msg );
    }
    #else
      (void) dt;
    #endif
  }

}  // wait_ptp270pex_ready



static /*HDR*/
/**
 * @brief Check if IRQ usage with a device is unsafe
 *
 * A few early PCI Express cards had a problem where usage
 * of IRQs was unsafe. Firmware upgrades are available to fix this.
 * See http://www.meinberg.de/english/info/pex-upgrades.htm
 *
 * @param[in]  pddev               Pointer to the device structure
 * @param[in]  req_fw_ver          Required firmware version
 * @param[in]  req_asic_ver_major  Required ASIC major version
 * @param[in]  req_asic_ver_minor  Required ASIC major version
 *
 * @return true if IRQ operation is unsafe
 */
bool pcps_check_pex_irq_unsafe( PCPS_DDEV *pddev, uint16_t req_fw_ver,
                                uint8_t req_asic_ver_major, uint8_t req_asic_ver_minor )
{
  bool b = !_pcps_pex_irq_is_safe( _pcps_ddev_fw_rev_num( pddev ), req_fw_ver,
                                   _pcps_ddev_asic_version( pddev ),
                                   req_asic_ver_major, req_asic_ver_minor );

  if ( b )
  {
    pddev->irq_stat_info |= PCPS_IRQ_STAT_UNSAFE;

    // Prevent the driver from writing IRQ ACK to the card even if IRQs
    // should be unintentionally enabled.
    pddev->irq_ack_port = 0;
    pddev->irq_ack_mask = 0;
  }

  return b;

}  // pcps_check_pex_irq_unsafe



#if MBG_TGT_SUPP_MEM_ACC

static __mbg_inline /*HDR*/
/**
 * @brief Check if a virtual address has been mapped for a device
 *
 * @param[in]  pddev  Pointer to the device structure
 *
 * @return  true if virtual address has been mapped
 */
bool has_mapped_sys_virtual_address( PCPS_DDEV *pddev )
{
  return pddev->mm_addr != NULL;

}  // has_mapped_sys_virtual_address



static __mbg_inline /*HDR*/
/**
 * @brief Map a virtual address for a device
 *
 * @param[in]  pddev  Pointer to the device structure
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
int map_sys_virtual_address( PCPS_DDEV *pddev )
{
  pddev->mm_tstamp_addr = NULL;  // unless configured below

  #if defined( MBG_TGT_WIN32 )
  {
    PHYSICAL_ADDRESS pAD;

    pAD.QuadPart = pddev->rsrc_info.mem[0].start;
    pddev->mm_addr = MmMapIoSpace( pAD, sizeof( *pddev->mm_addr ), MmNonCached );
  }
  #elif defined( MBG_TGT_LINUX )

    pddev->mm_addr = ioremap( ( (ulong) pddev->rsrc_info.mem[0].start ), sizeof( *pddev->mm_addr ) );

  #elif defined( MBG_TGT_FREEBSD )

    pddev->mm_addr = rman_get_virtual( pddev->rsrc_info.mem[0].bsd.res );

  #elif defined( MBG_TGT_NETBSD )

    pddev->mm_addr = bus_space_vaddr( pddev->rsrc_info.mem[0].bsd.bst, pddev->rsrc_info.mem[0].bsd.bsh );

  #else  // DOS, ...

    pddev->mm_addr = (PCPS_MM_LAYOUT FAR *) pddev->rsrc_info.mem[0].start;

  #endif  // target specific code

  if ( pddev->mm_addr == NULL )
  {
    _mbgddmsg_2( MBG_DBG_ERR, "%s: %s Failed to map virtual address",
                 pcps_driver_name, _pcps_ddev_type_name( pddev ) );
    return -1;
  }

  if ( _pcps_ddev_is_pci_mbgpex( pddev ) )
    pddev->mm_tstamp_addr = &pddev->mm_addr->mbgpex.tstamp;
  else
    if ( _pcps_ddev_is_pci_pex8311( pddev ) )
      pddev->mm_tstamp_addr = &pddev->mm_addr->pex8311.tstamp;

#if 0 && defined( DEBUG ) // ### TODO
  _mbgddmsg_3( MBG_DBG_INIT_DEV, "MM addr: base: 0x%p, tstamp: 0x%p, offs: 0x%02lX",
               pddev->mm_addr, pddev->mm_tstamp_addr,
               (ulong) ( (uint8_t *) pddev->mm_tstamp_addr - (uint8_t *) pddev->mm_addr ) );
#endif

  return MBG_SUCCESS;

}  // map_sys_virtual_address



static __mbg_inline /*HDR*/
/**
 * @brief Unmap a virtual address for a device, if it has been mapped before
 *
 * @param[in]  pddev  Pointer to the device structure
 */
void unmap_sys_virtual_address( PCPS_DDEV *pddev )
{

  if ( has_mapped_sys_virtual_address( pddev ) )
  {
    #if defined( MBG_TGT_WIN32 )
      MmUnmapIoSpace( pddev->mm_addr, sizeof( *pddev->mm_addr ) );
    #elif defined( MBG_TGT_LINUX )
      iounmap( pddev->mm_addr );
    #else  // DOS, ...
      // nothing to do
    #endif

    pddev->mm_addr = NULL;
    pddev->mm_tstamp_addr = NULL;
  }

}  // unmap_sys_virtual_address

#endif  // MBG_TGT_SUPP_MEM_ACC



#if ( DEBUG_ACCESS_TIMING || DEBUG_IO_TIMING )

#if ( 1 && defined( MBG_TGT_LINUX ) && defined( MBG_ARCH_X86 ) )
  #define DEBUG_PRINT_ACCESS_TIMES   1
#else
  #define DEBUG_PRINT_ACCESS_TIMES   0
#endif


#if DEBUG_PRINT_ACCESS_TIMES

static __mbg_inline /*HDR*/
/**
 * @brief Convert TSC cycles to picoseconds
 *
 * @note This has not yet been implemented for all target systems.
 *
 * A specific cycles frequency has to be provided by the target OS.
 *
 * @param[in]  cyc  The number of TSC cycles
 *
 * @return the computed number of picoseconds
 */
long long _cyc_to_ps( long long cyc )
{
  cyc *= 1000000000UL;

  #if defined( MBG_TGT_LINUX )
    // The variable cpu_khz is exported by the kernel, and the
    // do_div() function needs to be used in kernel space
    // for the division.
    do_div( cyc, cpu_khz );
  #else
    cyc = 0;  //### TODO FIXME for different targets
  #endif

  return cyc;
}

#endif  // DEBUG_PRINT_ACCESS_TIMES

#endif  // ( DEBUG_ACCESS_TIMING || DEBUG_IO_TIMING )



#if DEBUG_ACCESS_TIMING

static uint32_t debug_dummy_var;

static /*HDR*/
/**
 * @brief Report access timing for a device
 *
 * This is only used for testing.
 *
 * @param[in]  pddev           Pointer to the device structure
 * @param[in]  info            Informational string to be printed
 * @param[in]  t_after_cmd     Cycles value taken after command code has been written
 * @param[in]  t_after_reread  Cycles value taken after a data word has been re-read
 */
void report_access_timing( const PCPS_DDEV *pddev, const char *info,
                           MBG_PC_CYCLES t_after_cmd, MBG_PC_CYCLES t_after_reread )
{
  #if defined( MBG_TGT_LINUX )
    long long delta_cyc_write = t_after_cmd - pddev->acc_cycles;
    long long delta_cyc_read = t_after_reread - t_after_cmd;

    #if DEBUG_PRINT_ACCESS_TIMES
      long long delta_t_write = _cyc_to_ps( delta_cyc_write );
      long long delta_t_read = _cyc_to_ps( delta_cyc_read );
    #endif

    printk( KERN_INFO "%s %s: %lli/%lli cyc"
            #if DEBUG_PRINT_ACCESS_TIMES
              ", %lli/%lli ps @ %lu kHz"
            #endif
            "\n",
            _pcps_ddev_type_name( pddev ), info,
            delta_cyc_write, delta_cyc_read
            #if DEBUG_PRINT_ACCESS_TIMES
              , delta_t_write, delta_t_read, (ulong) cpu_khz
            #endif
          );
  #endif

}  // report_access_timing

#endif  // DEBUG_ACCESS_TIMING



#if DEBUG_IO_TIMING

static /*HDR*/
/**
 * @brief Report I/O timing for a device
 *
 * This is only used for testing.
 *
 * @param[in]  pddev         Pointer to the device structure
 * @param[in]  info          Informational string to be printed
 * @param[in]  cmd           The command by tent to the device
 * @param[in]  count         The number of bytes read from the device
 * @param[in]  t_after_cmd   Cycles value taken after command code has been written
 * @param[in]  t_after_busy  Cycles value taken after busy flag was cleard by the device
 * @param[in]  t_done        Cycles value taken after bytes have been read
 */
void report_io_timing( const PCPS_DDEV *pddev, const char *info,
                       uint8_t cmd, uint16_t count, MBG_PC_CYCLES t_after_cmd,
                       MBG_PC_CYCLES t_after_busy, MBG_PC_CYCLES t_done )
{
  #if defined( MBG_TGT_LINUX )  //##++++ TODO support other OSs
    long long cmd_cycles = t_after_cmd - pddev->acc_cycles;
    long long busy_cycles = t_after_busy - t_after_cmd;
    long long read_cycles = t_done - t_after_busy;
    long long cycles_per_read = 0;

    #if DEBUG_PRINT_ACCESS_TIMES
      long long cmd_time = _cyc_to_ps( cmd_cycles );
      long long busy_time = _cyc_to_ps( busy_cycles );
      long long read_time = _cyc_to_ps( read_cycles );
      long long time_per_read = 0;
    #endif

    if ( count )
    {
      cycles_per_read = read_cycles;
      do_div( cycles_per_read, count );

      #if DEBUG_PRINT_ACCESS_TIMES
        time_per_read = read_time;
        do_div( time_per_read, count );
      #endif
    }

    printk( KERN_INFO "%s %s cmd: 0x%02X (%u bytes), "
            "write %lli cyc, busy: %lli cyc, read: %lli/%lli cyc"
            #if DEBUG_PRINT_ACCESS_TIMES
              ",\n  write %lli ps, busy: %lli ps, read: %lli/%lli ps @ %lu kHz"
            #endif
            "\n",
            _pcps_ddev_type_name( pddev ), info, cmd, count,
            cmd_cycles, busy_cycles, read_cycles, cycles_per_read
            #if DEBUG_PRINT_ACCESS_TIMES
              , cmd_time, busy_time, read_time, time_per_read, (ulong) cpu_khz
            #endif
          );
    printk( KERN_ERR "cycles after cmd: %lli, after busy: %lli, when done: %lli\n",
            (long long) t_after_cmd,
            (long long) t_after_busy,
            (long long) t_done );

  #endif

}  // report_io_timing

#endif  // DEBUG_IO_TIMING



#if DEBUG_USB_IO

// This is the prototype of the standard usb_bulk_msg() prototype
// provided by the Linux kernel. We include it here to make sure
// our debug wrapper function usb_bulk_msg_dbg() has the same
// prototype, otherwise we'd expect a compiler warning.

int usb_bulk_msg( struct usb_device *usb_dev,
                  unsigned int pipe,
                  void *data,
                  int len,
                  int *actual_length,
                  int timeout );


  /*HDR*/  //### TODO Do we need a prototype?
int usb_bulk_msg_dbg( struct usb_device *usb_dev,
                      unsigned int pipe,
                      void *data,
                      int len,
                      int *actual_length,
                      int timeout )
{
  int rc = usb_bulk_msg( usb_dev, pipe, data, len, actual_length, timeout );

  _mbgddmsg_6( MBG_DBG_ERR, "usb_bulk_msg: pipe %u, data %p, len %i, actual_len %i, timeout %i: returned %i",
               pipe, data, len, *actual_length, timeout, rc );

  return rc;

}  // usb_bulk_msg_dbg

#endif  // DEBUG_USB_IO



#if defined( __GNUC__ )
// Avoid "no previous prototype" with some gcc versions.
__mbg_inline
int pcps_wait_busy( PCPS_DDEV *pddev );
#endif

__mbg_inline /*HDR*/
/**
 * @brief Wait as long as a device is busy, or until timeout
 *
 * Used by the @ref pcps_read_fncs to wait after the command byte has been
 * written until the requested data has been made available by the device.
 *
 * @param[in]  pddev   Pointer to the device structure
 *
 * @return ::MBG_SUCCESS on success, or ::MBG_ERR_TIMEOUT
 *
 * @see @ref pcps_read_fncs
 */
int pcps_wait_busy( PCPS_DDEV *pddev )
{
  if ( _pcps_ddev_status_busy( pddev ) )
  {
    #if defined( MBG_TGT_BSD )
      struct timeval tv_start;

      getmicrouptime( &tv_start );

      while ( _pcps_ddev_status_busy( pddev ) )
      {
        struct timeval tv_now;
        long long delta_ms;

        getmicrouptime( &tv_now );
        delta_ms = ( ( tv_now.tv_sec - tv_start.tv_sec ) * 1000 )
                 + ( ( tv_now.tv_usec - tv_start.tv_usec ) / 1000 );
        if ( delta_ms > _pcps_ddev_timeout_clk( pddev ) )
          return MBG_ERR_TIMEOUT;
      }
    #elif _PCPS_USE_CLOCK_TICK
      clock_t timeout_val = clock() + _pcps_ddev_timeout_clk( pddev );

      while ( _pcps_ddev_status_busy( pddev ) )
        if ( _pcps_time_after( clock(), timeout_val ) )
          return MBG_ERR_TIMEOUT;
    #else
      long cnt = _pcps_ddev_timeout_clk( pddev );

      for ( ; _pcps_ddev_status_busy( pddev ); cnt-- )
        if ( cnt == 0 )
          return MBG_ERR_TIMEOUT;
    #endif
  }

  return MBG_SUCCESS;

}  // pcps_wait_busy



/**
 * @defgroup pcps_read_fncs Device read functions
 *
 * These group of functions is used for low level access to Meinberg
 * bus-level devices. Which of the function is actually to be used
 * depends on the device's bus type and interface chip and is determined
 * by the function ::pcps_probe_device which is called at device startup.
 *
 * @see ::pcps_read_null
 * @see ::pcps_read_std
 * @see ::pcps_read_amcc_s5933
 * @see ::pcps_read_amcc_s5920
 * @see ::pcps_read_asic
 * @see ::pcps_read_usb
 *
 * @{ */


static /*HDR*/
/**
 * @brief Dummy read function
 *
 * Used if a clock is not properly initialized, in order to
 * avoid I/O access on unspecified ports.
 *
 * @param[in]  pddev   Pointer to the device structure
 * @param[in]  cmd     The command code for the board, see @ref PCPS_CMD_CODES
 * @param[out] buffer  A buffer to take the bytes to be read
 * @param[in]  count   The number of bytes to be read into the buffer
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 *
 * @ingroup pcps_read_fncs
 * @see @ref pcps_read_fncs
 */
int pcps_read_null( PCPS_DDEV *pddev, uint8_t cmd,
                    void FAR *buffer, uint16_t count )
{
  // avoid compiler warnings
  (void) pddev;
  (void) cmd;
  (void) buffer;
  (void) count;

  return MBG_ERR_TIMEOUT;

}  // pcps_read_null

#if !defined( AVOID_REDUNDANT_REDECLARATION )
  PCPS_READ_FNC pcps_read_null;
#endif


static /*HDR*/
/**
 * @brief Read function for devices with ISA or micro channel bus
 *
 * @param[in]  pddev   Pointer to the device structure
 * @param[in]  cmd     The command code for the board, see @ref PCPS_CMD_CODES
 * @param[out] buffer  A buffer to take the bytes to be read
 * @param[in]  count   The number of bytes to be read into the buffer
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 *
 * @ingroup pcps_read_fncs
 * @see @ref pcps_read_fncs
 */
int pcps_read_std( PCPS_DDEV *pddev, uint8_t cmd,
                   void FAR *buffer, uint16_t count )
{
  uint8_t FAR *p = (uint8_t FAR *) buffer;
  PCPS_IO_ADDR_MAPPED port = _pcps_ddev_io_base_mapped( pddev, 0 );
  int i;
  _pcps_irq_flags

  #if DEBUG_IO
    _mbgddmsg_1( MBG_DBG_INIT_DEV, "pcps_read_std: cmd %02X", cmd );
  #endif

  _pcps_disb_local_irq_save();
  mbg_get_pc_cycles( &pddev->acc_cycles );
  // write the command byte
  _mbg_outp8( pddev, 0, port, cmd );
  _pcps_local_irq_restore();

  // wait until BUSY flag goes low or timeout
  if ( pcps_wait_busy( pddev ) < 0 )
    return MBG_ERR_TIMEOUT;


  // no timeout: read bytes from the board's FIFO
  for ( i = 0; i < count; i++ )
  {
    *p = _mbg_inp8( pddev, 0, port );

    #if DEBUG_IO
      _mbgddmsg_1( MBG_DBG_DETAIL, "pcps_read_std: %02X", *p );
    #endif

    p++;
  }

  return MBG_SUCCESS;

}  // pcps_read_std

#if !defined( AVOID_REDUNDANT_REDECLARATION )
  PCPS_READ_FNC pcps_read_std;
#endif


#if _PCPS_USE_PCI

static /*HDR*/
/**
 * @brief Read function for devices with AMCC S5933 PCI interface chip
 *
 * @param[in]  pddev   Pointer to the device structure
 * @param[in]  cmd     The command code for the board, see @ref PCPS_CMD_CODES
 * @param[out] buffer  A buffer to take the bytes to be read
 * @param[in]  count   The number of bytes to be read into the buffer
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 *
 * @ingroup pcps_read_fncs
 * @see @ref pcps_read_fncs
 */
int pcps_read_amcc_s5933( PCPS_DDEV *pddev, uint8_t cmd,
                          void FAR *buffer, uint16_t count )
{
  uint8_t FAR *p = (uint8_t FAR *) buffer;
  PCPS_IO_ADDR_MAPPED port = _pcps_ddev_io_base_mapped( pddev, 0 );
  int i;
  _pcps_irq_flags


  #if DEBUG_IO
    _mbgddmsg_1( MBG_DBG_INIT_DEV, "pcps_read_amcc_s5933: cmd %02X", cmd );
  #endif

  // reset inbound mailbox and FIFO status
  _mbg_outp8( pddev, 0, port + AMCC_OP_REG_MCSR + 3, 0x0C );

  // set FIFO
  _mbg_outp8( pddev, 0, port + AMCC_OP_REG_INTCSR + 3, 0x3C );

  _pcps_disb_local_irq_save();
  mbg_get_pc_cycles( &pddev->acc_cycles );
  // write the command byte
  _mbg_outp8( pddev, 0, port + AMCC_OP_REG_OMB1, cmd );
  #if defined( MBG_ARCH_SPARC )
    udelay( 3 );
  #endif
  _pcps_local_irq_restore();

  // wait until BUSY flag goes low or timeout
  if ( pcps_wait_busy( pddev ) < 0 )
    return MBG_ERR_TIMEOUT;


  // no timeout: read bytes from the board's FIFO
  for ( i = 0; i < count; i++ )
  {
    if ( _mbg_inp16_to_cpu( pddev, 0, port + AMCC_OP_REG_MCSR ) & 0x20 )
      return MBG_ERR_FIFO;

    p[i] = _mbg_inp8( pddev, 0, port + AMCC_OP_REG_FIFO + ( i % sizeof( uint32_t) ) );

    #if DEBUG_IO
      if ( ( cmd == PCPS_GIVE_FW_ID_1 ) || ( cmd == PCPS_GIVE_FW_ID_2 ) )
        _mbgddmsg_2( MBG_DBG_DETAIL, "pcps_read_amcc_s5933: %02X '%c'", p[i], p[i] );
      else
        _mbgddmsg_1( MBG_DBG_DETAIL, "pcps_read_amcc_s5933: %02X", p[i] );
    #endif
  }

  return MBG_SUCCESS;

}  // pcps_read_amcc_s5933

#if !defined( AVOID_REDUNDANT_REDECLARATION )
  PCPS_READ_FNC pcps_read_amcc_s5933;
#endif

#endif  /* _PCPS_USE_PCI */



#if _PCPS_USE_PCI

static /*HDR*/
/**
 * @brief Read function for devices with AMCC S5920 PCI interface chip
 *
 * @param[in]  pddev   Pointer to the device structure
 * @param[in]  cmd     The command code for the board, see @ref PCPS_CMD_CODES
 * @param[out] buffer  A buffer to take the bytes to be read
 * @param[in]  count   The number of bytes to be read into the buffer
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 *
 * @ingroup pcps_read_fncs
 * @see @ref pcps_read_fncs
 */
int pcps_read_amcc_s5920( PCPS_DDEV *pddev, uint8_t cmd,
                          void FAR *buffer, uint16_t count )
{
  uint8_t FAR *p = (uint8_t FAR *) buffer;
  PCPS_IO_ADDR_MAPPED data_port = _pcps_ddev_io_base_mapped( pddev, 1 );
  int i;
  int dt_quot;
  int dt_rem;
  _pcps_irq_flags


  #if DEBUG_IO
    _mbgddmsg_5( MBG_DBG_INIT_DEV, "pcps_read_amcc_s5920: cmd %02X, port: %04lX, data_port: %04lX, buffer: %p, count: %u",
                 cmd, (ulong) _pcps_ddev_io_base_mapped( pddev, 0 ) + AMCC_OP_REG_OMB,
                 (ulong) data_port, buffer, count );
  #endif

  _pcps_disb_local_irq_save();
  mbg_get_pc_cycles( &pddev->acc_cycles );
  // write the command byte
  _mbg_outp8( pddev, 0, _pcps_ddev_io_base_mapped( pddev, 0 ) + AMCC_OP_REG_OMB, cmd );
  #if defined( MBG_ARCH_SPARC )
    udelay( 3 );
  #endif
  _pcps_local_irq_restore();

  dt_quot = count / 4;
  dt_rem = count % 4;

  // wait until BUSY flag goes low or timeout
  if ( pcps_wait_busy( pddev ) < 0 )
    return MBG_ERR_TIMEOUT;


  if ( count )
  {
    // do this only if we must read data

    uint32_t ul;

    // first read full 32 bit words
    for ( i = 0; i < dt_quot; i++ )
    {
      ul = _mbg_inp32_to_cpu( pddev, 1, data_port );
      #if DEBUG_IO
        if ( ( cmd == PCPS_GIVE_FW_ID_1 ) || ( cmd == PCPS_GIVE_FW_ID_2 ) )
        {
          _mbgddmsg_5( MBG_DBG_INIT_DEV, "pcps_read_amcc_s5920: %08X  \"%c%c%c%c\"", ul,
                       BYTE_OF( ul, 0 ), BYTE_OF( ul, 1 ), BYTE_OF( ul, 2 ), BYTE_OF( ul, 3 ) );
        }
        else
          _mbgddmsg_1( MBG_DBG_INIT_DEV, "pcps_read_amcc_s5920: %08X", ul );
      #endif
      _mbg_put_unaligned( ul, (uint32_t FAR *) p );
      p += sizeof( ul );
    }

    // then read the remaining bytes, if required
    if ( dt_rem )
    {
      ul = _mbg_inp32_to_cpu( pddev, 1, data_port );

      for ( i = 0; i < dt_rem; i++ )
      {
        #if DEBUG_IO
          _mbgddmsg_1( MBG_DBG_INIT_DEV, "pcps_read_amcc_s5920: %02X", BYTE_OF( ul, i ) );
        #endif

        *p++ = BYTE_OF( ul, i );
      }
    }
  }
  else
    _mbg_inp32( pddev, 1, data_port );  // do a dummy read

  return MBG_SUCCESS;

}  // pcps_read_amcc_s5920

#if !defined( AVOID_REDUNDANT_REDECLARATION )
  PCPS_READ_FNC pcps_read_amcc_s5920;
#endif

#endif  /* _PCPS_USE_PCI */



#if _PCPS_USE_PCI

static /*HDR*/
/**
 * @brief Read function for devices with Meinberg PCI interface ASIC
 *
 * @param[in]  pddev   Pointer to the device structure
 * @param[in]  cmd     The command code for the board, see @ref PCPS_CMD_CODES
 * @param[out] buffer  A buffer to take the bytes to be read
 * @param[in]  count   The number of bytes to be read into the buffer
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 *
 * @ingroup pcps_read_fncs
 * @see ::pcps_read_asic_mm
 * @see @ref pcps_read_fncs
 */
int pcps_read_asic( PCPS_DDEV *pddev, uint8_t cmd,
                    void FAR *buffer, uint16_t count )
{
  short ret_val = MBG_SUCCESS;
  uint8_t FAR *p = (uint8_t FAR *) buffer;
  PCPS_IO_ADDR_MAPPED data_port;
  PCI_ASIC_REG ar;
  int i;
  int dt_quot;
  int dt_rem;
  _pcps_irq_flags
  #if DEBUG_ACCESS_TIMING || DEBUG_IO_TIMING
    MBG_PC_CYCLES t_after_cmd = 0;
  #endif

  #if DEBUG_ACCESS_TIMING
    MBG_PC_CYCLES t_after_reread = 0;
  #endif

  #if DEBUG_IO_TIMING
    MBG_PC_CYCLES t_after_busy = 0;
    MBG_PC_CYCLES t_done = 0;
  #endif

  #if DEBUG_IO
    _mbgddmsg_3( MBG_DBG_INIT_DEV, "pcps_read_asic: cmd: 0x%02X (0x%08X), cnt: %u",
                 cmd, _cpu_to_mbg32( cmd ), count );
  #endif

  _pcps_disb_local_irq_save();

  // get current cycles and write the command byte
  mbg_get_pc_cycles( &pddev->acc_cycles );
  _mbg_outp32( pddev, 0, _pcps_ddev_io_base_mapped( pddev, 0 )
               + offsetof( PCI_ASIC, pci_data ), cmd );

  #if DEBUG_ACCESS_TIMING || DEBUG_IO_TIMING
    mbg_get_pc_cycles( &t_after_cmd );
  #endif

  #if DEBUG_ACCESS_TIMING
    debug_dummy_var = _mbg_inp32_to_cpu( pddev, 0, _pcps_ddev_io_base_mapped( pddev, 0 )
                                         + offsetof( PCI_ASIC, addon_data ) );
    mbg_get_pc_cycles( &t_after_reread );
  #endif


  _pcps_local_irq_restore();

  data_port = _pcps_ddev_io_base_mapped( pddev, 0 )
                + offsetof( PCI_ASIC, addon_data );
  dt_quot = count / 4;
  dt_rem = count % 4;

  // wait until BUSY flag goes low or timeout
  if ( pcps_wait_busy( pddev ) < 0 )
  {
    #if DEBUG_IO_TIMING
      mbg_get_pc_cycles( &t_after_busy );
    #endif
    ret_val = MBG_ERR_TIMEOUT;
    goto done;
  }


  #if DEBUG_IO_TIMING
    mbg_get_pc_cycles( &t_after_busy );
  #endif

  // no timeout: read bytes from the board's FIFO

  // first read full 32 bit words
  for ( i = 0; i < dt_quot; i++ )
  {
    ar.ul = _mbg_inp32_to_cpu( pddev, 0, data_port );
    #if DEBUG_IO
      _mbgddmsg_1( MBG_DBG_INIT_DEV, "pcps_read_asic: %08X", ar.ul );
    #endif
    _mbg_put_unaligned( ar.ul, (uint32_t FAR *) p );
    p += sizeof( ar.ul );
    data_port += sizeof( ar.ul );
  }

  // then read the remaining bytes, if required
  if ( dt_rem )
  {
    ar.ul = _mbg_inp32_to_cpu( pddev, 0, data_port );

    for ( i = 0; i < dt_rem; i++ )
    {
      #if DEBUG_IO
        _mbgddmsg_1( MBG_DBG_INIT_DEV, "pcps_read_asic: %02X", ar.b[i] );
      #endif

      *p++ = ar.b[i];
    }
  }

done:
  #if DEBUG_IO_TIMING
    mbg_get_pc_cycles( &t_done );
    report_io_timing( pddev, "ASIC", cmd, count, t_after_cmd, t_after_busy, t_done );
  #endif

  #if DEBUG_ACCESS_TIMING
    report_access_timing( pddev, "ASIC wr/rd", t_after_cmd, t_after_reread );
  #endif

  return ret_val;

}  // pcps_read_asic

#if !defined( AVOID_REDUNDANT_REDECLARATION )
  PCPS_READ_FNC pcps_read_asic;
#endif


#if _PCPS_USE_MM_IO

static /*HDR*/
/**
 * @brief Read function for devices with Meinberg PCI interface ASIC
 *
 * Unlike ::pcps_read_asic and most of the other read functions,
 * this function accesses the memory mapped registers rather than
 * the I/O ports, so this is significantly faster.
 *
 * @param[in]  pddev   Pointer to the device structure
 * @param[in]  cmd     The command code for the board, see @ref PCPS_CMD_CODES
 * @param[out] buffer  A buffer to take the bytes to be read
 * @param[in]  count   The number of bytes to be read into the buffer
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 *
 * @ingroup pcps_read_fncs
 * @see ::pcps_read_asic
 * @see @ref pcps_read_fncs
 */
int pcps_read_asic_mm( PCPS_DDEV *pddev, uint8_t cmd,
                       void FAR *buffer, uint16_t count )
{
  short ret_val = MBG_SUCCESS;
  uint8_t FAR *p = (uint8_t FAR *) buffer;
  volatile uint32_t *p_data_reg;
  PCI_ASIC_REG ar;
  int i;
  int dt_quot;
  int dt_rem;
  _pcps_irq_flags

  #if DEBUG_ACCESS_TIMING || DEBUG_IO_TIMING
    MBG_PC_CYCLES t_after_cmd = 0;
  #endif

  #if DEBUG_ACCESS_TIMING
    MBG_PC_CYCLES t_after_reread = 0;
  #endif

  #if DEBUG_IO_TIMING
    MBG_PC_CYCLES t_after_busy = 0;
    MBG_PC_CYCLES t_done = 0;
  #endif

  #if DEBUG_IO
    _mbgddmsg_3( MBG_DBG_INIT_DEV, "pcps_read_asic_mm: cmd: 0x%02X (0x%08X), cnt: %u",
                 cmd, _cpu_to_mbg32( cmd ), count );
  #endif

  _pcps_disb_local_irq_save();

  // get current cycles and write the command byte
  #if USE_CMD_PTR
  {
    volatile uint32_t *p_cmd_reg = &pddev->mm_addr->mbgpex.asic.pci_data.ul;
    mbg_get_pc_cycles( &pddev->acc_cycles );
    *p_cmd_reg = cmd;
  }
  #else
    mbg_get_pc_cycles( &pddev->acc_cycles );
    pddev->mm_addr->mbgpex.asic.pci_data.ul = cmd;
  #endif

  #if DEBUG_ACCESS_TIMING || DEBUG_IO_TIMING
    mbg_get_pc_cycles( &t_after_cmd );
  #endif

  #if DEBUG_ACCESS_TIMING
    debug_dummy_var = pddev->mm_addr->mbgpex.asic.pci_data.ul;
    mbg_get_pc_cycles( &t_after_reread );
  #endif

  _pcps_local_irq_restore();

  p_data_reg = &pddev->mm_addr->mbgpex.asic.addon_data.ul[0];
  dt_quot = count / 4;
  dt_rem = count % 4;

  // wait until BUSY flag goes low or timeout
  if ( pcps_wait_busy( pddev ) < 0 )
  {
    #if DEBUG_IO_TIMING
      mbg_get_pc_cycles( &t_after_busy );
    #endif
    ret_val = MBG_ERR_TIMEOUT;
    goto done;
  }


  #if DEBUG_IO_TIMING
    mbg_get_pc_cycles( &t_after_busy );
  #endif

  // no timeout: read bytes from the board's FIFO

  // first read full 32 bit words
  for ( i = 0; i < dt_quot; i++ )
  {
    ar.ul = *p_data_reg;
    #if DEBUG_IO
      _mbgddmsg_1( MBG_DBG_INIT_DEV, "pcps_read_asic_mm: %08X", ar.ul );
    #endif
    _mbg_put_unaligned( ar.ul, (uint32_t FAR *) p );
    p += sizeof( ar.ul );
    p_data_reg++;
  }

  // then read the remaining bytes, if required
  if ( dt_rem )
  {
    ar.ul = *p_data_reg;

    for ( i = 0; i < dt_rem; i++ )
    {
      #if DEBUG_IO
      #endif

      *p++ = ar.b[i];
    }
  }

done:
  #if DEBUG_IO_TIMING
    mbg_get_pc_cycles( &t_done );
    report_io_timing( pddev, "ASIC MM", cmd, count, t_after_cmd, t_after_busy, t_done );
  #endif

  #if DEBUG_ACCESS_TIMING
    report_access_timing( pddev, "ASIC MM wr/rd", t_after_cmd, t_after_reread );
  #endif

  return ret_val;

}  // pcps_read_asic_mm

#if !defined( AVOID_REDUNDANT_REDECLARATION )
  PCPS_READ_FNC pcps_read_asic_mm;
#endif

#endif  // _PCPS_USE_MM_IO

#endif  // _PCPS_USE_PCI



#if _PCPS_USE_USB

static /*HDR*/
/**
 * @brief Read function for USB devices
 *
 * @param[in]  pddev   Pointer to the device structure
 * @param[in]  cmd     The command code for the board, see @ref PCPS_CMD_CODES
 * @param[out] buffer  A buffer to take the bytes to be read
 * @param[in]  count   The number of bytes to be read into the buffer
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 *
 * @ingroup pcps_read_fncs
 * @see @ref pcps_read_fncs
 */
int pcps_read_usb( PCPS_DDEV *pddev, uint8_t cmd,
                   void FAR *buffer, uint16_t count )
{
  int actual_count = 0;
  short rc;

  mbg_get_pc_cycles( &pddev->acc_cycles );

  rc = _pcps_usb_write_var( pddev, &cmd );

  #if DEBUG_USB_IO
    if ( rc == 0 )
      _mbgddmsg_4( MBG_DBG_INFO, "%s: %s _pcps_usb_write_var cmd 0x%02X, %i bytes: succeeded",
                   pcps_driver_name, _pcps_ddev_type_name( pddev ), cmd, actual_count );
    else
      _mbgddmsg_5( MBG_DBG_ERR, "%s: %s _pcps_usb_write_var cmd 0x%02X, %i bytes: failed with rc %i",
                   pcps_driver_name, _pcps_ddev_type_name( pddev ), cmd, actual_count, rc );
  #endif  // DEBUG_USB_IO

  if ( ( rc == 0 ) && ( count && buffer ) )
  {
    #if defined( MBG_TGT_WIN32_PNP )
      #if USE_USB_MICRO_FRAMES
        int temp_fn1 = micro_frame_number_1 ? micro_frame_number_1 : frame_number_1;
        int temp_fn2 = micro_frame_number_2 ? micro_frame_number_2 : frame_number_2;
      #else
        int temp_fn1 = frame_number_1;
        int temp_fn2 = frame_number_2;
      #endif

      LARGE_INTEGER UsbPreCount  = Count1;
      LARGE_INTEGER UsbPostCount = Count2;
    #endif

    rc = _pcps_usb_read( pddev, buffer, count );

    #if DEBUG_USB_IO
      if ( rc == 0 )
        _mbgddmsg_4( MBG_DBG_INFO, "%s: %s _pcps_usb_read after cmd 0x%02X succeeded, bytes read: %i",
                     pcps_driver_name, _pcps_ddev_type_name( pddev ), cmd, actual_count );
      else
        _mbgddmsg_5( MBG_DBG_ERR, "%s: %s _pcps_usb_read after cmd 0x%02X failed after with rc %i, bytes read: %i",
                     pcps_driver_name, _pcps_ddev_type_name( pddev ), cmd, rc, actual_count );
    #endif  // DEBUG_USB_IO

    #if defined( MBG_TGT_WIN32_PNP )

      if ( cmd == PCPS_GIVE_HR_TIME && rc == 0 )
      {
        ULONGLONG usb_latency_cycles;
        ULONGLONG cycles_diff;
        ULONGLONG time_diff;
        ULONGLONG frame_length_cycles;
        int FrameNumberDiff;

      #if !USE_USB_MICRO_FRAMES
        if ( pddev->usb_20_mode )
        {
          // USB 2.0 microframe timing.
          // Just add an offset to compensate constant latency.
          // This value has been determined experimentally on different hardware platforms
          usb_latency_cycles = ( (ULONGLONG) PerfFreq.QuadPart ) / 20000UL; // represents 50 us
        }
        else
      #endif
        {
          // USB 1.1 mode with millisecond timing.
          // Compensate latency to millisecond frame boundaries.

          if ( (temp_fn2 - temp_fn1) < 0 )
            FrameNumberDiff = 2;
          else
            FrameNumberDiff = temp_fn2 - temp_fn1;

          cycles_diff = (ULONGLONG) ( UsbPostCount.QuadPart - UsbPreCount.QuadPart );

          #if USE_USB_MICRO_FRAMES
            if ( micro_frame_number_1 > 0 || micro_frame_number_2 > 0 )
              frame_length_cycles = (ULONGLONG) ( (ULONGLONG) PerfFreq.QuadPart ) / 8000UL;
            else
              frame_length_cycles = (ULONGLONG) ( (ULONGLONG) PerfFreq.QuadPart ) / 1000UL;
          #else
            frame_length_cycles = (ULONGLONG) ( (ULONGLONG) PerfFreq.QuadPart ) / 1000UL;
          #endif

          if ( ( temp_fn1 == 0 ) && ( temp_fn2 == 0 ) )
          {
            if ( cycles_diff > frame_length_cycles )
              usb_latency_cycles =  cycles_diff - frame_length_cycles;
            else
              usb_latency_cycles =  frame_length_cycles - cycles_diff;
          }
          else
            usb_latency_cycles =  cycles_diff - ( ( FrameNumberDiff - 1 ) * frame_length_cycles );

          #if defined( DEBUG )
            swprintf( pddev->wcs_msg, L"FD %d CD %I64u l %I64u fl %I64u", FrameNumberDiff,
                      cycles_diff, usb_latency_cycles, frame_length_cycles );
            _dbg_evt_msg_w( GlbDriverObject, pddev->wcs_msg );
          #endif
        }

        pddev->acc_cycles += usb_latency_cycles;
      }
    #endif
  }

  return rc;

}  // pcps_read_usb

#if !defined( AVOID_REDUNDANT_REDECLARATION )
  PCPS_READ_FNC pcps_read_usb;
#endif

#endif  // _PCPS_USE_USB

/** @} defgroup group_pcps_read_fnc */



/*HDR*/
/**
 * @brief Write data to a device
 *
 * @param[in]  pddev   Pointer to the device structure
 * @param[in]  cmd     The command code for the board, see @ref PCPS_CMD_CODES
 * @param[in]  buffer  A buffer with data to be written according to the cmd code
 * @param[in]  count   The number of bytes to be written according to the cmd code
 *
 * @return ::MBG_SUCCESS on success,
 *         ::MBG_ERR_TIMEOUT if device didn't respond in time,
 *         ::MBG_ERR_NBYTES if the number of parameter bytes did not match
 *                          the number of data bytes expected by the device,
 *         or one of the other @ref MBG_RETURN_CODES
 *
 * @ingroup pcps_io_fncs
 * @see @ref pcps_io_fncs
 */
int pcps_write( PCPS_DDEV *pddev, uint8_t cmd,
                const void FAR *buffer, uint16_t count )
{
  int rc;

#if _PCPS_USE_USB
  if ( _pcps_ddev_is_usb( pddev ) )
  {
    int actual_count = 0;  // required by macro
    int n = sizeof( cmd ) + count;
    uint8_t *p = _pcps_kmalloc( n );

    if ( p == NULL )
      return MBG_ERR_NO_MEM;

    p[0] = cmd;
    memcpy( &p[1], buffer, count );

    rc = _pcps_usb_write( pddev, p, n );

    if ( rc == MBG_SUCCESS )
    {
      rc = _pcps_usb_read( pddev, p, 1 );

      if ( rc == MBG_SUCCESS )
        rc = (int8_t) p[0];  // return the rc from the board
    }

    _pcps_kfree( p, n );
  }
  else
#endif // _PCPS_USE_USB
  {
    const uint8_t FAR *p = (const uint8_t FAR *) buffer;
    int i;
    uint8_t bytes_expected;
    int8_t write_rc;

    // Write the command and read one byte which will contain
    // the number of data bytes that must follow.
    rc = _pcps_read_var( pddev, cmd, bytes_expected );

    #if DEBUG_IO
      _mbgddmsg_4( MBG_DBG_DETAIL, "pcps_write: cmd %02X, %u bytes, expects %u, rc: %i",
                   cmd, count, bytes_expected, rc );
    #endif

    if ( rc < 0 )
      goto done;


    // Check if the number of data bytes to be written is correct.
    if ( bytes_expected != count )
    {
      rc = MBG_ERR_NBYTES;
      goto done;
    }


    // Write all bytes but the last one without reading anything.
    bytes_expected--;

    for ( i = 0; i < bytes_expected; i++ )
    {
      #if DEBUG_IO
        _mbgddmsg_2( MBG_DBG_DETAIL, "pcps_write: byte %i: 0x%02X", i, *p );
      #endif

      rc = _pcps_write_byte( pddev, *p++ );

      if ( rc < 0 )
        goto done;
    }

    // Write the last byte and read the completion code.
    #if DEBUG_IO
      _mbgddmsg_2( MBG_DBG_DETAIL, "pcps_write: last byte %i: 0x%02X", i, *p );
    #endif

    rc = _pcps_read_var( pddev, *p++, write_rc );

    // If an error code has been returned by the I/O function,
    // return that code, otherwise return the completion code
    // read from the board.
    if ( !( rc < 0 ) )
      rc = write_rc;
  }

done:
  #if DEBUG_IO
    _mbgddmsg_1( MBG_DBG_DETAIL, "pcps_write: return %i", rc );
  #endif

  return rc;

}  // pcps_write

#if !defined( AVOID_REDUNDANT_REDECLARATION )
  PCPS_WRITE_FNC pcps_write;
#endif



/*HDR*/
/**
 * @brief Generic I/O function
 *
 * @param[in]  pddev     Pointer to the device structure
 * @param[in]  type      The type of data to be read/written, see @ref PCPS_CMD_CODES
 * @param[in]  in_buff   A buffer with data to be written according to the type code
 * @param[in]  in_cnt    The number of bytes to be written according to the type code
 * @param[out] out_buff  A buffer with data to be read according to the type  code
 * @param[in]  out_cnt   The number of bytes to be read according to the type code
 *
 * @return ::MBG_SUCCESS on success,
 *         ::MBG_ERR_TIMEOUT if device didn't respond in time,
 *         ::MBG_ERR_NBYTES if the number of parameter bytes did not match
 *                          the number of data bytes expected by the device,
 *         or one of the other @ref MBG_RETURN_CODES
 *
 * @ingroup pcps_io_fncs
 * @see @ref pcps_io_fncs
 */
int pcps_generic_io( PCPS_DDEV *pddev, uint8_t type,
                     const void FAR *in_buff, uint8_t in_cnt,
                     void FAR *out_buff, uint8_t out_cnt )
{
  const uint8_t FAR *p;
  int i;
  int rc;
  uint8_t tmp_byte;
  int8_t data_read[PCPS_FIFO_SIZE];
  uint8_t bytes_to_read;

  #if DEBUG_IO
    #if defined( MBG_TGT_DOS )
      #define FP_FMT  "%Fp"
    #else
      #define FP_FMT  "%p"
    #endif
    _mbgddmsg_5( MBG_DBG_DETAIL, "pcps_generic_io: type 0x%02X, in_buff: " FP_FMT
                 " (%u), out_buf " FP_FMT " (%u)",
                 type, in_buff, in_cnt, out_buff, out_cnt );
  #endif

  // Write the command and read one byte which will contain
  // the number of data bytes that must follow.
  rc = _pcps_read_var( pddev, PCPS_GENERIC_IO, tmp_byte );

  if ( mbg_rc_is_error( rc ) )
    return rc;


  // Check if the number of data bytes to be written is correct.
  if ( tmp_byte != 3 )
    return MBG_ERR_NBYTES;


  #if DEBUG_IO
    _mbgddmsg_3( MBG_DBG_DETAIL, "pcps_generic_io: going to write type 0x%02X, in_sz %i, out_sz %i",
                 type, in_cnt, out_cnt );
  #endif

  // Write the 3 bytes which are expected:
  rc = _pcps_write_byte( pddev, type );

  if ( mbg_rc_is_error( rc ) )
    goto done;


  rc = _pcps_write_byte( pddev, in_cnt );

  if ( mbg_rc_is_error( rc ) )
    goto done;


  if ( in_cnt == 0 )
    tmp_byte = out_cnt;
  else
  {
    rc = _pcps_write_byte( pddev, out_cnt );

  if ( mbg_rc_is_error( rc ) )
      goto done;


    // Write the input parameters
    #if DEBUG_IO
      _mbgddmsg_0( MBG_DBG_DETAIL, "pcps_generic_io: going to write input bytes" );
    #endif

    p = (const uint8_t FAR *) in_buff;
    tmp_byte = in_cnt - 1;

    for ( i = 0; i < tmp_byte; i++ )
    {
      rc = _pcps_write_byte( pddev, *p++ );

      if ( mbg_rc_is_error( rc ) )
        goto done;
    }

    tmp_byte = *p;
  }


  bytes_to_read = 2 + out_cnt;

  if ( bytes_to_read > sizeof( data_read ) )
    bytes_to_read = sizeof( data_read );


  // Write the last byte and read the completion code.
  rc = _pcps_read( pddev, tmp_byte, data_read, bytes_to_read );

  if ( out_cnt )   //### TODO should do some more plausibility checks
    if ( mbg_rc_is_success( rc ) )
    {
      _fmemcpy( out_buff, &data_read[2], out_cnt );
    }

done:
  // If an error code has been returned by the I/O function,
  // return that code, otherwise return the completion code
  // read from the board.
  return mbg_rc_is_error( rc ) ? rc : data_read[0];

}  // pcps_generic_io



static /*HDR*/
/**
 * @brief Get a block of data from a GPS device
 *
 * This static function is used by ::pcps_read_gps.
 *
 * @param[in]  pddev        Pointer to the device structure
 * @param[in]  data_type    The code assigned to the dadta type, see @ref PC_GPS_CMD_CODES
 * @param[out] buffer       A buffer with data to be read according to the data_type
 * @param[in]  buffer_size  The number of bytes to be read according to the data_type
 * @param[in]  block_num    A buffer with data to be written according to the type code
 * @param[in]  block_size   The number of bytes to be written according to the type code
 *
 * @return ::MBG_SUCCESS on success,
 *         ::MBG_ERR_TIMEOUT if device didn't respond in time,
 *         ::MBG_ERR_NBYTES if the number of parameter bytes did not match
 *                          the number of data bytes expected by the device,
 *         or one of the other @ref MBG_RETURN_CODES
 */
int pcps_read_gps_block( PCPS_DDEV *pddev,
                         uint8_t data_type,
                         void FAR *buffer,
                         uint16_t buffer_size,
                         uint8_t block_num,
                         uint8_t block_size )
{
  int rc;
  uint16_t n_bytes;
  uint8_t size_n_bytes;
  uint8_t uc;


  /* Determine which interface buffer size is supported
     and use the appropriate size specification */
  if ( _pcps_ddev_has_gps_data_16( pddev ) )
    size_n_bytes = 2;
  else
  {
    if ( buffer_size > 255 )
      return MBG_ERR_NBYTES;   // Error ...

    size_n_bytes = 1;
  }

  #if DEBUG_IO
    _mbgddmsg_4( MBG_DBG_DETAIL,
       "pcps_read_gps_block: cmd 0x%02X, block %u (%u), size_n_bytes = %u",
       data_type, block_num, block_size, size_n_bytes );
  #endif

  // Write the command, expect to read one byte.
  rc = _pcps_read_var( pddev, PCPS_READ_GPS_DATA, uc );

  if ( mbg_rc_is_error( rc ) )
    return rc;

  if ( uc != 1 ) // The board doesn't expect exactly one more byte
  {
    if ( uc == 0 )
    {
      // The board can't respond now. This may occur if a
      // GPS receiver is still initializing after power-up.
      #if DEBUG_IO
        _mbgddmsg_0( MBG_DBG_DETAIL, "pcps_read_gps_block: board not yet initialized" );
      #endif

      return MBG_ERR_NOT_READY;
    }
    else
    {
      #if DEBUG_IO
        _mbgddmsg_1( MBG_DBG_DETAIL, "pcps_read_gps_block: board expects %u bytes rather than 1", uc );
      #endif

      return MBG_ERR_NBYTES;   // Error ...
    }
  }

  // Write the code corresponding to the type of data we
  // want to read, expect to read the expected size.
  n_bytes = 0;
  rc = _pcps_read( pddev, data_type, &n_bytes, size_n_bytes );

  if ( mbg_rc_is_error( rc ) )   // Error ...
    return rc;

  #if defined( MBG_ARCH_BIG_ENDIAN )
    // Swap n_bytes regardless of whether we have actually read 1 or 2 bytes.
    // If we have read only 1 byte then the other one is 0 anyway.
    n_bytes = _mbg16_to_cpu( n_bytes );
  #endif

  #if DEBUG_IO
    _mbgddmsg_2( MBG_DBG_DETAIL, "pcps_read_gps_block: device expects data size %u, buffer size %u",
                 n_bytes, buffer_size );
  #endif

  if ( n_bytes == 0 )
    return MBG_ERR_INV_TYPE;

  if ( n_bytes != buffer_size )  // Size of data structure does not match.
    return MBG_ERR_NBYTES;


  // Write the block number and read n bytes of data.
  rc = _pcps_read( pddev, block_num, buffer, block_size );

  return rc;

}  // pcps_read_gps_block



/*HDR*/
/**
 * @brief Read a large data structure from a device
 *
 * Read data structures which exceed ::PCPS_FIFO_SIZE bytes.
 * This can't be handled in a single read cycle, and due to
 * limitations of the device's microprocessor the execution time
 * can be up to 20 milliseconds, depending on the device type.
 * This has been introduced with the first GPS devices but is
 * now in fact also used with non-GPS devices.
 *
 * @param[in]  pddev        Pointer to the device structure
 * @param[in]  data_type    The code assigned to the dadta type, see @ref PC_GPS_CMD_CODES
 * @param[out] buffer       A buffer with data to be read according to the data_type
 * @param[in]  buffer_size  The number of bytes to be read according to the data_type
 *
 * @return ::MBG_SUCCESS on success,
 *         ::MBG_ERR_TIMEOUT if device didn't respond in time,
 *         ::MBG_ERR_NBYTES if the number of parameter bytes did not match
 *                          the number of data bytes expected by the device,
 *         or one of the other @ref MBG_RETURN_CODES
 *
 * @ingroup pcps_io_fncs
 * @see @ref pcps_io_fncs
 */
int pcps_read_gps( PCPS_DDEV *pddev,
                   uint8_t data_type,
                   void FAR *buffer,
                   uint16_t buffer_size )
{
  uint8_t FAR *p = (uint8_t FAR *) buffer;
  int rc = 0;
  int dt_quot;
  int dt_rem;
  int block_num;


  #if DEBUG_IO
    _mbgddmsg_3( MBG_DBG_DETAIL, "Going to read GPS data, type: 0x%02X, addr: %p, size: %u",
                 data_type, buffer, buffer_size );
  #endif

  // Split buffer size to a number of blocks of PCPS_FIFO_SIZE
  // and a number of remaining bytes (less than PCPS_FIFO_SIZE).
  dt_quot = buffer_size / PCPS_FIFO_SIZE;
  dt_rem = buffer_size % PCPS_FIFO_SIZE;

  // Read dt_quot full blocks of data.
  for ( block_num = 0; block_num < dt_quot; block_num++ )
  {
    rc = pcps_read_gps_block( pddev, data_type, p, buffer_size,
                              (uint8_t) block_num, PCPS_FIFO_SIZE );

    if ( mbg_rc_is_error( rc ) )   // Error ...
      goto done;

    // Move the destination pointer to the next free byte.
    p += PCPS_FIFO_SIZE;
  }


  // Read dt_rem additional bytes of data.
  if ( dt_rem )
    rc = pcps_read_gps_block( pddev, data_type, p, buffer_size,
                              (uint8_t) block_num, (uint8_t) dt_rem );

done:
  #if DEBUG_IO
    _mbgddmsg_1( MBG_DBG_DETAIL, "Done reading GPS data, rc: %i", rc );
  #endif

  return rc;

}  // pcps_read_gps

#if !defined( AVOID_REDUNDANT_REDECLARATION )
  PCPS_READ_FNC pcps_read_gps;
#endif



/*HDR*/
/**
 * @brief Write a large data structure to a device
 *
 * This has been introduced with the first GPS devices but is
 * now in fact also used with non-GPS devices.
 *
 * @param[in]  pddev        Pointer to the device structure
 * @param[in]  data_type    The code assigned to the dadta type, see @ref PC_GPS_CMD_CODES
 * @param[in]  buffer       A buffer with data to be written according to the data_type
 * @param[in]  buffer_size  The number of bytes to be written according to the data_type
 *
 * @return ::MBG_SUCCESS on success,
 *         ::MBG_ERR_TIMEOUT if device didn't respond in time,
 *         ::MBG_ERR_NBYTES if the number of parameter bytes did not match
 *                          the number of data bytes expected by the device,
 *         or one of the other @ref MBG_RETURN_CODES
 *
 * @ingroup pcps_io_fncs
 * @see @ref pcps_io_fncs
 */
int pcps_write_gps( PCPS_DDEV *pddev,
                    uint8_t data_type,
                    const void FAR *buffer,
                    uint16_t buffer_size )
{
  const uint8_t FAR *p = (const uint8_t FAR *) buffer;
  int rc;
  short i;
  uint16_t n_bytes;
  uint8_t size_n_bytes;
  uint8_t uc;


  /* Determine which interface buffer size is supported
     and use the appropriate size specification */
  if ( _pcps_ddev_has_gps_data_16( pddev ) )
    size_n_bytes = 2;
  else
  {
    if ( buffer_size > 255 )
      return MBG_ERR_NBYTES;   // Error ...

    size_n_bytes = 1;
  }

  #if DEBUG_IO
    _mbgddmsg_1( MBG_DBG_DETAIL, "pcps_write_gps: size_n_bytes = %u", size_n_bytes );
  #endif

  // Write the command, expect to read one byte.
  rc = _pcps_read_var( pddev, PCPS_WRITE_GPS_DATA, uc );

  if ( mbg_rc_is_error( rc ) )   // Error ...
    return rc;

  if ( uc != 1 ) // The board doesn't expect exactly one more byte
  {
    if ( uc == 0 )
    {
      // The board can't respond now. This may occur if a
      // GPS receiver is still initializing after power-up.
      #if DEBUG_IO
        _mbgddmsg_0( MBG_DBG_DETAIL, "pcps_write_gps: board not yet initialized" );
      #endif

      return MBG_ERR_NOT_READY;
    }
    else
    {
      #if DEBUG_IO
        _mbgddmsg_1( MBG_DBG_DETAIL, "pcps_write_gps: board expects %u bytes rather than 1", uc );
      #endif

      return MBG_ERR_NBYTES;   // Error ...
    }
  }

  // Write the code corresponding to the type of data we
  // want to write, expect to read the expected size.
  n_bytes = 0;
  rc = _pcps_read( pddev, data_type, &n_bytes, size_n_bytes );

  if ( mbg_rc_is_error( rc ) )   // Error ...
    return rc;

  #if defined( MBG_ARCH_BIG_ENDIAN )
    // Swap n_bytes regardless of whether we have actuall read 1 or 2 bytes.
    // If we have read only 1 byte then the other one is 0.
    n_bytes = _mbg16_to_cpu( n_bytes );
  #endif

  #if DEBUG_IO
    _mbgddmsg_2( MBG_DBG_DETAIL, "pcps_write_gps: device expects data size %u, buffer size is %u",
                 n_bytes, buffer_size );
  #endif

  if ( n_bytes != buffer_size )  // The board doesn't expect the number
    return MBG_ERR_NBYTES;       // of bytes we were going to write.


  // Write all bytes but the last one without reading.
  buffer_size--;

  for ( i = 0; i < buffer_size; i++ )
  {
    rc = _pcps_write_byte( pddev, *p++ );

    if ( mbg_rc_is_error( rc ) )   // Error ...
      return rc;
  }


  // Write the last byte and read the completion code.
  rc = _pcps_read_var( pddev, *p, n_bytes );

  // If an error code has been returned by read_fnc, return that
  // code, otherwise return the completion code read from the board.
  return rc ? rc : n_bytes;

}  // pcps_write_gps

#if !defined( AVOID_REDUNDANT_REDECLARATION )
  PCPS_WRITE_FNC pcps_write_gps;
#endif



static /*HDR*/
/**
 * @brief Read the firmware ID from a device
 *
 * This is usually done first when probing a device to check if the
 * device is properly installed and can be accessed without problems.
 *
 * @param[in]  pddev  Pointer to a device structure
 * @param[out] fw_id  A buffer to be filled withtthe firmware ID
 *
 * @return ::MBG_SUCCESS if the the signature was found, or signature was specified, or
 *         ::MBG_ERR_FW_ID if the firmware ID is unknown or invalid
 */
int pcps_get_fw_id( PCPS_DDEV *pddev, PCPS_ID_STR FAR fw_id )
{
  int rc;

  memset( fw_id, 0, sizeof( PCPS_ID_STR ) );  // sizeof( fw_id ) would yield pointer size only

  // read the first part of the firmware ID
  rc = _pcps_read( pddev, PCPS_GIVE_FW_ID_1, &fw_id[0], PCPS_FIFO_SIZE );

  if ( mbg_rc_is_error( rc ) )
    goto out;

  // read the second part of the firmware ID
  rc = _pcps_read( pddev, PCPS_GIVE_FW_ID_2, &fw_id[PCPS_FIFO_SIZE], PCPS_FIFO_SIZE );

  // force termination with 0
  fw_id[PCPS_ID_SIZE - 1] = 0;

out:
  return rc;

}  // pcps_get_fw_id



static /*HDR*/
/**
 * @brief Check an ASCIIZ string for a valid signature.
 *
 * This function is only used to determine the type of a legacy ISA card.
 * For PCI and USB devices the type can be determined from the device ID.
 *
 * @param[in] pddev  Pointer to a device structure
 * @param[in] ref    The reference signature, i.e. the expected signature
 *                   at the beginning of the device's firmware ID
 *
 * @return ::MBG_SUCCESS if a signature was specified and the signature was found, or
 *         ::MBG_ERR_FW_ID if the firmware ID is unknown or invalid
 */
int pcps_check_id( PCPS_DDEV *pddev, const char FAR *ref )
{
  // check if the first characters of the string match the reference

  if ( ref )
    if ( _fstrncmp( _pcps_ddev_fw_id( pddev ), ref, _fstrlen( ref ) ) )
      return MBG_ERR_FW_ID;

  return MBG_SUCCESS;

}  // pcps_check_id



static /*HDR*/
/**
 * @brief Retrieve a version number from a firmware ID string
 *
 * @param[in] idstr  The ID string, e.g. a ::PCPS_ID_STR
 *
 * @return A version number in hex format, e.g. 0x0270 for version 2.7,
 *         or 0 if no version number could be found.
 */
PCPS_FW_REV_NUM pcps_get_rev_num( char FAR *idstr )
{
  int len = _fstrlen( idstr );
  int i;

  if ( len < 2 )
    goto no_rev_num;

  len -= 2;

  for ( i = 0; i < len; i++ )
  {
    if ( idstr[i + 1] == '.' )
    {
      uint8_t rev_num_hi = idstr[i] & 0x0F;
      uint8_t rev_num_lo = ( idstr[i + 2] & 0x0F ) << 4;

      // If the minor part of the version string has 2 digits
      // then evaluate the 2nd digit, too.

      char c = idstr[i + 3];

      if ( c >= '0' && c <= '9' )
        rev_num_lo |= c & 0x0F;

      return ( rev_num_hi << 8 ) | rev_num_lo;
    }
  }

no_rev_num:
  return 0;

}  // pcps_get_rev_num



/*HDR*/
/**
 * @brief Read the serial number from a device, if supported
 *
 * If the serial number could be read successfully then it is
 * stored in one of the sub-structures of pddev.
 *
 * @param[in,out] pddev  Pointer to a device structure
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
int pcps_read_sernum( PCPS_DDEV *pddev )
{
  char *cp;
  int i;
  int rc = MBG_SUCCESS;


  memset( pddev->dev.cfg.sernum, 0, sizeof( pddev->dev.cfg.sernum ) );

  // There are different ways to read the clock's S/N. Check which
  // way is supported, and read the S/N from the device.
  //
  // Never just return a previous copy of the serial number which
  // has been read earlier since the S/N may just have been set
  // by a configuration API call.

  // Read directly. This is supported by newer devices.
  if ( _pcps_ddev_has_sernum( pddev ) )
  {
    #if DEBUG_SERNUM
      _mbgddmsg_0( MBG_DBG_DETAIL, "getting S/N via PCPS_GIVE_SERNUM cmd" );
    #endif

    rc = _pcps_read( pddev, PCPS_GIVE_SERNUM, pddev->dev.cfg.sernum, PCPS_FIFO_SIZE );

    if ( mbg_rc_is_error( rc ) )
    {
      _mbgddmsg_2( MBG_DBG_INIT_DEV, "PCPS read SERNUM %X: rc = %i",
                   _pcps_ddev_dev_id( pddev ), rc );
      goto fail;
    }

    goto check;
  }


  // The S/N is part of the RECEIVER_INFO structure,
  // so use that one, if supported.
  if ( _pcps_ddev_has_receiver_info( pddev ) )
  {
    RECEIVER_INFO *p_ri = _ri_addr( pddev );

    #if DEBUG_SERNUM
      _mbgddmsg_0( MBG_DBG_DETAIL, "getting S/N from receiver info" );
    #endif

    rc = _pcps_read_gps_var( pddev, PC_GPS_RECEIVER_INFO, *p_ri );

    if ( mbg_rc_is_error( rc ) )
    {
      _mbgddmsg_2( MBG_DBG_INIT_DEV, "PCPS read GPS receiver info %X: rc = %i",
                   _pcps_ddev_dev_id( pddev ), rc );
      goto fail;
    }

    _mbg_swab_receiver_info( p_ri );

    strncpy( pddev->dev.cfg.sernum, p_ri->sernum,
             sizeof( pddev->dev.cfg.sernum ) );
    goto check;
  }


  // Older GPS clocks store the S/N in an IDENT structure
  // which needs to be decoded to get the S/N.
  if ( _pcps_ddev_has_ident( pddev ) )
  {
    static_wc IDENT ident = { { 0 } };

    #if DEBUG_SERNUM
      _mbgddmsg_0( MBG_DBG_DETAIL, "getting S/N from ident" );
    #endif

    rc = _pcps_read_gps_var( pddev, PC_GPS_IDENT, ident );

    #if !defined( MBG_TGT_LINUX ) && !defined( MBG_TGT_BSD )
      assert( sizeof( ident ) < sizeof( pddev->dev.cfg.sernum ) );
    #endif

    if ( mbg_rc_is_error( rc ) )
    {
      _mbgddmsg_2( MBG_DBG_INIT_DEV, "PCPS read GPS ident %X: rc = %i",
                   _pcps_ddev_dev_id( pddev ), rc );
      goto fail;
    }

    // The ident union must never be swapped due to endianess since we are
    // using it only as an array of characters.

    #if DEBUG_SERNUM
      for ( i = 0; i < sizeof( ident ); i += 4 )
        _mbgddmsg_5( MBG_DBG_DETAIL, "ident[%02i]: %02X %02X %02X %02X",
                     i, ident.c[i], ident.c[i+1], ident.c[i+2], ident.c[i+3] );
    #endif

    mbg_gps_ident_decode( _pcps_ddev_sernum( pddev ), &ident );

    #if DEBUG_SERNUM
      _mbgddmsg_1( MBG_DBG_DETAIL, "S/N decoded from ident: %s (raw)",
                   _pcps_ddev_sernum( pddev ) );
    #endif

    // Some old devices may have non-digits appended
    // to the S/N string. Truncate such a string.
    for ( cp = _pcps_ddev_sernum( pddev ); *cp; cp++ )
      if ( ( *cp < '0' ) || ( *cp > '9' ) )
      {
        *cp = 0;
        break;
      }

    #if DEBUG_SERNUM
      _mbgddmsg_1( MBG_DBG_DETAIL, "S/N decoded from ident: %s",
                   _pcps_ddev_sernum( pddev ) );
    #endif

    goto check;
  }

  // The clock doesn't support a S/N.
  // Assume the rc is still set to MBG_SUCCESS.
  strcpy( _pcps_ddev_sernum( pddev ), "N/A" );
  goto done;


check:
  // Remove unprintable characters.
  for ( i = 0, cp = pddev->dev.cfg.sernum;
        i < sizeof( pddev->dev.cfg.sernum ); i++, cp++ )
    if ( ( *cp < 0x20 ) || ( *cp >= 0x7F ) )
    {
      #if DEBUG_SERNUM
        *cp = '#';
      #else
        *cp = 0;
      #endif
    }

#if DEBUG_SERNUM

fail:
  goto done;

#else
  // Remove trailing spaces or 'F' characters which may
  // unfortunately be returned by some devices.
  for ( i = strlen( pddev->dev.cfg.sernum ); ; )
  {
    if ( i == 0 )
      break;

    --i;
    cp = &pddev->dev.cfg.sernum[i];
    if ( ( *cp > ' ' ) && ( *cp != 'F' ) )
      break;  // done

    *cp = 0;
  }

  if ( strlen( pddev->dev.cfg.sernum ) )
    goto done;


fail:
  // No valid serial number has been found, though the device
  // should have one. In order to distinguish from devices which
  // don't even support a serial number we return a number of '?'
  // rather than "N/A".
  memset( pddev->dev.cfg.sernum, '?', 8 );
  pddev->dev.cfg.sernum[8] = 0;
#endif

done:
  // Make sure the S/N is terminated by 0.
  pddev->dev.cfg.sernum[sizeof( pddev->dev.cfg.sernum ) - 1] = 0;

  return rc;

}  // pcps_read_sernum



#if _PCPS_USE_RSRCMGR

static /*HDR*/
/**
 * @brief Try to claim an I/O port resource range
 *
 * @param[in,out] pddev   The device structure
 *
 * @return ::MBG_SUCCESS if the port range could be claimed, else ::MBG_ERR_CLAIM_RSRC
 */
int pcps_rsrc_claim( PCPS_DDEV *pddev )
{
  ushort decode_width;
  int i;

  if ( _pcps_ddev_is_pci( pddev ) )
    decode_width = PCPS_DECODE_WIDTH_PCI;
  else
    if ( _pcps_ddev_is_mca( pddev ) )
      decode_width = PCPS_DECODE_WIDTH_MCA;
    else
      decode_width = PCPS_DECODE_WIDTH_ISA;

  for ( i = 0; i < pddev->rsrc_info.num_rsrc_io; i++ )
  {
    PCPS_IO_RSRC *p = &_pcps_ddev_io_rsrc( pddev, i );
    ushort rc;

    rc = _rsrc_alloc_ports( pddev, p->base_raw, p->num, decode_width );

    // If the resource manager was unable to alloc the resources
    // then the selected range of ports is probably in use
    // by another hardware device and/or driver
    if ( mbg_rc_is_error( rc ) )
    {
      _pcps_ddev_set_err_flags( pddev, PCPS_EF_IO_RSRC );
      return MBG_ERR_CLAIM_RSRC;
    }
  }

  return MBG_SUCCESS;

}  // pcps_rsrc_claim



/*HDR*/
/**
 * @brief Release an I/O port resource range that has been claimed before
 *
 * @param[in,out] pddev   The device structure
 */
void pcps_rsrc_release( PCPS_DDEV *pddev )
{
  int i;

  for ( i = 0; i < pddev->rsrc_info.num_rsrc_io; i++ )
  {
    PCPS_IO_RSRC *p = &_pcps_ddev_io_rsrc( pddev, i );

    if ( p->base_raw )
      _rsrc_dealloc_ports( pddev, p->base_raw, p->num );
  }

}  // pcps_rsrc_release



#if defined( MBG_TGT_OS2 )

static /*HDR*/
/**
 * @brief Register a device under OS/2
 *
 * @param[in] pddev   The device structure
 */
void pcps_rsrc_register_device( PCPS_DDEV *pddev )
{
  #define RSRC_BASE_NAME "RADIOCLK_#  Meinberg Radio Clock "
  static const char rsrc_type_dcf77[] = RSRC_BASE_NAME "(DCF77)";
  static const char rsrc_type_gps[] = RSRC_BASE_NAME "(GPS)";
  static const char rsrc_type_irig[] = RSRC_BASE_NAME "(IRIG)";

  uchar bus_type;
  ushort rc;
  const char *cp;

  #if _PCPS_USE_USB
    #error USB not supported for this target environment!
  #endif

  if ( _pcps_ddev_is_pci( pddev ) )
    bus_type = RSRC_BUS_PCI;
  else
    if ( _pcps_ddev_is_mca( pddev ) )
      bus_type = RSRC_BUS_MCA;
    else
      bus_type = RSRC_BUS_ISA;

  if ( _pcps_ddev_is_irig_rx( pddev ) )
    cp = rsrc_type_irig;
  else
    if ( _pcps_ddev_is_gps( pddev ) )
      cp = rsrc_type_gps;
    else
      cp = rsrc_type_dcf77;

  rc = rsrc_register_device( &pddev->hDev, &pddev->rsrc, n_ddevs - 1, cp, bus_type );

}  // pcps_rsrc_register_device

#endif  // defined( MBG_TGT_OS2 )

#endif  // _PCPS_USE_RSRCMGR



#if _PCPS_USE_MCA

/*--------------------------------------------------------------
 *  PS31 only:
 *
 *  The scheme below shows the way the bits of the POS 103
 *  configuration byte are mapped to the PS31's programmable
 *  address decoder:
 *
 *   MSB            LSB
 *    |      ||      |
 *    0bbbbbb1000bxxxx  <-- 16 bit port base address (binary)
 *     ||||||    |
 *      \\\\\\   |               b:  configurable bit
 *        \\\\\\ |               x:  don't care bit
 *         |||||||
 *        1bbbbbbb      <--  8 bit configuration byte (POS 103)
 *        |
 *        |
 *    decoder enable bit, always 1 if adapter enabled
 *
 *-------------------------------------------------------------*/

static /*HDR*/
/**
 * @brief Convert the code read from PS/2 POS to the port base address
 *
 * @param[in] pos  The POS code to be converted
 *
 * @return  The decoded I/O port base address
 */
uint16_t pcps_port_from_pos( uint16_t pos )
{
  uint16_t us = ( ( pos & 0x007E ) << 8 ) | 0x0100;

  if ( pos & 0x0001 )
    us |= 0x0010;

  return us;

}  // pcps_port_from_pos



static /*HDR*/
/**
 * @brief Convert the port base address to a PS/2 POS code
 *
 * @param[in] port  The I/O port base address to be encoded
 *
 * @return  The encoded POS code
 */
uint8_t pcps_pos_from_port( uint16_t port )
{
  uint8_t uc;


  uc = *( ( (uchar *) (&port) ) + 1 ) & 0x7E;

  if ( port & 0x0010 )
    uc |= 1;

  return uc;

}  // pcps_pos_from_port



static /*HDR*/
/**
 * @brief Detect MCA devices and allocate device info structures
 *
 * @param[in]  alloc_fnc  Pointer to function called to allocate a device structure for each detected device.
 */
void pcps_detect_mca_devices( PCPS_DDEV_ALLOC_FNC *alloc_fnc )
{
  short rc;
  ushort type_idx;

  rc = mca_fnc_init();

  if ( rc != MCA_SUCCESS )
    return;


  // MCA is installed, now try to find a MCA clock with
  // known ID.
  for ( type_idx = 0; type_idx < N_PCPS_DEV_TYPE; type_idx++ )
  {
    static_wc MCA_POS_DATA pos_data;
    PCPS_DDEV *pddev;
    PCPS_DEV_TYPE *p = &pcps_dev_type[type_idx];

    static_wc uchar slot_num;   // the slot in which the board is installed


    if ( !( p->bus_flags & PCPS_BUS_MCA ) )
      continue;


    rc = mca_find_adapter( p->dev_id, &slot_num, &pos_data );

    if ( rc != MCA_SUCCESS )
      continue;


    // New device found, try to add to list.
    pddev = alloc_fnc();

    if ( pddev )  // Setup only if successful.
    {
      pddev->dev.type = *p;
      pcps_add_rsrc_io( pddev, pcps_port_from_pos( pos_data.pos_103 ),
                        PCPS_NUM_PORTS_MCA );

      //##++ Should try to read the interrupt line assigned to the clock.
      // The standard functions, however, don't use any interrupt.

      pcps_probe_device( pddev, 0, slot_num );
    }
  }

  mca_fnc_deinit();

}  // pcps_detect_mca_devices

#endif  /* _PCPS_USE_MCA */



/*HDR*/
/**
 * @brief Lookup a specific device in the device table
 *
 * The function below takes a bus flag and device ID to search
 * the table of known devices for a device which matches the
 * given criteria.
 *
 * @param[in]  bus_mask  Mask of the bus type to look up, see @ref PCPS_BUS_FLAG_MASKS
 * @param[in]  dev_id    The device ID to lookup, see @ref MEINBERG_PCI_DEVICE_IDS
 *                       or @ref MBG_USB_DEVICE_IDS, depending on the bus_mask
 *
 * @return  A pointer to the device table entry, or NULL if no entry found
 */
PCPS_DEV_TYPE *pcps_get_dev_type_table_entry( PCPS_BUS_FLAGS bus_mask, PCPS_DEV_ID dev_id )
{
  int i;

  for ( i = 0; i < N_PCPS_DEV_TYPE; i++ )
  {
    PCPS_DEV_TYPE *p = &pcps_dev_type[i];

    if ( !( p->bus_flags & bus_mask ) )
      continue;

    if ( p->dev_id == dev_id )
      return p;
  }

  return NULL;

}  // pcps_get_dev_type_table_entry



/*HDR*/
/**
 * @brief Allocate a device info structure for a device
 *
 * @return  A pointer to the allocated structure, or NULL if failed
 */
PCPS_DDEV *pcps_alloc_ddev( void )
{
  PCPS_DDEV *pddev;

  #if !_PCPS_STATIC_DEV_LIST
    pddev = _pcps_kmalloc( sizeof( *pddev ) );
  #else
    if ( n_ddevs >= N_SUPP_DEV_BUS )
    {
      _mbgddmsg_0( MBG_DBG_INIT_DEV,
                   "Unable to add new device: max count reached" );

      return NULL;
    }

    pddev = &pcps_ddev[n_ddevs];
    n_ddevs++;
  #endif

  if ( pddev )
  {
    memset( pddev, 0, sizeof( *pddev ) );

    // If mutexes or spinlocks need to be destroyed on the target OS
    // when the driver shuts down then they are initialized now and
    // destroyed in the complementary function pcps_free_ddev().
    // However, there are target OSs where those semaphores don't need
    // to be destroyed, and sometimes even *must not* be initialized
    // at this early point of driver initialization, e.g. under Windows,
    // in which case the semaphores will be initialized later.
    #if defined( _mbg_mutex_destroy )
      _mbg_mutex_init( &pddev->dev_mutex );
    #endif
    #if defined( _mbg_spin_lock_destroy )
      _mbg_spin_lock_init( &pddev->mm_lock );
      _mbg_spin_lock_init( &pddev->irq_lock );
    #endif
  }

  return pddev;

}  // pcps_alloc_ddev



/*HDR*/
/**
 * @brief Free a previously allocated device info structure
 *
 * @param[in] pddev  Pointer to the device structure to be freed
 */
void pcps_free_ddev( PCPS_DDEV *pddev )
{
  if ( pddev )
  {
    #if defined( _mbg_mutex_destroy )
      _mbg_mutex_destroy( &pddev->dev_mutex );
    #endif
    #if defined( _mbg_spin_lock_destroy )
      _mbg_spin_lock_destroy( &pddev->mm_lock );
      _mbg_spin_lock_destroy( &pddev->irq_lock );
    #endif

    #if !_PCPS_STATIC_DEV_LIST
      _pcps_kfree( pddev, sizeof( *pddev ) );
    #else
      memset( pddev, 0, sizeof( *pddev ) );

      if ( n_ddevs )
        n_ddevs--;
    #endif
  }

}  // pcps_free_ddev



static /*HDR*/
/**
 * @brief Convert a raw I/O base address to a short format
 *
 * See notes for ::PCPS_SHORT_PORT_RSRC.

 * @param[out]  p_short_port_rsrc  Pointer to a variable that takes up the converted address
 * @param[in]   p_io_rsrc          Pointer to a variable with the address to be converted
 */
void rsrc_port_to_cfg_port( PCPS_SHORT_PORT_RSRC *p_short_port_rsrc, const PCPS_IO_RSRC *p_io_rsrc )
{
  p_short_port_rsrc->base = (PCPS_SHORT_PORT_ADDR) p_io_rsrc->base_raw;
  p_short_port_rsrc->num = (uint16_t) p_io_rsrc->num;

}  // rsrc_port_to_cfg_port



/*HDR*/
/**
 * @brief Add an I/O address range resource to the device structure
 *
 * @param[in,out]  pddev  Pointer to the device structure
 * @param[in]      base   Base address of the I/O address range
 * @param[in]      num    Number of addresses of the I/O address range
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
int pcps_add_rsrc_io( PCPS_DDEV *pddev, ulong base, ulong num )
{
  PCPS_RSRC_INFO *prsrci = &pddev->rsrc_info;

  if ( prsrci->num_rsrc_io < N_PCPS_PORT_RSRC )
  {
    PCPS_IO_RSRC *p = &prsrci->port[prsrci->num_rsrc_io];

    p->base_mapped = (PCPS_IO_ADDR_MAPPED) _pcps_ioremap( base, num );
    p->base_raw = (PCPS_IO_ADDR_RAW) base;
    p->num = num;

    prsrci->num_rsrc_io++;

    _mbgddmsg_3( MBG_DBG_INIT_DEV, "Adding I/O rsrc #%i: %03lX(%lu)",
                 prsrci->num_rsrc_io, (ulong) base, (ulong) num );

    return MBG_SUCCESS;
  }

  return MBG_ERR_GENERIC;

}  // pcps_add_rsrc_io



/*HDR*/
/**
 * @brief Add a memory address range resource to the device structure
 *
 * @param[in,out]  pddev  Pointer to the device structure
 * @param[in]      start  Start address of the memory range
 * @param[in]      len    Size of the memory range
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
int pcps_add_rsrc_mem( PCPS_DDEV *pddev, MBG_MEM_ADDR start, ulong len )
{
  PCPS_RSRC_INFO *prsrci = &pddev->rsrc_info;

  if ( prsrci->num_rsrc_mem < N_PCPS_MEM_RSRC )
  {
    PCPS_MEM_RSRC *p = &prsrci->mem[prsrci->num_rsrc_mem];
    p->start = start;
    p->len = len;
    prsrci->num_rsrc_mem++;

    #if defined( MBG_TGT_POSIX )
      _mbgddmsg_3( MBG_DBG_INIT_DEV, "Adding mem rsrc #%i: %08llX(%lu)",
                   prsrci->num_rsrc_mem, (unsigned long long) start, len );
    #else
      _mbgddmsg_3( MBG_DBG_INIT_DEV, "Adding mem rsrc #%i: %08lX(%lu)",
                   prsrci->num_rsrc_mem, (unsigned long) start, len );
    #endif

    return MBG_SUCCESS;
  }

  return MBG_ERR_GENERIC;

}  // pcps_add_rsrc_mem



/*HDR*/
/**
 * @brief Add an IRQ number resource to the device structure
 *
 * @param[in,out]  pddev    Pointer to the device structure
 * @param[in]      irq_num  Start address of the memory range
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
int pcps_add_rsrc_irq( PCPS_DDEV *pddev, int16_t irq_num )
{
  PCPS_RSRC_INFO *prsrci = &pddev->rsrc_info;

  if ( prsrci->num_rsrc_irq == 0 )
  {
    prsrci->irq.num = irq_num;
    prsrci->num_rsrc_irq++;

    _mbgddmsg_2( MBG_DBG_INIT_DEV, "Adding IRQ rsrc #%i: %i",
                 prsrci->num_rsrc_irq, irq_num );

    return MBG_SUCCESS;
  }

  return MBG_ERR_GENERIC;

}  // pcps_add_rsrc_irq



#if _PCPS_USE_PNP

/*HDR*/
/**
 * @brief Initialize an allocated device structure for a specific device
 *
 * @param[in,out]  pddev     Pointer to the device structure
 * @param[in]      bus_mask  Mask of the bus type to look up, see @ref PCPS_BUS_FLAG_MASKS
 * @param[in]      dev_id    The device ID to lookup, see @ref MEINBERG_PCI_DEVICE_IDS
 *                           or @ref MBG_USB_DEVICE_IDS, depending on the bus_mask
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
int pcps_init_ddev( PCPS_DDEV *pddev, PCPS_BUS_FLAGS bus_mask, PCPS_DEV_ID dev_id )
{
  // First check if we really support the device to be added.
  PCPS_DEV_TYPE *pdt = pcps_get_dev_type_table_entry( bus_mask, dev_id );

  if ( pdt == NULL )
  {
    _mbgddmsg_1( MBG_DBG_INIT_DEV, "PCPS add PNP device %X: unknown type",
                 dev_id );

    return MBG_ERR_DEV_NOT_SUPP;
  }


  pddev->dev.type = *pdt;

  _mbgddmsg_2( MBG_DBG_INIT_DEV, "PCPS add PNP device %X: found %s",
               dev_id, pdt->name );

  return MBG_SUCCESS;

}  // pcps_init_ddev

#endif // _PCPS_USE_PNP



#if defined( DEBUG )

static /*HDR*/
/**
 * @brief Get the name assigned to one of the @ref PCPS_FEATURE_MASKS flags
 *
 * @param[in]  flag_mask  one of the @ref PCPS_FEATURE_MASKS flags
 *
 * @return The associated feature name, or "unknown"
 *
 * @see @ref PCPS_FEATURE_MASKS
 */
const char *get_pcps_feature_name( PCPS_FEATURES flag_mask )
{
  int i;

  for ( i = 0; i < N_PCPS_FEATURE_BITS; i++ )
    if ( ( 1UL << i ) == flag_mask )
      return pcps_feature_names[i];

  return "unknown";

}  // get_pcps_feature_name

#endif



static /*HDR*/
/**
 * @brief Check the firmware to see if a specific feature is supported
 *
 * Some devices provide a specific feature only starting with a specific
 * firmware version, but the device provides no way to test this, so
 * the feature flag in the device structure is set if the firmware
 * has the required version.
 *
 * @param[in,out]  pddev        Pointer to the device structure
 * @param[in]      req_rev_num  Revision number required for the feature to be supported
 * @param[in]      flag_mask    Flag mask of the feature to be tested
 */
void check_feature( PCPS_DDEV *pddev, ushort req_rev_num,
                    PCPS_FEATURES flag_mask )
{
  int supported = _pcps_ddev_fw_rev_num( pddev ) >= req_rev_num;

  if ( supported )
    pddev->dev.cfg.features |= flag_mask;

  #if defined( DEBUG )
    _mbgddmsg_5( MBG_DBG_INIT_DEV, "%s v%03X: feature 0x%08lX (%s) %ssupported",
                 _pcps_ddev_type_name( pddev ),
                 _pcps_ddev_fw_rev_num( pddev ),
                 (ulong) flag_mask, get_pcps_feature_name( flag_mask ),
                 supported ? "" : "not " );
  #endif

}  // check_feature



static /*HDR*/
/**
 * @brief Check the receiver info features to see if a specific feature is supported
 *
 * Some devices provide a ::RECEIVER_INFO which provides some flags that
 * can be mapped to the feature flags we use internally, so we scan these features
 * and update the internal features mask accordingly.
 *
 * @param[in,out]  pddev  Pointer to the device structure
 */
void check_ri_features( PCPS_DDEV *pddev )
{
  const RECEIVER_INFO *p_ri = _ri_addr( pddev );
  int gps_feat_bit;

  for ( gps_feat_bit = 0; gps_feat_bit < N_GPS_FEATURE; gps_feat_bit++ )
  {
    int supported = ( p_ri->features & ( 1UL << gps_feat_bit ) ) != 0;
    PCPS_FEATURES pcps_flags = ri_feat_tbl[gps_feat_bit];

    if ( supported )
    {
      if ( pcps_flags )
      {
        pddev->dev.cfg.features |= pcps_flags;

        #if defined( DEBUG )
        {
          int i;

          for ( i = 0; i < 8 * sizeof( pcps_flags ); i++ )
            if ( pcps_flags & ( 1UL << i ) )
              _mbgddmsg_4( MBG_DBG_INIT_DEV, "%s v%03X: Supported RI feature \"%s\" maps to PCPS feature %s",
                           _pcps_ddev_type_name( pddev ),
                           _pcps_ddev_fw_rev_num( pddev ),
                           gps_ri_feature_names[gps_feat_bit],
                           pcps_feature_names[i] );
        }
        #endif
      }
      else
      {
        #if defined( DEBUG )
          _mbgddmsg_3( MBG_DBG_INIT_DEV, "%s v%03X: RI feature \"%s\" supported, but no corresponding PCPS feature",
                       _pcps_ddev_type_name( pddev ),
                       _pcps_ddev_fw_rev_num( pddev ),
                       gps_ri_feature_names[gps_feat_bit] );
        #endif
      }
    }
    else
    {
      #if defined( DEBUG )
        _mbgddmsg_3( MBG_DBG_INIT_DEV, "%s v%03X: RI feature \"%s\" not supported by device",
                     _pcps_ddev_type_name( pddev ),
                     _pcps_ddev_fw_rev_num( pddev ),
                     gps_ri_feature_names[gps_feat_bit] );
      #endif
    }
  }

}  // check_ri_features



/*HDR*/
/**
 * @brief Check if a specific feature of a specific feature type is supported
 *
 * There are different structures where information can be stored
 * if a specific feature is supported. All information is set up
 * when the ::pcps_probe_device function is called to probe and
 * initialize the device.
 * This generic low-level function can be called by API functions
 * to check if a specific feature is supported.
 *
 * @param[in]  p_ddev         Pointer to the device structure
 * @param[in]  feat_req_type  See ::DEV_FEAT_REQ_TYPES
 * @param[in]  feat_num       Number and range depending on feat_req_type value
 *
 * @return ::MBG_SUCCESS if the requested feature is supported, ::MBG_ERR_NOT_SUPP_BY_DEV
 *         if not supported, or one of the other @ref MBG_ERROR_CODES
 *
 * @see ::pcps_probe_device
 */
int pcps_chk_dev_feat( PCPS_DDEV *p_ddev, uint32_t feat_req_type, uint32_t feat_num )
{
  // NOTE In the code below we currently check if the bit to be tested
  // is really defined in an enumeration, and return an error if a bit
  // is tested which is not defined.
  // Alternatively we could instead just check if the bit number is valid
  // according to the buffer size.

  switch ( feat_req_type )
  {
    case DEV_FEAT_REQ_TYPE_BUILTIN:
      _mbgddmsg_3( MBG_DBG_DETAIL, "%s: chk_dev_feat: builtin_feat %i/%i",
                   pcps_driver_name, feat_num, N_GPS_BUILTIN_FEATURE_BITS );

      if ( feat_num < N_GPS_BUILTIN_FEATURE_BITS )
        return _check_feat_supp_bit( p_ddev->builtin_features, feat_num );

      break;  // invalid


    case DEV_FEAT_REQ_TYPE_REF_TYPE:
      _mbgddmsg_3( MBG_DBG_DETAIL, "%s: chk_dev_feat: ref_type feat %i/%i",
                   pcps_driver_name, feat_num, N_PCPS_REF );

      if ( feat_num < N_PCPS_REF )
        return ( feat_num == _pcps_ddev_ref_type( p_ddev ) ) ? MBG_SUCCESS : MBG_ERR_NOT_SUPP_BY_DEV;

      break;  // invalid


    case DEV_FEAT_REQ_TYPE_PCPS:
      _mbgddmsg_3( MBG_DBG_DETAIL, "%s: chk_dev_feat: pcps_feat %i/%i",
                   pcps_driver_name, feat_num, N_PCPS_FEATURE_BITS );

      if ( feat_num < N_PCPS_FEATURE_BITS )
        return _check_feat_supp_bit( p_ddev->dev.cfg.features, feat_num );

      break;  // invalid


    case DEV_FEAT_REQ_TYPE_RI:
      _mbgddmsg_3( MBG_DBG_DETAIL, "%s: chk_dev_feat: ri_feat %i/%i",
                   pcps_driver_name, feat_num, N_GPS_FEATURE );

      if ( feat_num < N_GPS_FEATURE )
        return _check_feat_supp_bit( p_ddev->xdev_features.receiver_info.features, feat_num );

      break;  // invalid



    case DEV_FEAT_REQ_TYPE_XFEAT:
      _mbgddmsg_3( MBG_DBG_DETAIL, "%s: chk_dev_feat: x_feat %i/%i",
                   pcps_driver_name, feat_num, N_MBG_XFEATURE );

      if ( feat_num < N_MBG_XFEATURE )
        return check_feat_supp_byte_array( feat_num, p_ddev->xdev_features.xfeature_buffer.b,
                     sizeof( p_ddev->xdev_features.xfeature_buffer ) );
      break;  // invalid



    case DEV_FEAT_REQ_TYPE_TLV_FEAT:
      _mbgddmsg_3( MBG_DBG_DETAIL, "%s: chk_dev_feat: tlv_feat %i/%i",
                   pcps_driver_name, feat_num, N_MBG_TLV_FEAT_TYPES );

      if ( feat_num < N_MBG_TLV_FEAT_TYPES )
        return check_feat_supp_byte_array( feat_num, p_ddev->xdev_features.tlv_info.supp_tlv_feat.b,
                     sizeof( p_ddev->xdev_features.tlv_info.supp_tlv_feat.b ) );
      break;  // invalid
  }

  _mbgddmsg_3( MBG_DBG_WARN, "%s: chk_dev_feat: *invalid* request %i:%i",
                   pcps_driver_name, feat_req_type, feat_num );

  return MBG_ERR_INV_PARM;

}  // pcps_chk_dev_feat



/*HDR*/
/**
 * @brief Probe if a device is supported, and allocate and setup the device structure
 *
 * This function should be called by the probe routines of any
 * target-specific kernel drivers.
 * If the device is supported then all specific information including
 * supported features is read from the device and stored in sub-structures
 * of the device structure addressed by pdev.
 *
 * @param[in,out]  pddev        Pointer to the device structure which has been initialized and will be set up
 * @param[in]      bus_num      The bus number if supported (e.g. PCI), else 0
 * @param[in]      dev_fnc_num  The device/function number if supported (e.g. PCI), else 0
 *
 * @return ::MBG_SUCCESS if the requested feature is supported, ::MBG_ERR_NOT_SUPP_BY_DEV
 *         if not supported, or one of the other @ref MBG_ERROR_CODES
 *
 * @see ::pcps_cleanup_device
 * @see ::pcps_chk_dev_feat
 */
int pcps_probe_device( PCPS_DDEV *pddev,
                       PCPS_BUS_NUM bus_num,
                       PCPS_SLOT_NUM dev_fnc_num )
{
  RECEIVER_INFO *p_ri = _ri_addr( pddev );
  ushort port_rsrc_len[N_PCPS_PORT_RSRC] = { 0 };
  int port_ranges_required = 0;
  ushort status_port_offs = 0;
  int i;
  int rc;

  _mbgddmsg_1( MBG_DBG_INIT_DEV, "PCPS start device %X",
               _pcps_ddev_dev_id( pddev ) );

  pddev->read = pcps_read_null;
  pddev->dev.cfg.bus_num = bus_num;
  pddev->dev.cfg.slot_num = dev_fnc_num;

  if ( _pcps_ddev_chk_err_flags( pddev, PCPS_EF_IO_INIT | PCPS_EF_IO_ENB ) )
  {
    _mbgddmsg_1( MBG_DBG_INIT_DEV, "PCPS start device %X: failing due to error flags.",
                 _pcps_ddev_dev_id( pddev ) );
    goto fail;
  }


  // If mutexes or spinlocks need to be destroyed on the target OS
  // when the driver shuts down then they have already been initialized
  // in the device allocation routine and will be destroyed in the
  // complementary device deallocation routine.
  // However, there are target OSs where those semaphores don't need
  // to be destroyed, and sometimes even *must not* be initialized
  // when the device structure is allocated (e.g under Windows),
  // in which case the semaphores need to be initialized now.
  #if !defined( _mbg_mutex_destroy )
    #if defined( _mbg_mutex_init )
      _mbg_mutex_init( &pddev->dev_mutex );
    #endif
  #endif
  #if !defined( _mbg_spin_lock_destroy )
    #if defined( _mbg_spin_lock_init )
      _mbg_spin_lock_init( &pddev->mm_lock );
      _mbg_spin_lock_init( &pddev->irq_lock );
    #endif
  #endif


  // Next we do some setup depending on the interface type
  // and chip.

  switch ( _pcps_ddev_bus_flags( pddev ) )
  {
    #if _PCPS_USE_USB
      case PCPS_BUS_USB:
      case PCPS_BUS_USB_V2:
        // No direct port I/O possible.
        pddev->read = pcps_read_usb;

        // In case of an USB device, do some additional
        // USB initializiation
        #if defined( MBG_TGT_WIN32_PNP )
          rc = pcps_usb_init( pddev );
        #elif defined( MBG_TGT_LINUX )
          rc = usb_set_interface( pddev->udev, 0, 0 );

          if ( rc )
            _mbgddmsg_1( MBG_DBG_INIT_DEV, "usb_set_interface returned %d", rc );
          else
          {
            struct usb_host_interface *iface_desc = pddev->intf->cur_altsetting;
            int i;

            pddev->n_usb_ep = 0;

            for ( i = 0; i < iface_desc->desc.bNumEndpoints; i++ )
            {
              struct usb_endpoint_descriptor *endpoint = &iface_desc->endpoint[i].desc;
              PCPS_USB_EP *pep = &pddev->ep[i];
              pep->addr = endpoint->bEndpointAddress;
              pep->max_packet_size = le16_to_cpu( endpoint->wMaxPacketSize );
              _mbgddmsg_3( MBG_DBG_INIT_DEV, "endpoint %d: addr %02X, size: %d",
                           i, pep->addr, pep->max_packet_size );
              pddev->n_usb_ep++;
            }
          }
        #else
          #error USB endpoint configuration can not be determined for this target.
        #endif

        if ( mbg_rc_is_success( rc ) )
          if ( pddev->n_usb_ep < MBGUSB_MIN_ENDPOINTS_REQUIRED )
          {
            #if defined( MBG_TGT_LINUX )
              printk( KERN_INFO "device supports only %d endpoints while %d are required\n",
                      pddev->n_usb_ep, MBGUSB_MIN_ENDPOINTS_REQUIRED );
            #elif defined( MBG_TGT_WIN32_PNP )
              swprintf( pddev->wcs_msg, L"device supports only %d endpoints while %d are required",
                        pddev->n_usb_ep, MBGUSB_MIN_ENDPOINTS_REQUIRED );
              _evt_msg_w( GlbDriverObject, pddev->wcs_msg );
            #else
              //##++
            #endif

            rc = MBG_ERR_GENERIC;
          }
          #if defined( MBG_TGT_WIN32_PNP )
            else
            {
              LARGE_INTEGER Count1, Count2, PerfFreq;
              PCPS_HR_TIME t;

              uint8_t irq_cmd = PCPS_IRQ_1_SEC;
              rc = _pcps_usb_write_var( pddev, &irq_cmd );

              // get access time to determine latency compensation mode
              Count1 = KeQueryPerformanceCounter( &PerfFreq );
              rc = _pcps_read_var( pddev, PCPS_GIVE_HR_TIME, t );
              Count2 = KeQueryPerformanceCounter( NULL );

              // If access time is below 1 ms then there might be a V2.0 Hub between device and host.
              // In this case, you can expect that there is the 125 us microframe timing of USB 2.0.
              pddev->usb_20_mode = ( (ULONGLONG) ( Count2.QuadPart - Count1.QuadPart ) ) < ( ( (ULONGLONG) PerfFreq.QuadPart ) / 1000UL );
            }
          #endif

        if ( mbg_rc_is_error( rc ) )
        {
          _pcps_ddev_set_err_flags( pddev, PCPS_EF_IO_INIT );
          goto fail;
        }
        break;
      #endif

    case PCPS_BUS_PCI_PEX8311:
      port_rsrc_len[0] = 0;
      port_rsrc_len[1] = sizeof( PCI_ASIC );  // same as ASIC
      port_ranges_required = 2;  // will be swapped later
      status_port_offs = offsetof( PCI_ASIC, status_port ); // same as ASIC
      pddev->read = pcps_read_asic;
      break;

    case PCPS_BUS_PCI_ASIC:
    case PCPS_BUS_PCI_MBGPEX:
      port_rsrc_len[0] = sizeof( PCI_ASIC );
      port_ranges_required = 1;
      status_port_offs = offsetof( PCI_ASIC, status_port );
      pddev->read = pcps_read_asic;
      break;

    case PCPS_BUS_PCI_S5920:
      port_rsrc_len[0] = AMCC_OP_REG_RANGE_S5920;
      port_rsrc_len[1] = 16;  //##++
      port_ranges_required = 2;
      status_port_offs = AMCC_OP_REG_IMB4 + 3;
      pddev->read = pcps_read_amcc_s5920;
      break;

    case PCPS_BUS_PCI_S5933:
      port_rsrc_len[0] = AMCC_OP_REG_RANGE_S5933;
      port_ranges_required = 1;
      status_port_offs = AMCC_OP_REG_IMB4 + 3;
      pddev->read = pcps_read_amcc_s5933;
      break;

    case PCPS_BUS_MCA:
    case PCPS_BUS_ISA:
      // resource lengths have always been set
      port_ranges_required = 1;
      status_port_offs = 1;
      pddev->read = pcps_read_std;
      break;

    default:
      #if defined( MBG_TGT_LINUX )
        printk( KERN_ERR "%s: unhandled bus flags %04X for device %s\n",
                pcps_driver_name, _pcps_ddev_bus_flags( pddev ), _pcps_ddev_type_name( pddev ) );
      #elif defined( MBG_TGT_BSD )
        printf( "%s: unhandled bus flags %04X for device %s\n",
                pcps_driver_name, _pcps_ddev_bus_flags( pddev ), _pcps_ddev_type_name( pddev ) );
      #elif defined( MBG_TGT_WIN32 )
        swprintf( pddev->wcs_msg, L"unhandled bus flags %04X for device %S",
                  _pcps_ddev_bus_flags( pddev ), _pcps_ddev_type_name( pddev ) );
        _evt_msg_w( GlbDriverObject, pddev->wcs_msg );
      #endif
      goto fail;

  }  // switch ( _pcps_ddev_bus_flags( pddev ) )

  #if _PCPS_USE_MM_IO
  if ( !_pcps_ddev_is_pci_mbgpex( pddev ) || force_io_access )
  #endif
  {
    // check if all required I/O resources have been assigned
    if ( pddev->rsrc_info.num_rsrc_io < port_ranges_required )
    {
      _mbgddmsg_3( MBG_DBG_INIT_DEV, "PCPS start device %X fails: port ranges (%u) less than required (%u)",
                   _pcps_ddev_dev_id( pddev ), pddev->rsrc_info.num_rsrc_io, port_ranges_required );
      _pcps_ddev_set_err_flags( pddev, PCPS_EF_IO_INIT );
      goto fail;
    }
  }

  #if _PCPS_USE_MM_IO
    pddev->use_mm_io = 0;    // default, eventually changed later
  #endif

  if ( _pcps_ddev_is_pci_mbgpex( pddev ) )
  {
    pddev->irq_enb_disb_port = _pcps_ddev_io_base_mapped( pddev, 0 )
                             + offsetof( PCI_ASIC, control_status );
    pddev->irq_flag_port = pddev->irq_enb_disb_port;
    pddev->irq_flag_mask = PCI_ASIC_PCI_IRQF;

    pddev->irq_ack_port = pddev->irq_enb_disb_port;
    pddev->irq_ack_mask = PCI_ASIC_PCI_IRQF;

    _mbgddmsg_1( MBG_DBG_INIT_DEV, "%s: pcps_probe_device: interface is MBGPEX",
                 pcps_driver_name );

    #if _PCPS_USE_MM_IO
      if ( force_io_access )
      {
        #if defined( MBG_TGT_LINUX )  //##+++++++++++++++++++++++
          printk( KERN_INFO "%s: Forced I/O access for %s\n",
                  pcps_driver_name, _pcps_ddev_type_name( pddev ) );
        #endif
      }
      else
      {
        if ( map_sys_virtual_address( pddev ) < 0 )
          goto fail_with_cleanup;

        #if DEBUG_IO && defined( MBG_TGT_LINUX )
          printk( KERN_ERR "io_addr: 0x%llX, cmd: 0x%llX\n",
                  (unsigned long long) _pcps_ddev_io_base_mapped( pddev, 0 ),
                  (unsigned long long) _pcps_ddev_io_base_mapped( pddev, 0 ) + offsetof( PCI_ASIC, pci_data )
                );

          printk( KERN_ERR "mm_addr: %p, cmd: %p, tstamp: %p\n",
                  pddev->mm_addr,
                  &pddev->mm_addr->mbgpex.asic.pci_data.ul,
                  pddev->mm_tstamp_addr
                );
        #endif

        pddev->read = pcps_read_asic_mm;
        pddev->use_mm_io = 1;
      }
    #endif  // _PCPS_USE_MM_IO

    goto chip_setup_done;
  }


  // Setup additional properties depending on the
  // type of bus interface chip.

  if ( _pcps_ddev_is_pci_pex8311( pddev ) )
  {
    // I/O and memory ranges must be swapped for the
    // low level functions because otherwise the first
    // range addressed the chip configuration registers
    // while the second range addressed data registers.

    PCPS_MEM_RSRC tmp_mem_rsrc;
    PCPS_IO_RSRC tmp_io_rsrc;

    tmp_mem_rsrc = pddev->rsrc_info.mem[0];
    pddev->rsrc_info.mem[0] = pddev->rsrc_info.mem[1];
    pddev->rsrc_info.mem[1] = tmp_mem_rsrc;

    #if DEBUG_IO
      _mbgddmsg_4( MBG_DBG_DETAIL, "Ports before swapping: %04lX (%08lX), %04lX (%08lX)",
        (ulong) pddev->rsrc_info.port[0].base_raw, (ulong) pddev->rsrc_info.port[0].base_mapped,
        (ulong) pddev->rsrc_info.port[1].base_raw, (ulong) pddev->rsrc_info.port[1].base_mapped );
    #endif

    tmp_io_rsrc = pddev->rsrc_info.port[0];
    pddev->rsrc_info.port[0] = pddev->rsrc_info.port[1];
    pddev->rsrc_info.port[1] = tmp_io_rsrc;

    #if DEBUG_IO
      _mbgddmsg_4( MBG_DBG_DETAIL, "Ports after swapping: %04lX (%08lX), %04lX (%08lX)",
        (ulong) pddev->rsrc_info.port[0].base_raw, (ulong) pddev->rsrc_info.port[0].base_mapped,
        (ulong) pddev->rsrc_info.port[1].base_raw, (ulong) pddev->rsrc_info.port[1].base_mapped );
    #endif

    // Attention: the interrupt control/status register is located in
    // the PLX configuration space which is addressed by a different
    // port address range than the normal data ports !!
    pddev->irq_enb_disb_port = _pcps_ddev_io_base_mapped( pddev, 1 ) + PLX_LCS_INTCSR;
    pddev->irq_enb_mask = PLX_LCS_INTCSR_INT_ENB;
    pddev->irq_disb_mask = PLX_LCS_INTCSR_INT_ENB;

    pddev->irq_flag_port = _pcps_ddev_io_base_mapped( pddev, 1 ) + PLX_LCS_INTCSR;
    pddev->irq_flag_mask = PLX_LCS_INTCSR_INT_FLAG;

    pddev->irq_ack_port = _pcps_ddev_io_base_mapped( pddev, 0 ) + offsetof( PCI_ASIC, control_status );
    pddev->irq_ack_mask = PCI_ASIC_PCI_IRQF;
    goto chip_setup_done;
  }

  if ( _pcps_ddev_is_pci_asic( pddev ) )
  {
    pddev->irq_enb_disb_port = _pcps_ddev_io_base_mapped( pddev, 0 )
                             + offsetof( PCI_ASIC, control_status );
    pddev->irq_flag_port = pddev->irq_enb_disb_port;
    pddev->irq_flag_mask = PCI_ASIC_PCI_IRQF;

    pddev->irq_ack_port = pddev->irq_enb_disb_port;
    pddev->irq_ack_mask = PCI_ASIC_PCI_IRQF;
    goto chip_setup_done;
  }

  if ( _pcps_ddev_is_pci_amcc( pddev ) )
  {
    pddev->irq_enb_disb_port = _pcps_ddev_io_base_mapped( pddev, 0 )
                             + AMCC_OP_REG_INTCSR;
    pddev->irq_enb_mask = AMCC_INT_ENB;
    pddev->irq_disb_mask = AMCC_INT_MASK;

    pddev->irq_flag_port = pddev->irq_enb_disb_port;
    pddev->irq_flag_mask = AMCC_INT_FLAG;

    pddev->irq_ack_port = pddev->irq_enb_disb_port;
    pddev->irq_ack_mask = AMCC_INT_ACK;
    goto chip_setup_done;
  }

chip_setup_done:

  // Chip-specific setup done. Continue with common device setup.

  pddev->status_port = _pcps_ddev_io_base_mapped( pddev, 0 ) + status_port_offs;

  // Set up the resource list in pddev->dev.cfg which
  // isn't really required anymore, but just informational:

  for ( i = 0; i < N_PCPS_PORT_RSRC; i++ )
  {
    PCPS_IO_RSRC *prsrc;

    if ( i >= port_ranges_required )
      break;

    prsrc = &pddev->rsrc_info.port[i];

    // If the resource len has not yet been set
    // then use the default resource len.
    if ( prsrc->num == 0 )
      prsrc->num = port_rsrc_len[i];

    rsrc_port_to_cfg_port( &pddev->dev.cfg.port[i], &pddev->rsrc_info.port[i] );
  }

  pddev->dev.cfg.irq_num = pddev->rsrc_info.num_rsrc_irq ?
                           pddev->rsrc_info.irq.num : -1;
  pddev->dev.cfg.short_status_port = _pcps_ddev_short_port_base( pddev, 0 ) + status_port_offs;

  pddev->dev.cfg.timeout_clk = PCPS_TIMEOUT_CNT;

  #if DEBUG_PORTS
    _mbgddmsg_3( MBG_DBG_DETAIL, "IRQ enb/disb port: %04lX, enb: %08lX, disb: %08lX",
                 (ulong) pddev->irq_enb_disb_port,
                 (ulong) pddev->irq_enb_mask,
                 (ulong) pddev->irq_disb_mask
               );
    _mbgddmsg_2( MBG_DBG_DETAIL, "IRQ flag port: %04lX, mask: %08lX",
                 (ulong) pddev->irq_flag_port,
                 (ulong) pddev->irq_flag_mask
               );
    _mbgddmsg_2( MBG_DBG_DETAIL, "IRQ ack port: %04lX, mask: %08lX",
                 (ulong) pddev->irq_ack_port,
                 (ulong) pddev->irq_ack_mask
               );
    _mbgddmsg_1( MBG_DBG_DETAIL, "status port: %04lX", (ulong) pddev->status_port );
  #endif

  #if _PCPS_USE_RSRCMGR
    rc = pcps_rsrc_claim( pddev );

    if ( mbg_rc_is_error( rc ) )
    {
      _mbgddmsg_1( MBG_DBG_INIT_DEV, "PCPS start device %X: failed to alloc resources",
                   _pcps_ddev_dev_id( pddev ) );

      goto fail_with_cleanup;
    }
  #endif

  // There are some BIOSs out there which don't configure some PEX cards
  // properly, and thus the cards can not be accessed properly.
  // See note near the definition of ::_pcps_pci_cfg_err for details.
  if ( _pcps_ddev_pci_cfg_err( pddev ) )
  {
    #if defined( MBG_TGT_LINUX )
      printk( KERN_WARNING "%s: duplicate base address 0x%04lX, device %s will not work properly (BIOS faulty)\n",
              pcps_driver_name, (ulong) _pcps_ddev_io_base_raw( pddev, 0 ), _pcps_ddev_type_name( pddev ) );
    #elif defined( MBG_TGT_BSD )
      printf( "%s: duplicate base address 0x%04lX, device %s will not work properly (BIOS faulty)\n",
              pcps_driver_name, (ulong) _pcps_ddev_io_base_raw( pddev, 0 ), _pcps_ddev_type_name( pddev ) );
    #elif defined( MBG_TGT_WIN32 )
      swprintf( pddev->wcs_msg, L"duplicate base address 0x%04lX, device %s will not work properly (BIOS faulty)",
                (ulong) _pcps_ddev_io_base_raw( pddev, 0 ), _pcps_ddev_type_name( pddev ) );
      _evt_msg_w( GlbDriverObject, pddev->wcs_msg );
    #endif
  }


  #if 0 && DEBUG //###++++++++++++++ TODO testing only
  {
    MBG_SYS_UPTIME uptime;
    mbg_get_sys_uptime( &uptime );
    mbg_sleep_sec( 1 );
    mbg_get_sys_uptime( &uptime );
  }
  #endif

  // A PTP270PEX card must have finished booting
  // before it can be accessed.
  if ( pcps_ddev_is_ptp270pex( pddev ) )
    wait_ptp270pex_ready( pddev );

  // try to read firmware ID
  rc = pcps_get_fw_id( pddev, pddev->dev.cfg.fw_id );

  if ( mbg_rc_is_error( rc ) )
  {
    if ( _pcps_ddev_is_isa( pddev ) )
    {
      // ISA devices are detected by trying to read a firmware ID via
      // a given port, so if the firmware ID could not be read then this
      // just means there is no device using the given port address.
      #if defined( MBG_TGT_WIN32 )
        swprintf( pddev->wcs_msg, L"No ISA card found at port %03lXh.",
                  (ulong) _pcps_ddev_io_base_raw( pddev, 0 ) );
        _evt_msg_w( GlbDriverObject, pddev->wcs_msg );
      #else
        _mbgddmsg_1( MBG_DBG_INIT_DEV, "No ISA card found at port %03lXh.",
                     (ulong) _pcps_ddev_io_base_raw( pddev, 0 ) );
      #endif
    }
    else
    {
      // Non-ISA devices are detected by some other means, so if the firmware
      // ID could not be read this is a serious error.
      #if defined( MBG_TGT_WIN32 )  //##+++ debug or not debug ... ;-)
        _evt_msg_w( GlbDriverObject, L"StartDevice: failed to read firmware ID" );
      #else
        _mbgddmsg_1( MBG_DBG_INIT_DEV, "PCPS start device %X: failed to read firmware ID",
                     _pcps_ddev_dev_id( pddev ) );
      #endif
    }

    _pcps_ddev_set_err_flags( pddev, PCPS_EF_TIMEOUT );
    goto fail_with_cleanup;
  }

  if ( _pcps_ddev_bus_flags( pddev ) == PCPS_BUS_ISA )
  {
    ushort dev_type;

    // Still need to find out which type of ISA clock we have found.
    // Check EPROM ID to find out which kind of clock is installed.
    if ( pcps_check_id( pddev, fw_id_ref_gps ) == MBG_SUCCESS )
      dev_type = PCPS_TYPE_GPS167PC;
    else
    {
      if ( mbg_rc_is_success( pcps_check_id( pddev, fw_id_ref_pcps ) ) )
      {
        // Device is a PC31, or a PC32 if it has signature code.
        // If no support for MCA has been compiled in, it may even
        // be a PS31 which is software compatible with a PC31.
        dev_type =
          ( _mbg_inp16_to_cpu( pddev, 0, _pcps_ddev_io_base_mapped( pddev, 0 ) + 2 )
            == pcps_dev_type[PCPS_TYPE_PC32].dev_id ) ?
         PCPS_TYPE_PC32 : PCPS_TYPE_PC31;
      }
      else
      {
        _pcps_ddev_set_err_flags( pddev, PCPS_EF_INV_FW_ID );
        goto fail_with_cleanup;
      }
    }

    pddev->dev.type = pcps_dev_type[dev_type];
  }

  #if defined( MBG_TGT_OS2 )
    pcps_rsrc_register_device( pddev );
  #endif

  // Extract the firmware revision number from the ID string.
  pddev->dev.cfg.fw_rev_num = pcps_get_rev_num( _pcps_ddev_fw_id( pddev ) );

  // If the device has an ASIC or EPLD read the ASIC version number
  if ( _pcps_ddev_has_asic_version( pddev ) )
  {
  #if _PCPS_USE_MM_IO
    if ( pddev->use_mm_io )
    {
      pddev->raw_asic_version = pddev->mm_addr->mbgpex.asic.raw_version;
      _mbg_swab_asic_version( &pddev->raw_asic_version );
      _mbgddmsg_2( MBG_DBG_INIT_DEV, "%s: pcps_probe_device: raw ASIC version %04X read via MM",
                   pcps_driver_name, pddev->raw_asic_version );
    }
    else
  #endif  // _PCPS_USE_MM_IO
    {
      pddev->raw_asic_version = _mbg_inp32_to_cpu( pddev, 0, _pcps_ddev_io_base_mapped( pddev, 0 )
                                + offsetof( PCI_ASIC, raw_version ) );
      _mbg_swab_asic_version( &pddev->raw_asic_version );
      _mbgddmsg_2( MBG_DBG_INIT_DEV, "%s: pcps_probe_device: raw ASIC version %04X read via I/O",
                   pcps_driver_name, pddev->raw_asic_version );
    }

    pddev->asic_version = _convert_asic_version_number( pddev->raw_asic_version );
  }

  // Setup some feature flags which depend on the device type
  // and firmware version.
  switch ( _pcps_ddev_type_num( pddev ) )
  {
    case PCPS_TYPE_PC31:
    case PCPS_TYPE_PS31_OLD:
    case PCPS_TYPE_PS31:
      pddev->builtin_features = BUILTIN_FEAT_UNDEFINED;
      pddev->dev.cfg.features = PCPS_FEAT_PC31PS31;
      check_feature( pddev, REV_CAN_SET_TIME_PC31PS31, PCPS_CAN_SET_TIME );
      check_feature( pddev, REV_HAS_SERIAL_PC31PS31, PCPS_HAS_SERIAL );
      check_feature( pddev, REV_HAS_SYNC_TIME_PC31PS31, PCPS_HAS_SYNC_TIME );
      check_feature( pddev, REV_HAS_UTC_OFFS_PC31PS31, PCPS_HAS_UTC_OFFS );
      break;

    case PCPS_TYPE_PC32:
      pddev->builtin_features = BUILTIN_FEAT_UNDEFINED;
      pddev->dev.cfg.features = PCPS_FEAT_PC32;
      break;

    case PCPS_TYPE_PCI32:
      pddev->builtin_features = BUILTIN_FEAT_UNDEFINED;
      pddev->dev.cfg.features = PCPS_FEAT_PCI32;
      break;

    case PCPS_TYPE_GPS167PC:
      pddev->builtin_features = BUILTIN_FEAT_GPS167PC;
      pddev->dev.cfg.features = PCPS_FEAT_GPS167PC;
      check_feature( pddev, REV_HAS_HR_TIME_GPS167PC, PCPS_HAS_HR_TIME );
      check_feature( pddev, REV_HAS_CABLE_LEN_GPS167PC, PCPS_HAS_CABLE_LEN );
      break;

    case PCPS_TYPE_GPS167PCI:
      pddev->builtin_features = BUILTIN_FEAT_GPS167PCI;
      pddev->dev.cfg.features = PCPS_FEAT_GPS167PCI;
      check_feature( pddev, REV_HAS_CABLE_LEN_GPS167PCI, PCPS_HAS_CABLE_LEN );
      check_feature( pddev, REV_CAN_CLR_UCAP_BUFF_GPS167PCI, PCPS_CAN_CLR_UCAP_BUFF );
      check_feature( pddev, REV_HAS_UCAP_GPS167PCI, PCPS_HAS_UCAP );
      break;

    case PCPS_TYPE_PCI509:
      pddev->builtin_features = BUILTIN_FEAT_UNDEFINED;
      pddev->dev.cfg.features = PCPS_FEAT_PCI509;
      break;

    case PCPS_TYPE_GPS168PCI:
      pddev->builtin_features = BUILTIN_FEAT_GPS168PCI;
      pddev->dev.cfg.features = PCPS_FEAT_GPS168PCI;
      check_feature( pddev, REV_CAN_CLR_UCAP_BUFF_GPS168PCI, PCPS_CAN_CLR_UCAP_BUFF );
      check_feature( pddev, REV_HAS_UCAP_GPS168PCI, PCPS_HAS_UCAP );
      break;

    case PCPS_TYPE_PCI510:
      pddev->builtin_features = BUILTIN_FEAT_UNDEFINED;
      pddev->dev.cfg.features = PCPS_FEAT_PCI510;
      break;

    case PCPS_TYPE_GPS169PCI:
      pddev->builtin_features = BUILTIN_FEAT_GPS169PCI;
      pddev->dev.cfg.features = PCPS_FEAT_GPS169PCI;
      check_feature( pddev, REV_HAS_GPS_DATA_16_GPS169PCI, PCPS_HAS_GPS_DATA_16 );
      break;

    case PCPS_TYPE_TCR510PCI:
      pddev->builtin_features = BUILTIN_FEAT_UNDEFINED;
      pddev->dev.cfg.features = PCPS_FEAT_TCR510PCI;
      check_feature( pddev, REV_HAS_HR_TIME_TCR510PCI, PCPS_HAS_HR_TIME );
      break;

    case PCPS_TYPE_TCR167PCI:
      pddev->builtin_features = BUILTIN_FEAT_TCR167PCI;
      pddev->dev.cfg.features = PCPS_FEAT_TCR167PCI;
      break;

    case PCPS_TYPE_GPS170PCI:
      pddev->builtin_features = BUILTIN_FEAT_GPS170PCI;
      pddev->dev.cfg.features = PCPS_FEAT_GPS170PCI;
      break;

    case PCPS_TYPE_PCI511:
      pddev->builtin_features = BUILTIN_FEAT_UNDEFINED;
      pddev->dev.cfg.features = PCPS_FEAT_PCI511;
      check_feature( pddev, REV_HAS_HR_TIME_PCI511, PCPS_HAS_HR_TIME );
      break;

    case PCPS_TYPE_TCR511PCI:
      pddev->builtin_features = BUILTIN_FEAT_UNDEFINED;
      pddev->dev.cfg.features = PCPS_FEAT_TCR511PCI;
      check_feature( pddev, REV_HAS_IRIG_CTRL_BITS_TCR511PCI, PCPS_HAS_IRIG_CTRL_BITS );
      check_feature( pddev, REV_HAS_IRIG_TIME_TCR511PCI, PCPS_HAS_IRIG_TIME );
      check_feature( pddev, REV_HAS_RAW_IRIG_DATA_TCR511PCI, PCPS_HAS_RAW_IRIG_DATA );
      break;

    case PCPS_TYPE_PEX511:
      pddev->builtin_features = BUILTIN_FEAT_UNDEFINED;
      pddev->dev.cfg.features = PCPS_FEAT_PEX511;
      // HR time support for the PEX511 requires both a certain ASIC
      // version plus a certain firmware version.
      if ( _pcps_asic_version_greater_equal( _pcps_ddev_asic_version( pddev ),
           PCI_ASIC_MAJOR_PEX511, PCI_ASIC_HR_TIME_MINOR_PEX511 ) )
        check_feature( pddev, REV_HAS_HR_TIME_PEX511, PCPS_HAS_HR_TIME );

      pcps_check_pex_irq_unsafe( pddev, REV_HAS_IRQ_FIX_MINOR_PEX511,
           PCI_ASIC_MAJOR_PEX511, PCI_ASIC_FIX_IRQ_MINOR_PEX511 );
      break;

    case PCPS_TYPE_TCR511PEX:
      pddev->builtin_features = BUILTIN_FEAT_UNDEFINED;
      pddev->dev.cfg.features = PCPS_FEAT_TCR511PEX;
      check_feature( pddev, REV_HAS_IRIG_CTRL_BITS_TCR511PEX, PCPS_HAS_IRIG_CTRL_BITS );
      check_feature( pddev, REV_HAS_IRIG_TIME_TCR511PEX, PCPS_HAS_IRIG_TIME );
      check_feature( pddev, REV_HAS_RAW_IRIG_DATA_TCR511PEX, PCPS_HAS_RAW_IRIG_DATA );
      pcps_check_pex_irq_unsafe( pddev, REV_HAS_IRQ_FIX_MINOR_TCR511PEX,
           PCI_ASIC_MAJOR_TCR511PEX, PCI_ASIC_FIX_IRQ_MINOR_TCR511PEX );
      break;

    case PCPS_TYPE_GPS170PEX:
      pddev->builtin_features = BUILTIN_FEAT_GPS170PEX;
      pddev->dev.cfg.features = PCPS_FEAT_GPS170PEX;
      pcps_check_pex_irq_unsafe( pddev, REV_HAS_IRQ_FIX_MINOR_GPS170PEX,
           PCI_ASIC_MAJOR_GPS170PEX, PCI_ASIC_FIX_IRQ_MINOR_GPS170PEX );
      break;

    case PCPS_TYPE_USB5131:
      pddev->builtin_features = BUILTIN_FEAT_UNDEFINED;
      pddev->dev.cfg.features = PCPS_FEAT_USB5131;
      break;

    case PCPS_TYPE_TCR51USB:
      pddev->builtin_features = BUILTIN_FEAT_UNDEFINED;
      pddev->dev.cfg.features = PCPS_FEAT_TCR51USB;
      check_feature( pddev, REV_HAS_IRIG_CTRL_BITS_TCR51USB, PCPS_HAS_IRIG_CTRL_BITS );
      check_feature( pddev, REV_HAS_IRIG_TIME_TCR51USB, PCPS_HAS_IRIG_TIME );
      check_feature( pddev, REV_HAS_RAW_IRIG_DATA_TCR51USB, PCPS_HAS_RAW_IRIG_DATA );
      break;

    case PCPS_TYPE_MSF51USB:
      pddev->builtin_features = BUILTIN_FEAT_UNDEFINED;
      pddev->dev.cfg.features = PCPS_FEAT_MSF51USB;
      break;

    case PCPS_TYPE_PTP270PEX:
      pddev->builtin_features = BUILTIN_FEAT_PTP270PEX;
      pddev->dev.cfg.features = PCPS_FEAT_PTP270PEX;
      break;

    case PCPS_TYPE_FRC511PEX:
      pddev->builtin_features = BUILTIN_FEAT_FRC511PEX;
      pddev->dev.cfg.features = PCPS_FEAT_FRC511PEX;
      break;

    case PCPS_TYPE_TCR170PEX:
      pddev->builtin_features = BUILTIN_FEAT_TCR170PEX;
      pddev->dev.cfg.features = PCPS_FEAT_TCR170PEX;
      break;

    case PCPS_TYPE_WWVB51USB:
      pddev->builtin_features = BUILTIN_FEAT_WWVB511;
      pddev->dev.cfg.features = PCPS_FEAT_WWVB51USB;
      break;

    case PCPS_TYPE_GPS180PEX:
      pddev->builtin_features = BUILTIN_FEAT_GPS180PEX;
      pddev->dev.cfg.features = PCPS_FEAT_GPS180PEX;
      break;

    case PCPS_TYPE_TCR180PEX:
      pddev->builtin_features = BUILTIN_FEAT_TCR180PEX;
      pddev->dev.cfg.features = PCPS_FEAT_TCR180PEX;
      break;

    case PCPS_TYPE_DCF600USB:
      pddev->builtin_features = BUILTIN_FEAT_UNDEFINED;
      pddev->dev.cfg.features = PCPS_FEAT_DCF600USB;
      break;

    case PCPS_TYPE_PZF180PEX:
      pddev->builtin_features = BUILTIN_FEAT_PZF180PEX;
      pddev->dev.cfg.features = PCPS_FEAT_PZF180PEX;
      break;

    case PCPS_TYPE_TCR600USB:
      pddev->builtin_features = BUILTIN_FEAT_UNDEFINED;
      pddev->dev.cfg.features = PCPS_FEAT_TCR600USB;
      break;

    case PCPS_TYPE_MSF600USB:
      pddev->builtin_features = BUILTIN_FEAT_UNDEFINED;
      pddev->dev.cfg.features = PCPS_FEAT_MSF600USB;
      break;

    case PCPS_TYPE_WVB600USB:
      pddev->builtin_features = BUILTIN_FEAT_UNDEFINED;
      pddev->dev.cfg.features = PCPS_FEAT_WVB600USB;
      break;

    case PCPS_TYPE_GLN180PEX:
      pddev->builtin_features = BUILTIN_FEAT_GLN180PEX;
      pddev->dev.cfg.features = PCPS_FEAT_GLN180PEX;
      break;

    case PCPS_TYPE_GPS180AMC:
      pddev->builtin_features = BUILTIN_FEAT_GPS180AMC;
      pddev->dev.cfg.features = PCPS_FEAT_GPS180AMC;
      break;

    case PCPS_TYPE_GNS181PEX:
      pddev->builtin_features = BUILTIN_FEAT_GNS181PEX;
      pddev->dev.cfg.features = PCPS_FEAT_GNS181PEX;
      break;

    default:
      #if defined( MBG_TGT_LINUX )
        printk( KERN_WARNING "%s: no feature detection for device %s\n",
                pcps_driver_name, _pcps_ddev_type_name( pddev ) );
      #elif defined( MBG_TGT_BSD )
        printf( "%s: no feature detection for device %s\n",
                pcps_driver_name, _pcps_ddev_type_name( pddev ) );
      #elif defined( MBG_TGT_WIN32 )
        swprintf( pddev->wcs_msg, L"no feature detection for device %S",
                  _pcps_ddev_type_name( pddev ) );
        _evt_msg_w( GlbDriverObject, pddev->wcs_msg );
      #endif
      goto fail_with_cleanup;

  }  // switch


  if ( _pcps_ddev_has_receiver_info( pddev ) )
  {
    rc = _pcps_read_gps_var( pddev, PC_GPS_RECEIVER_INFO, *p_ri );

    if ( mbg_rc_is_success( rc ) )
    {
      _mbg_swab_receiver_info( p_ri );
      _mbgddmsg_1( MBG_DBG_INIT_DEV, "Successfully read receiver info from dev %X",
                   _pcps_ddev_dev_id( pddev ) );

      goto receiver_info_done;
    }

    _mbgddmsg_1( MBG_DBG_INIT_DEV, "Failed to read receiver info from dev %X",
                 _pcps_ddev_dev_id( pddev ) );

  }

  _mbgddmsg_1( MBG_DBG_INIT_DEV, "Setting up default receiver info for dev %X",
               _pcps_ddev_dev_id( pddev ) );

  if ( _pcps_ddev_is_gps( pddev ) )
    _setup_default_receiver_info_gps( p_ri );
  else
    _setup_default_receiver_info_dcf( p_ri, &pddev->dev );

receiver_info_done:

  #if defined( DEBUG )
    _mbgddmsg_1( MBG_DBG_DETAIL, "ri.model_code: 0x%04X",
                 p_ri->model_code );

    _mbgddmsg_3( MBG_DBG_DETAIL, "ri.model_name: \"%-*.*s\"",
                 (int) sizeof( p_ri->model_name ),
                 (int) sizeof( p_ri->model_name ),
                 p_ri->model_name );

    _mbgddmsg_4( MBG_DBG_DETAIL, "ri.sw_rev: code 0x%04X, name \"%-*.*s\"",
                 p_ri->sw_rev.code,
                 (int) sizeof( p_ri->sw_rev.name ),
                 (int) sizeof( p_ri->sw_rev.name ),
                 p_ri->sw_rev.name );

    _mbgddmsg_3( MBG_DBG_DETAIL, "ri.sernum: \"%-*.*s\"",
                 (int) sizeof( p_ri->sernum ),
                 (int) sizeof( p_ri->sernum ),
                 p_ri->sernum );

    _mbgddmsg_3( MBG_DBG_DETAIL, "ri.epld_name: \"%-*.*s\"",
                 (int) sizeof( p_ri->epld_name ),
                 (int) sizeof( p_ri->epld_name ),
                 p_ri->epld_name );

    _mbgddmsg_1( MBG_DBG_DETAIL, "ri.n_channels: %u",
                 p_ri->n_channels );

    _mbgddmsg_1( MBG_DBG_DETAIL, "ri.ticks_per_sec: %lu",
                 (ulong) p_ri->ticks_per_sec );

    _mbgddmsg_1( MBG_DBG_DETAIL, "ri.features: 0x%08lX",
                 (ulong) p_ri->features );

    _mbgddmsg_2( MBG_DBG_DETAIL, "ri.fixed_freq: khz_val %u, range %i",
                 p_ri->fixed_freq.khz_val,
                 p_ri->fixed_freq.range );

    _mbgddmsg_2( MBG_DBG_DETAIL, "ri.osc type: %u, flags: 0x%02X",
                 p_ri->osc_type, p_ri->osc_flags );

    _mbgddmsg_1( MBG_DBG_DETAIL, "ri.n_ucaps: %u",
                 p_ri->n_ucaps );

    _mbgddmsg_1( MBG_DBG_DETAIL, "ri.n_com_ports: %u",
                 p_ri->n_com_ports );

    _mbgddmsg_1( MBG_DBG_DETAIL, "ri.n_str_type: %u",
                 p_ri->n_str_type );

    _mbgddmsg_1( MBG_DBG_DETAIL, "ri.n_prg_out: %u",
                 p_ri->n_prg_out );

    _mbgddmsg_1( MBG_DBG_DETAIL, "ri.flags: 0x%04X",
                 p_ri->flags );
#endif

#if 0 //###+++++++++ TODO check if this is reasonable

  // Make sure this program supports at least as many ports as
  // the current clock device.
  if ( pddev->ri.n_com_ports > MAX_PARM_PORT )
  {
    _mbgddmsg_3( MBG_DBG_INIT_DEV, "%s provides %i COM ports, but this driver only supports %i",
                 _pcps_ddev_type_name( pddev ), pddev->ri.n_com_ports, MAX_PARM_PORT );
    pddev->ri.n_com_ports = MAX_PARM_PORT;
  }

  // Make sure this program supports at least as many string types
  // as the current clock device.
  if ( pddev->ri.n_str_type > MAX_PARM_STR_TYPE )
  {
    _mbgddmsg_3( MBG_DBG_INIT_DEV, "%s supports %i serial string formats, but this driver only supports %i",
                 _pcps_ddev_type_name( pddev ), pddev->ri.n_str_type, MAX_PARM_STR_TYPE );
    pddev->ri.n_str_type = MAX_PARM_STR_TYPE;
  }
#endif

  // detect the presence of some optional features at run time
  _mbgddmsg_3( MBG_DBG_INIT_DEV, "%s v%03X RECEIVER_INFO features: 0x%08lX",
               _pcps_ddev_type_name( pddev ), _pcps_ddev_fw_rev_num( pddev ),
               (ulong) *_ri_addr( pddev ).features );

  check_ri_features( pddev );

  _mbgddmsg_3( MBG_DBG_INIT_DEV, "%s v%03X actual features 1: 0x%08lX",
               _pcps_ddev_type_name( pddev ), _pcps_ddev_fw_rev_num( pddev ),
               (ulong) _pcps_ddev_features( pddev ) );

  #if !defined( MBG_TGT_OS2 ) && !defined( MBG_TGT_BSD )
    // Function strstr may not be supported at kernel level,
    // but this is not required, in most cases, anyway.
    if ( strstr( _pcps_ddev_fw_id( pddev ), "CERN" ) != NULL )
      pddev->dev.cfg.features |= PCPS_HAS_EVENT_TIME;
  #endif


  // Check if the device supports extended features

  rc = pcps_chk_dev_feat( pddev, DEV_FEAT_REQ_TYPE_RI, GPS_FEAT_XFEATURE );

  if ( mbg_rc_is_success( rc ) )
  {
    _mbgddmsg_1( MBG_DBG_INIT_DEV, "%s going to read XFEATURE_BUFFER", _pcps_ddev_type_name( pddev ) );

    rc = _pcps_read_gps_var( pddev, PC_GPS_XFEATURE_BUFFER, *_xfeat_addr( pddev ) );

    if ( mbg_rc_is_error( rc ) )
    {
      #if defined( MBG_TGT_LINUX )
        printk( KERN_ERR "%s: failed to read XFEATURE_BUFFER from device %s\n",
                pcps_driver_name, _pcps_ddev_type_name( pddev ) );
      #endif
      memset( _xfeat_addr( pddev ), 0, sizeof( *_xfeat_addr( pddev ) ) );
    }
  }


  // Check if the device supports the TLV API

  rc = pcps_chk_dev_feat( pddev, DEV_FEAT_REQ_TYPE_XFEAT, MBG_XFEATURE_TLV_API );

  if ( mbg_rc_is_success( rc ) )
  {
    _mbgddmsg_1( MBG_DBG_INIT_DEV, "%s going to read TLV_INFO", _pcps_ddev_type_name( pddev ) );

    rc = _pcps_read_gps_var( pddev, PC_GPS_TLV_INFO, *_tlv_info_addr( pddev ) );

    if ( mbg_rc_is_error( rc ) )
    {
      #if defined( MBG_TGT_LINUX )
        printk( KERN_ERR "%s: failed to read TLV_INFO from device %s\n",
                pcps_driver_name, _pcps_ddev_type_name( pddev ) );
      #endif
      memset( _tlv_info_addr( pddev ), 0, sizeof( *_tlv_info_addr( pddev ) ) );
    }
  }


  #if DEBUG_IO && defined( MBG_TGT_LINUX )
  {
    PCPS_TIME t = { 0 };
    printk( KERN_INFO "reading current time as test:\n" );
    rc = _pcps_read( pddev, PCPS_GIVE_TIME, &t, sizeof( t ) );
    printk( KERN_INFO "read time, sz: %lu, returned %i\n", (ulong) sizeof( t ), rc );
    printk( KERN_INFO "  sec100 %02X, sec %02X, min %02X hour %02X\n",
            t.sec100, t.sec, t.min, t.hour );
    printk( KERN_INFO "  mday %02X, wday %02X, month %02X year %02X\n",
            t.mday, t.wday, t.month, t.year );
    printk( KERN_INFO "  status %02X, sig %02X, offs_utc %02X\n",
            t.status, t.signal, t.offs_utc );
  }
  #endif

  if ( _pcps_ddev_has_asic_features( pddev ) )
  {
    pddev->asic_features = _mbg_inp32_to_cpu( pddev, 0, _pcps_ddev_io_base_mapped( pddev, 0 )
       + offsetof( PCI_ASIC, features ) );

    _mbg_swab_asic_features( &pddev->asic_features );

    _mbgddmsg_3( MBG_DBG_INIT_DEV, "%s v%03X asic_features: 0x%08lX",
                 _pcps_ddev_type_name( pddev ), _pcps_ddev_fw_rev_num( pddev ),
                 (ulong) pddev->asic_features );

    #if MBG_TGT_SUPP_MEM_ACC
      if ( pddev->asic_features & PCI_ASIC_HAS_MM_IO )
        pddev->dev.cfg.features |= PCPS_HAS_FAST_HR_TSTAMP;
      else
        if ( pddev->dev.cfg.features & PCPS_HAS_FAST_HR_TSTAMP )
        {
          // The device supports memory mapped time stamps by default.
          // However, this is not reflected by the ASIC features.
          _mbgddmsg_0( MBG_DBG_INIT_DEV,
                       "Warning: ASIC features don't reflect memory mapped time stamp support." );
        }

      if ( !has_mapped_sys_virtual_address( pddev ) )
        if ( pddev->dev.cfg.features & PCPS_HAS_FAST_HR_TSTAMP )
          if ( map_sys_virtual_address( pddev ) < 0 )
            goto fail_with_cleanup;
    #endif
  }

  pcps_read_sernum( pddev );

  _mbgddmsg_3( MBG_DBG_INIT_DEV, "%s v%03X actual features: 0x%08lX",
               _pcps_ddev_type_name( pddev ), _pcps_ddev_fw_rev_num( pddev ),
               (ulong) _pcps_ddev_features( pddev ) );

  return MBG_SUCCESS;


fail_with_cleanup:
  pcps_cleanup_device( pddev );

fail:
  return MBG_ERR_GENERIC;

}  // pcps_probe_device



/*HDR*/
/**
 * @brief Clean up function called by ::pcps_probe_device on error
 *
 * @param[in,out]  pddev  Pointer to the device structure
 *
 * @see ::pcps_probe_device
 */
void pcps_cleanup_device( PCPS_DDEV *pddev )
{
  pddev->read = pcps_read_null;

  #if MBG_TGT_SUPP_MEM_ACC
    unmap_sys_virtual_address( pddev );
  #endif

  #if _PCPS_USE_RSRCMGR
    pcps_rsrc_release( pddev );
  #endif

}  // pcps_cleanup_device



#if _PCPS_USE_PCI_BIOS

static /*HDR*/
/**
 * @brief Read PCI recources for a PCI device on non-PnP systems
 *
 * This function should be called by the probe routines of any
 * target-specific kernel drivers.
 * If the device is supported then all specific information including
 * supported features is read from the device and stored in sub-structures
 * of the device structure addressed by pdev.
 *
 * @param[in]      bus_num      The PCI bus number
 * @param[in]      dev_fnc_num  The PCI device/function number
 * @param[in,out]  pddev        Pointer to the device structure to be set up
 *
 * @return 0 on success, or a combination of @ref PCPS_ERR_FLAG_MASKS masks on error
 */
PCPS_ERR_FLAGS pcps_read_pci_rsrc( PCPS_BUS_NUM bus_num,
                                   PCPS_SLOT_NUM dev_fnc_num,
                                   PCPS_DDEV *pddev )
{
  PCPS_ERR_FLAGS err_flags = 0;
  uchar irq;
  short rc;
  PCI_DWORD dw;
  int i;

  // Clear resources
  memset( &pddev->rsrc_info, 0, sizeof( pddev->rsrc_info ) );

  for ( i = 0; i < MAX_PCPS_RSRC; i++ )
  {
    rc = _mbg_pci_read_cfg_dword( bus_num, dev_fnc_num,
           PCI_CS_BASE_ADDRESS_0 + i * sizeof( uint32_t ), &dw );

    if ( rc != PCI_SUCCESS )
      break;

    if ( dw == 0 )   // base address register not used
      continue;

    if ( dw & 0x0001 )   // is an I/O resource
    {
      if ( dw & 0xFFFF0000UL )
      {
        // The PCI interface chip is not initialized. This
        // should occur ONLY at the first-time installation
        // at the factory.
        err_flags |= PCPS_EF_IO_INIT;
        goto done;
      }

      pcps_add_rsrc_io( pddev, (uint16_t) ( dw & ~0x0001 ), 0 );
    }
    else
      pcps_add_rsrc_mem( pddev, dw, 0 );  //##++ range length?
  }

  // Read the interrupt line assigned to the clock.
  // The standard functions, however, don't use any
  // interrupt.
  rc = _mbg_pci_read_cfg_byte( bus_num, dev_fnc_num,
                               PCI_CS_INTERRUPT_LINE, &irq );

  if ( rc == PCI_SUCCESS )
    pcps_add_rsrc_irq( pddev, irq );

done:
  return err_flags;

}  // pcps_read_pci_rsrc



static /*HDR*/
/**
 * @brief Enable a PCI device which has not yet been enabled
 *
 * In PnP systems only PCI devices that are required for booting
 * are always enabled by the PCI BIOS. Other PCI devices may only
 * be fully enabled if a PC BIOS setup option "PNP OS installed"
 * is set to "NO".
 * If this option is set to "YES" then I/O access to the board
 * may still be disabled, and a PnP operating system is expected
 * to enable remaining devices when it boots up.
 * Drivers can call this function to enable I/O and memory access,
 * if required, in case this hasn't been done by the PCI BIOS
 * or operating system.
 *
 * @param[in]  bus_num       The PCI bus number
 * @param[in]  dev_fnc_num   The PCI device/function number
 * @param[in]  num_rsrc_mem  If != 0 then memory access is also enabled
 *
 * @return 0 on success, or a combination of @ref PCPS_ERR_FLAG_MASKS masks on error
 */
PCPS_ERR_FLAGS pcps_enable_pci_dev( PCPS_BUS_NUM bus_num,
                                    PCPS_SLOT_NUM dev_fnc_num,
                                    int num_rsrc_mem )
{
  PCPS_ERR_FLAGS err_flags = 0;
  uint16_t pci_command;      // current value read from the PCI cfg "command" register
  uint16_t new_pci_command;  // new value to be written to the PCI cfg "command" register
  int rc;


  rc = _mbg_pci_read_cfg_word( bus_num, dev_fnc_num,
                               PCI_CS_COMMAND, &pci_command );
  new_pci_command = pci_command | PCI_CMD_ENB_IO_ACC;

  if ( num_rsrc_mem )
    new_pci_command |= PCI_CMD_ENB_MEM_ACC;

  if ( new_pci_command != pci_command )
  {
    rc = _mbg_pci_write_cfg_word( bus_num, dev_fnc_num,
                                  PCI_CS_COMMAND, pci_command );
    if ( rc != PCI_SUCCESS )
    {
      err_flags |= PCPS_EF_IO_ENB;

      _mbgddmsg_1( MBG_DBG_INIT_DEV,
                   "PCI enable device returned %d", rc );
    }
  }

  return err_flags;

}  // pcps_enable_pci_dev



/*HDR*/
/**
 * @brief Setup and start a PCI device in a non-PnP system
 *
 * @param[in,out]  pddev        Pointer to the device structure to be set up
 * @param[in]      bus_num      The PCI bus number returned by the PCI BIOS
 * @param[in]      dev_fnc_num  The PCI device/function number returned by the PCI BIOS
 */
void pcps_setup_and_start_pci_dev( PCPS_DDEV *pddev,
            PCPS_BUS_NUM bus_num, PCPS_SLOT_NUM dev_fnc_num )
{
  PCPS_ERR_FLAGS err_flags;

  err_flags = pcps_read_pci_rsrc( bus_num, dev_fnc_num, pddev );
  _pcps_ddev_set_err_flags( pddev, err_flags );

  if ( !( err_flags & PCPS_EF_IO_INIT ) )
  {
    err_flags = pcps_enable_pci_dev( bus_num, dev_fnc_num,
                  pddev->rsrc_info.num_rsrc_mem );
    _pcps_ddev_set_err_flags( pddev, err_flags );
  }

  pcps_probe_device( pddev, bus_num, dev_fnc_num );

}  // pcps_setup_and_start_pci_dev



/*HDR*/
/**
 * @brief Detect and initialize PCI devices in a non-PnP system
 *
 * @param[in]  alloc_fnc     Pointer to function called to allocate a device structure for each detected device.
 * @param[in]  cleanup_fnc   Pointer to function called if the device structure needs to be de-allocated in case of error.
 * @param[in]  vendor_id     The PCI vendor ID code.
 * @param[in]  dev_type      An array with known PCI devices for the specified vendor ID
 * @param[in]  n_dev_types   The number of entries in the PCI device table
 */
void pcps_detect_pci_devices( PCPS_DDEV_ALLOC_FNC *alloc_fnc,
                              PCPS_DDEV_CLEANUP_FNC *cleanup_fnc,
                              ushort vendor_id, PCPS_DEV_TYPE dev_type[],
                              int n_dev_types )
{
  #if defined( MBG_TGT_QNX )
    #if defined( MBG_TGT_QNX_NTO )
      unsigned int pci_handle;  // specific to QNX Neutrino
    #endif
    unsigned int pci_hardware_mechanism;
    unsigned int pci_last_bus_number;
    unsigned int pci_interface_level_version;
  #elif defined( MBG_TGT_LINUX )
    // not yet supported/used
  #else
    uchar pci_hardware_mechanism;
    uchar pci_last_bus_number;
    ushort pci_interface_level_version;
  #endif
  ushort type_idx;
  int rc;


  #ifdef _mbg_pci_fnc_init
    rc = _mbg_pci_fnc_init();

    if ( rc != PCI_SUCCESS )
      return;
  #endif


  // See if PCI BIOS is installed on the machine.
  rc = _mbg_pci_find_bios( &pci_hardware_mechanism,
                           &pci_interface_level_version,
                           &pci_last_bus_number
                         );

  if ( rc == PCI_SUCCESS )
  {
    // PCI BIOS is installed, now try to find a PCI clock with
    // known ID (the list is terminated with a ID of 0).
    for ( type_idx = 0; type_idx < n_dev_types; type_idx++ )
    {
      ushort dev_idx;
      PCPS_DEV_TYPE *p = &dev_type[type_idx];

      if ( !( p->bus_flags & PCPS_BUS_PCI ) )
        continue;


      for ( dev_idx = 0; ; dev_idx++ )
      {
        PCPS_DDEV *pddev;
        #if defined( MBG_TGT_QNX )
          unsigned bus_num;
          unsigned dev_fnc_num;
        #else
          uchar bus_num;
          uchar dev_fnc_num;
        #endif

        rc = _mbg_pci_find_device( p->dev_id, vendor_id,
                                   dev_idx, &bus_num, &dev_fnc_num );

        if ( rc != PCI_SUCCESS )
          break;  // go to try next device ID


        _mbgddmsg_2( MBG_DBG_INIT_DEV, "Found PCI device %s (0x%04X)",
                     p->name, p->dev_id );

        // New device found, try to add to list.
        pddev = alloc_fnc();

        if ( pddev )  // Setup only if successful.
        {
          #if _PCPS_USE_PCI_PNP  //##++
            // This can be used to test the PNP functions in a
            // non-PNP environment.
            pcps_init_ddev( pddev, PCPS_BUS_PCI, p->dev_id );
          #else
            pddev->dev.type = *p;
          #endif

          pcps_setup_and_start_pci_dev( pddev, bus_num, dev_fnc_num );

          #if !_ACCEPT_UNINITD_CLOCKS
            if ( pddev->dev.cfg.err_flags )
            {
              _mbgddmsg_1( MBG_DBG_INIT_DEV,
                           "Remove PCI device: err_flags " FMT_08X "h",
                           (ulong) pddev->dev.cfg.err_flags );

              if ( cleanup_fnc )
                cleanup_fnc( pddev );
            }
          #else
            (void) cleanup_fnc;  // avoid compiler warning
          #endif
        }
      }
    }
  }

  #ifdef _mbg_pci_fnc_deinit
    _mbg_pci_fnc_deinit();
  #endif

}  // pcps_detect_pci_devices

#endif  // _PCPS_USE_PCI_BIOS



#if !_PCPS_USE_ISA_PNP

/*HDR*/
/**
 * @brief Detect and initialize ISA devices in a non-PnP system
 *
 * @param[in]  alloc_fnc     Pointer to function called to allocate a device structure for each detected device.
 * @param[in]  cleanup_fnc   Pointer to function called if the device structure needs to be de-allocated in case of error.
 * @param[in]  register_fnc  Pointer to function called to register a detected device.
 * @param[in]  isa_ports     An array with potential I/O base addresses for ISA devices.
 * @param[in]  isa_irqs      An array with potential IRQ numbers assigned to ISA devices.
 */
void pcps_detect_isa_devices( PCPS_DDEV_ALLOC_FNC *alloc_fnc,
                              PCPS_DDEV_CLEANUP_FNC *cleanup_fnc,
                              PCPS_DDEV_REGISTER_FNC *register_fnc,
                              int isa_ports[PCPS_MAX_ISA_CARDS],
                              int isa_irqs[PCPS_MAX_ISA_CARDS] )
{
  int *p_port = isa_ports;
  int *p_irq = isa_irqs;
  PCPS_DDEV *pddev;
  int i;

  if ( p_port == NULL )   // No list has been passed
    return;               // so don't try to detect ISA clocks.


  for( i = 0; i < PCPS_MAX_ISA_CARDS;
       i++, p_port++, p_irq ? ( p_irq++ ) : p_irq )
  {
    int irq_num;

    if ( *p_port == 0 )
      continue;

    irq_num = p_irq ? *p_irq : -1;

    _mbgddmsg_2( MBG_DBG_INIT_DEV,
                 "Check ISA device at port " FMT_03X "h, irq %d",
                 *p_port, irq_num );

    // Assume ISA device is available,
    // but clock type is unknown, yet.
    pddev = alloc_fnc();

    if ( pddev )  // Setup only if successfull.
    {
      pddev->dev.type.bus_flags = PCPS_BUS_ISA;

      // Set up basic cfg for ISA devices.
      pcps_add_rsrc_io( pddev, (uint16_t) *p_port, PCPS_NUM_PORTS_ISA );

      if ( irq_num != -1 )
        pcps_add_rsrc_irq( pddev, (uint16_t) *p_irq );

      // Init the device structure. This includes registration
      // of I/O ports with the OS's resource manager (if supported),
      // and reading the firmware ID.
      pcps_probe_device( pddev, 0, 0 );

      // If an error has occurred, then remove the last
      // device from the list and try next.
      if ( pddev->dev.cfg.err_flags )
      {
        _mbgddmsg_1( MBG_DBG_INIT_DEV,
                     "ISA device not found: err_flags " FMT_08X "h",
                     (ulong) pddev->dev.cfg.err_flags );
        if ( cleanup_fnc )
          cleanup_fnc( pddev );

        continue;
      }

      // Register the device with the OS, if required.
      if ( register_fnc )
        register_fnc( pddev );  //##++
    }
  }

}  // pcps_detect_isa_devices

#endif  //!_PCPS_USE_ISA_PNP



#if !_PCPS_USE_PNP

static /*HDR*/
/**
 * @brief Detect all bus-level devices in a non-PnP system, use specific alloc/cleanup functions
 *
 * @note If a DOS TSR is installed, be sure it is disabled (BUSY flag set)
 * when this function is called.
 *
 * @param[in]  alloc_fnc     Pointer to function called to allocate a device structure for each detected device.
 * @param[in]  cleanup_fnc   Pointer to function called if the device structure needs to be de-allocated in case of error.
 * @param[in]  isa_ports     An array with potential I/O base addresses for ISA devices.
 * @param[in]  isa_irqs      An array with potential IRQ numbers assigned to ISA devices.
 */
void _MBG_INIT_CODE_ATTR pcps_detect_devices_alloc( PCPS_DDEV_ALLOC_FNC *alloc_fnc,
                                                    PCPS_DDEV_CLEANUP_FNC *cleanup_fnc,
                                                    int isa_ports[PCPS_MAX_ISA_CARDS],
                                                    int isa_irqs[PCPS_MAX_ISA_CARDS] )
{
  #if defined( MBG_TGT_OS2 )
    rsrc_register_driver();  // register driver and init resource manager
  #endif

  #if _PCPS_USE_PCI_BIOS
    pcps_detect_pci_devices( alloc_fnc, cleanup_fnc, PCI_VENDOR_MEINBERG,
                             pcps_dev_type, N_PCPS_DEV_TYPE );
  #endif

  #if _PCPS_USE_MCA
    pcps_detect_mca_devices( alloc_fnc );
  #endif

  #if !_PCPS_USE_ISA_PNP
    pcps_detect_isa_devices( alloc_fnc, cleanup_fnc, NULL, isa_ports, isa_irqs );
  #endif

}  // pcps_detect_devices_alloc



/*HDR*/
/**
 * @brief Detect all bus-level devices in a non-PnP system
 *
 * @note If a DOS TSR is installed, be sure it is disabled (BUSY flag set)
 * when this function is called.
 *
 * @param[in]  isa_ports     An array with potential I/O base addresses for ISA devices.
 * @param[in]  isa_irqs      An array with potential IRQ numbers assigned to ISA devices.
 */
void _MBG_INIT_CODE_ATTR pcps_detect_devices( int isa_ports[PCPS_MAX_ISA_CARDS],
                                              int isa_irqs[PCPS_MAX_ISA_CARDS] )
{
  pcps_detect_devices_alloc( pcps_alloc_ddev, pcps_free_ddev, isa_ports, isa_irqs );

}  // pcps_detect_devices

#endif  // !_PCPS_USE_PNP


