
/**************************************************************************
 *
 *  $Id: toolutil.c 1.5 2017/07/05 07:32:38 martin REL_M $
 *
 *  Description:
 *    Common functions which can be used with Meinberg command line
 *    utility programs.
 *
 * -----------------------------------------------------------------------
 *  $Log: toolutil.c $
 *  Revision 1.5  2017/07/05 07:32:38  martin
 *  Renamed mbg_check_devices() to mbg_handle_devices() and added
 *  a new parameter which controls if only one or all devices
 *  should be handled if no device has been specified.
 *  New function mbg_open_device_by_param() which can be used e.g. to
 *  check if a parameter string is an device file name (MBG_DEV_FN),
 *  or a device index number, or a device name in MBG_DEV_NAME format,
 *  and calls the appropriate open function to open the device.
 *  Made mbg_snprint_hr_tstamp() more versatile by passing a new
 *  optional UTC offset parameter.
 *  Account for PCPS_HRT_BIN_FRAC_SCALE renamed to MBG_FRAC32_UNITS_PER_SEC.
 *  Account for frac_sec_from_bin() obsoleted by bin_frac_32_to_dec_frac().
 *  Print device options with OS-specific device names.
 *  Print special firmware version info, if appropriate.
 *  Reworked error checking after opening device.
 *  mbg_print_program_info() which is usually called first after program
 *  start now makes stdout unbuffered if output is redirected.
 *  Use safe string functions from str_util.c.
 *  Support Windows and QNX Neutrino targets.
 *  Added doxygen comments.
 *  Revision 1.4  2012/10/15 09:33:48  martin
 *  Use common way to handle version information.
 *  Open device with O_RDWR flag.
 *  Fixed printing delta timestamps.
 *  Added function to print PCPS_TIME.
 *  Added function to show PZF correlation.
 *  Added function mbg_print_hr_time() which optionally prints hex status.
 *  Let the display functions for HR timestamps optionally show
 *  the raw (hex) timestamps.
 *  Revision 1.3  2009/06/19 12:12:14  martin
 *  Added function mbg_print_hr_timestamp().
 *  Revision 1.2  2009/02/18 09:15:55  martin
 *  Support TAI and GPS time scales in mbg_snprint_hr_time().
 *  Revision 1.1  2008/12/17 10:45:13  martin
 *  Initial revision.
 *  Revision 1.1  2008/12/15 08:35:07  martin
 *  Initial revision.
 *
 **************************************************************************/

#define _TOOLUTIL
  #include <toolutil.h>
#undef _TOOLUTIL

// include Meinberg headers
#include <cfg_hlp.h>
#include <pcpsutil.h>
#include <timeutil.h>
#include <str_util.h>

// include system headers
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>



#if defined( MBG_TGT_WIN32 )
  #define isatty   _isatty
  #define fileno   _fileno
#endif



static __mbg_inline
/**
 * @brief Compute the difference between two ::PCPS_TIME_STAMP values
 *
 * @param[in]  p_ts      Pointer to the current time stamp.
 * @param[in]  p_prv_ts  Pointer to a previous time stamp.
 *
 * @return The time difference as double, in microseconds.
 */
double delta_timestamps( const PCPS_TIME_STAMP *p_ts, const PCPS_TIME_STAMP *p_prv_ts )
{
  uint64_t ts = pcps_time_stamp_to_uint64( p_ts );
  uint64_t prv_ts = pcps_time_stamp_to_uint64( p_prv_ts );

  // We divide by MBG_FRAC32_UNITS_PER_SEC to get the correct fractions
  // and we multiply by 1E6 to get the result in microseconds.
  return (double) ( (int64_t) ( ts - prv_ts ) ) * 1E6 / MBG_FRAC32_UNITS_PER_SEC;

}  // delta_timestamps



/*HDR*/
/**
 * @brief Print the program version to a string buffer
 *
 * @param[out]  s              Address of a string buffer to be filled.
 * @param[in]   max_len        Size of the string buffer.
 * @param[in]   micro_version  Micro version code for the application.
 *
 * @return The number of characters printed to the buffer.
 */
int mbg_program_version_str( char *s, size_t max_len, int micro_version )
{
  int n;

  #if defined( MBG_MICRO_VERSION_CODE_DEV )
    micro_version = MBG_MICRO_VERSION_CODE_DEV;
  #endif

  n = snprintf_safe( s, max_len, "%i.%i.%i", MBG_MAJOR_VERSION_CODE,
                     MBG_MINOR_VERSION_CODE, micro_version );

  return n;

}  // mbg_program_version_str



/*HDR*/
/**
 * @brief Print some program info to a string buffer
 *
 * @param[out]  s              Address of a string buffer to be filled.
 * @param[in]   max_len        Size of the string buffer.
 * @param[in]   pname          The program name.
 * @param[in]   micro_version  Micro version code for the application.
 * @param[in]   first_year     First copyright year.
 * @param[in]   last_year      Last copyright year.
 *
 * @return The number of characters printed to the buffer.
 */
int mbg_program_info_str( char *s, size_t max_len, const char *pname,
                          int micro_version, int first_year, int last_year )
{
  int n = 0;

  #if defined( MBG_MICRO_VERSION_CODE_DEV )
    micro_version = MBG_MICRO_VERSION_CODE_DEV;
  #endif

  if ( last_year == 0 )
    last_year = MBG_CURRENT_COPYRIGHT_YEAR;

  n += snprintf_safe( &s[n], max_len - n, "%s v", pname );

  n += mbg_program_version_str( &s[n], max_len - n, micro_version );

  n += snprintf_safe( &s[n], max_len - n, " copyright Meinberg " );

  if ( first_year != last_year )
    n += snprintf_safe( &s[n], max_len - n, "%04i-", first_year );

  n += snprintf_safe( &s[n], max_len - n, "%04i", last_year );

  return n;

}  // mbg_program_info_str



/*HDR*/
/**
 * @brief Print program info to console
 *
 * @param[in]   pname          The program name.
 * @param[in]   micro_version  Micro version code for the application.
 * @param[in]   first_year     First copyright year.
 * @param[in]   last_year      Last copyright year.
 */
void mbg_print_program_info( const char *pname, int micro_version, int first_year, int last_year )
{
  char ws[256];

  // If the output has been redirected then make stdout unbuffered,
  // e.g. to see the output immediately even though piped through 'tee'.
  if ( !isatty( fileno( stdout ) ) )
    setvbuf( stdout, NULL, _IONBF, 0 );

  mbg_program_info_str( ws, sizeof( ws ), pname, micro_version, first_year, last_year );

  printf( "\n%s\n\n", ws );

}  // mbg_print_program_info



/*HDR*/
/**
 * @brief Print usage intro to console
 *
 * @param[in]   pname  The program name.
 * @param[in]   info   An optional additional info string, may be NULL.
 */
void mbg_print_usage_intro( const char *pname, const char *info )
{
  printf( "Usage:  %s [[opt] [opt] ...] [[dev] [dev] ...]\n\n", pname );

  if ( info )
    printf( "%s\n\n", info );


}  // mbg_print_usage_intro



/*HDR*/
/**
 * @brief Print info on a single program option / argument
 *
 * @param[in]   opt_name  The option name, optional, may be NULL.
 * @param[in]   opt_info  The option info, optional, may be NULL.
 */
void mbg_print_opt_info( const char *opt_name, const char *opt_info )
{
  if ( opt_name == NULL )
    opt_name = str_empty;

  if ( opt_info == NULL )
    opt_info = str_empty;

  printf( "  %8s   %s\n", opt_name, opt_info );

}  // mbg_print_opt_info



/*HDR*/
/**
 * @brief Print info on common program help arguments
 *
 * Lists program parameters causing printing
 * of help / usage information.
 */
void mbg_print_help_options( void )
{
  puts( "where \"opt\" is one of the options:" );
  mbg_print_opt_info( "-? or -h", "print this usage information" );

}  // mbg_print_help_options



/*HDR*/
/**
 * @brief Print common info on how to specify devices on the command line
 */
void mbg_print_device_options( void )
{
  printf( "\n" );

  puts( "where \"dev\" can be:" );

  #if MBG_TGT_HAS_DEV_FN
    printf( "    a device file name, e.g. \"%s\" or \"%s\"\n",
            EXAMPLE_DEV_FN_1, EXAMPLE_DEV_FN_2 );
  #endif
    printf( "    a device type name, with or with S/N appended, e.g. \"%s\" or \"%s\"\n",
            EXAMPLE_DEV_NAME_1, EXAMPLE_DEV_NAME_2 );

    printf( "    a device index, e.g. \"0\", \"1\",\"2\", ...\n" );

}  // mbg_print_device_options



/*HDR*/
/**
 * @brief Print program info and default usage information
 *
 * @param[in]   pname      The program name.
 * @param[in]   prog_info  An optional additional info string, may be NULL.
 */
void mbg_print_default_usage( const char *pname, const char *prog_info )
{
  mbg_print_usage_intro( pname, prog_info );
  mbg_print_help_options();
  mbg_print_device_options();
  puts( "" );

}  // mbg_print_default_usage



/*HDR*/
/**
 * @brief Retrieve and print some common device info
 *
 * @param[in]   dh        Valid ::MBG_DEV_HANDLE handle to a Meinberg device.
 * @param[in]   dev_name  A device name string to be printed
 * @param[out]  p_dev     Pointer to a device info structure to be read from the device
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
int mbg_get_show_dev_info( MBG_DEV_HANDLE dh, const char *dev_name, PCPS_DEV *p_dev )
{
  RECEIVER_INFO ri;
  unsigned long port;
  int irq_num;
  int rc;

  if ( dev_name )
    printf( "%s:\n", dev_name );

  // get information about the device
  rc = mbg_get_device_info( dh, p_dev );

  if ( mbg_cond_err_msg( rc, "mbg_get_device_info" ) )
    goto out;

  rc = mbg_setup_receiver_info( dh, p_dev, &ri );

  if ( mbg_cond_err_msg( rc, "mbg_setup_receiver_info" ) )
    goto out;


  printf( "%s", _pcps_type_name( p_dev ) );

  if ( strlen( _pcps_sernum( p_dev ) ) &&
       strcmp( _pcps_sernum( p_dev ), "N/A" ) )
    printf( " %s", _pcps_sernum( p_dev ) );

  printf( " (FW %X.%02X",
          _pcps_fw_rev_num_major( _pcps_fw_rev_num( p_dev ) ),
          _pcps_fw_rev_num_minor( _pcps_fw_rev_num( p_dev ) )
        );

  if ( chk_sw_rev_name( &ri.sw_rev, 0 ) )
    printf( " \"%s\"", ri.sw_rev.name );

  if ( _pcps_has_asic_version( p_dev ) )
  {
    PCI_ASIC_VERSION av;
    int rc = mbg_get_asic_version( dh, &av );

    if ( mbg_rc_is_success( rc ) )
    {
      av = _convert_asic_version_number( av );

      printf( ", ASIC %u.%02u",
              _pcps_asic_version_major( av ),
              _pcps_asic_version_minor( av )
            );
    }
  }

  printf( ")" );

  port = _pcps_short_port_base( p_dev, 0 );

  if ( port )
    printf( " at port 0x%03lX", port );

  port = _pcps_short_port_base( p_dev, 1 );

  if ( port )
    printf( "/0x%03lX", port );

  irq_num = _pcps_irq_num( p_dev );

  if ( irq_num != -1 )
    printf( ", irq %i", irq_num );

  rc = MBG_SUCCESS;

out:
  puts( str_empty );
  return rc;

}  // mbg_get_show_dev_info



/*HDR*/
/**
 * @brief Print device info and take some action on a specific device
 *
 * @param[in]   dh        Valid ::MBG_DEV_HANDLE handle to a Meinberg device.
 * @param[in]   dev_name  A device name string to be printed.
 * @param[in]   fnc       Pointer to a callback function that actually takes the action.
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
int mbg_handle_device( MBG_DEV_HANDLE dh, const char *dev_name,
                       MBG_DEV_HANDLER_FNC *fnc )
{
  PCPS_DEV dev;
  int rc;

  rc = mbg_get_show_dev_info( dh, dev_name, &dev );

  if ( mbg_rc_is_success( rc ) )
    if ( fnc )
      rc = fnc( dh, &dev );

  mbg_close_device( &dh );

  puts( "" );

  return rc;

}  // mbg_handle_device



/*HDR*/
/**
 * @brief Get the number of devices actually present
 *
 * Do a real search only when called for the first time.
 *
 * @return The number of devices currently present.
 */
int chk_get_num_devices( void )
{
  static int num_devices;
  static bool has_searched_devices;

  if ( !has_searched_devices )
  {
    num_devices = mbg_find_devices();
    has_searched_devices = true;
  }

  return num_devices;

}  // chk_get_num_devices



/*HDR*/
int mbg_open_device_by_param( MBG_DEV_HANDLE *p_dh, const char *dev_param_str,
                              int dev_idx, int devices_specified,
                              char *dev_name_buffer, size_t dev_name_buffer_size,
                              int chk_dev_flags )
{
  MBG_DEV_HANDLE dh = MBG_INVALID_DEV_HANDLE;
  int devices_found;
  int rc = MBG_ERR_GENERIC;

  if ( dev_name_buffer_size )  // This should always be the case
    dev_name_buffer[0] = 0;    // Make string empty

  if ( dev_param_str )
  {
    // A device parameter string has been specified.
    // This can be:
    //  - A device file name, see ::MBG_DEV_FN
    //  - A device model name like "GPS180PEX", eventually with the serial number
    //    appended after an underscore, as in "GPS180PEX_029511026220".
    //    See ::MBG_DEV_NAME.
    //  - A device index number as string, e.g "0" or "3".

    char *endptr;
    long l;

    #if MBG_TGT_HAS_DEV_FN
      // Check if the device parameter represents a device file name
      if ( strncmp( dev_param_str, mbg_dev_fn_base, strlen( mbg_dev_fn_base ) ) == 0 )
      {
        snprintf_safe( dev_name_buffer, dev_name_buffer_size, "%s", dev_param_str );
        dh = mbg_open_device_by_dev_fn( dev_param_str );
        goto out_chk;
      }
    #endif


    // Check if the device parameter string is an index number.
    endptr = NULL;
    l = strtol( dev_param_str, &endptr, 0 );

    // If endptr now points past the string then the
    // whole string has been interpreted as number.
    if ( endptr == &dev_param_str[strlen( dev_param_str )] )
    {
      long l_max = (long) ( ( (unsigned int) -1 ) >> 1 );

      if ( l > l_max )
      {
        // The decoded number exceeds the maximum number
        // that can be passed to ::mbg_open_device.
        rc = MBG_ERR_RANGE;
        goto out;
      }

      // We use the decoded number as device index.
      dev_idx = (int) l;
      goto open_dev_with_idx;
    }


    // Finally assume the device name is a model name, eventually
    // with serial number appended.
    snprintf_safe( dev_name_buffer, dev_name_buffer_size, "%s", dev_param_str );
    dh = mbg_open_device_by_name( dev_param_str, MBG_MATCH_MODEL );
    goto out_chk;
  }


  // No device has been specified on the command line.

#if 0 // TODO
  if ( chk_dev_flags & CHK_DEV_WITHOUT_DEV )
  {
    // TODO Check if this branch is useful
    // We have been called without a device to be used.
    // This may be useful e.g. if we just want to print
    // some general usage information.
    if ( fnc )
    {
      *p_dh = MBG_INVALID_DEV_HANDLE;
      rc = fnc( *p_dh, NULL );
    }
    else
    {
      // TODO Should we report an error if we have been called
      // without callback function and without device?
    }
    goto out;
  }
#endif


open_dev_with_idx:
  devices_found = chk_get_num_devices();

  if ( devices_found == 0 )  // no device found
  {
    // Don't continue without any device, unless
    // explicitly requested to do so.
    if ( !( chk_dev_flags & CHK_DEV_WITHOUT_DEV ) )
      return MBG_ERR_NO_DEV;

    // We may continue even if no device is available,
    // but have to process the callback loop only once.
    devices_found = 1;
  }

  // Unless explicitly requested to handle all devices
  // we have found, we only handle the first one.
  if ( !( chk_dev_flags & CHK_DEV_ALL_DEVICES ) )
    devices_found = 1;


  // Try to open the device with the specified index number.
  mbg_dev_fn_from_dev_idx( dev_name_buffer, dev_name_buffer_size, dev_idx );
  dh = mbg_open_device( dev_idx );

out_chk:
  // Now check if the "open" function we've just called before
  // was successful.
  if ( dh == MBG_INVALID_DEV_HANDLE )
  {
    char msg[256];
    int n = 0;
    rc  = mbg_get_last_error( NULL );

    n = snprintf_safe( msg, sizeof( msg ), "** Failed to open device" );

    if ( dev_name_buffer[0] )  // string not empty
      n += snprintf_safe( &msg[n], sizeof( msg ) - n, " %s", dev_name_buffer );

    n += snprintf_safe( &msg[n], sizeof( msg ) - n, ": %s", mbg_strerror( rc ) );
    fprintf( stderr, "%s\n\n", msg );
    goto out;
  }

  rc = MBG_SUCCESS;


out:
  *p_dh = dh;

  return rc;

}  // mbg_open_device_by_param



/*HDR*/
/**
 * @brief Main action handler that can be called by utility programs
 *
 * This function checks the command line parameters passed to the program.
 * Those which have not been evaluated before are interpreted as device specifiers.
 *
 * Device specifiers can be device file names like "/dev/mbgclock0" on target
 * platforms that support such device names, see ::MBG_DEV_FN,
 * or device type names like "GPS180PEX" which can optionally have a serial
 * number appended, like "GPS180PEX_029511026220", to be able to distinguish
 * between several devices of the same type (see ::MBG_DEV_NAME),
 * or just an index number as required for the ::mbg_open_device call.
 *
 * For each of the specified devices the callback function @p fnc is called to
 * take some specific action on the device.
 *
 * If no device has been specified in the argument list then ::mbg_find_devices
 * is called to set up a device list, and depending on whether ::CHK_DEV_ALL_DEVICES
 * is set in @p chk_dev_flags the callback function @p fnc is either called
 * for every device from the list, or only for the first one.
 *
 * @param[in]   argc     Number of parameters of the program argument list.
 * @param[in]   argv     Array of program argument strings.
 * @param[in]   optind   Number of the command line arguments that have already been handled.
 * @param[in]   fnc      Pointer to a callback function that actually takes an action on each specified device.
 * @param[in]   chk_dev_flags    Bit mask controlling the behavior of the function, see ::CHK_DEV_FLAGS
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
int mbg_handle_devices( int argc, char *argv[], int optind,
                        MBG_DEV_HANDLER_FNC *fnc, int chk_dev_flags )
{
  MBG_DEV_FN tmp_dev_fn;
  MBG_DEV_HANDLE dh;
  int devices_specified = argc - optind;
  int devices_found = 0;
  const char *dev_param_str = NULL;
  int rc = MBG_ERR_NO_DEV;
  int i;

  if ( devices_specified )
  {
    // One or more device names have been specified
    // on the command line, so handle each device.
    for ( i = optind; i < argc; i++ )
    {
      dev_param_str = argv[i];

      rc = mbg_open_device_by_param( &dh, dev_param_str, 0, devices_specified,
                                     tmp_dev_fn, sizeof( tmp_dev_fn ), chk_dev_flags );

      if ( mbg_rc_is_success( rc ) )
        rc = mbg_handle_device( dh, ( devices_specified > 1 ) ? tmp_dev_fn : NULL, fnc );

      // Don't continue if one of the specified devices failed.
      if ( mbg_rc_is_error( rc ) )
        break;
    }

    goto out;
  }


  // No device has been specified on the command line.

  if ( chk_dev_flags & CHK_DEV_WITHOUT_DEV )
  {
    // TODO Check if this branch is useful
    // We have been called without a device to be used.
    // This may be useful e.g. if we just want to print
    // some general usage information.
    if ( fnc )
    {
      dh = MBG_INVALID_DEV_HANDLE;
      rc = fnc( dh, NULL );
    }
    else
    {
      // TODO Should we report an error if we have been called
      // without callback function and without device?
    }
    goto out;
  }


  devices_found = chk_get_num_devices();

  if ( devices_found == 0 )  // no device found
  {
    // Don't continue without any device, unless
    // explicitly requested to do so.
    if ( !( chk_dev_flags & CHK_DEV_WITHOUT_DEV ) )
    {
      printf( "No device found.\n" );  // TODO
      rc = MBG_ERR_NO_DEV;
      goto out;
    }

    // We may continue even if no device is available,
    // but must process the callback loop only once.
    devices_found = 1;
  }

  // Unless explicitly requested to handle all devices
  // we have found, we only handle the first one.
  if ( !( chk_dev_flags & CHK_DEV_ALL_DEVICES ) )
    devices_found = 1;

  for ( i = 0; i < devices_found; i++ )
  {
    rc = mbg_open_device_by_param( &dh, NULL, i, 0, tmp_dev_fn,
                                   sizeof( tmp_dev_fn ), chk_dev_flags );

    if ( mbg_rc_is_success( rc ) )
      rc = mbg_handle_device( dh, ( devices_found > 1 ) ? tmp_dev_fn : NULL, fnc );

    // If one of the unspecified devices failed we continue anyway
    // and don't break here.
  }


out:
  return rc;

}  // mbg_handle_devices



/*HDR*/
/**
 * @brief Print date and time from a ::PCPS_TIME structure to a string
 *
 * @param[out]  s        Address of a string buffer to be filled.
 * @param[in]   max_len  Size of the string buffer.
 * @param[in]   p        Pointer to a ::PCPS_TIME structure to be evaluated.
 * @param[in]   verbose  Increase verbosity of the output:<br>
 *                       > 0: append UTC offset and status<br>
 *                       > 1: append signal value
 *
 * @return The number of characters printed to the buffer.
 */
size_t mbg_snprint_date_time( char *s, size_t max_len, const PCPS_TIME *p, int verbose )
{
  size_t n = 0;

  n += snprintf_safe( &s[n], max_len - n, "%04u-%02u-%02u %02u:%02u:%02u.%02u",
                      pcps_exp_year( p->year, mbg_exp_year_limit ), p->month, p->mday,
                      p->hour, p->min, p->sec, p->sec100 );

  if ( verbose > 0 )
    n += snprintf_safe( &s[n], max_len - n, " (UTC%+ih), st: %02Xh",
                        p->offs_utc, p->status );

  if ( verbose > 1 )
    n += snprintf_safe( &s[n], max_len - n, ", sig: %i", p->signal );

  return n;

}  // mbg_snprint_date_time



/*HDR*/
/**
 * @brief Print date and time from a ::PCPS_TIME_STAMP structure to a string
 *
 * @param[out]  s         Address of a string buffer to be filled.
 * @param[in]   max_len   Size of the string buffer.
 * @param[in]   p         Pointer to a ::PCPS_TIME_STAMP structure to be evaluated.
 * @param[in]   utc_offs  A local time offset added to the time stamp before converted to date and time.
 * @param[in]   show_raw  Flag indicating if a raw timestamp (hex) is to be printed, too.
 *
 * @return The number of characters printed to the buffer.
 */
size_t mbg_snprint_hr_tstamp( char *s, size_t max_len, const PCPS_TIME_STAMP *p,
                              long utc_offs, int show_raw )
{
  long l;
  time_t t;
  struct tm tm = { 0 };
  size_t n = 0;
  int rc;

  if ( show_raw )
    n += snprintf_safe( &s[n], max_len - n, "raw: 0x%08lX.%08lX, ",
                        (ulong) p->sec, (ulong) p->frac );

  // We'll use the standard C library functions to convert the seconds
  // to broken-down calendar date and time.
  l = (long) p->sec + utc_offs;
  t = cvt_to_time_t( l );

  // Our time stamp may be UTC, or have been converted to local time.
  // Anyway, since we don't want to account for the system's time zone
  // settings, we always use the gmtime() function for conversion,
  // or our own wrapper, respectively:
  rc = mbg_gmtime( &tm, &t );

  if ( mbg_rc_is_success( rc ) )
    n += snprintf_safe( &s[n], max_len - n, "%04i-%02i-%02i %02i:%02i:%02i." PCPS_HRT_FRAC_SCALE_FMT,
                        tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday,
                        tm.tm_hour, tm.tm_min, tm.tm_sec,
                        (ulong) bin_frac_32_to_dec_frac( p->frac, PCPS_HRT_FRAC_SCALE ) );
  else
    n += snprint_gmtime_error( &s[n], max_len - n, rc, t, __func__ );

  return n;

}  // mbg_snprint_hr_tstamp



/*HDR*/
/**
 * @brief Print date and time from a ::PCPS_HR_TIME structure to a string
 *
 * Converts the UTC timestamp first to local time according to
 * the ::PCPS_HR_TIME::utc_offs value.
 *
 * @param[out]  s         Address of a string buffer to be filled.
 * @param[in]   max_len   Size of the string buffer.
 * @param[in]   p         Pointer to a ::PCPS_HR_TIME structure to be evaluated.
 * @param[in]   show_raw  Flag indicating if a raw timestamp (hex) is to be printed, too.
 *
 * @return The number of characters printed to the buffer.
 */
size_t mbg_snprint_hr_time( char *s, int max_len, const PCPS_HR_TIME *p, int show_raw )
{
  char ws[80];
  PCPS_TIME_STAMP ts = p->tstamp;
  size_t n;
  const char *time_scale_name;
  const char *cp;

  // If the local time offset is not 0 then add this to the time stamp
  // and set up a string telling the offset.
  if ( p->utc_offs )
  {
    // The local time offset is in seconds and may be negative, so we
    // convert to absolute hours and minutes first.
    ldiv_t ldt = ldiv( labs( p->utc_offs ) / 60, 60 );

    snprintf_safe( ws, sizeof( ws ), "%c%lu:%02luh",
                   ( p->utc_offs < 0 ) ? '-' : '+',
                   ldt.quot, ldt.rem );
    cp = ws;
  }
  else
    cp = "";   // no local time offset


  // Convert the local time stamp to calendar date and time.
  n = mbg_snprint_hr_tstamp( s, max_len, &ts, p->utc_offs, show_raw );

  // By default the time stamp represents UTC plus an optional local time offset.
  time_scale_name = "UTC";

  // However, some cards may be configured to output TAI or GPS time.
  if ( p->status & PCPS_SCALE_TAI )
    time_scale_name = "TAI";      // time stamp represents TAI
  else
    if ( p->status & PCPS_SCALE_GPS )
      time_scale_name = "GPS";    // time stamp represents GPS system time

  n += snprintf_safe( &s[n], max_len - n, " %s%s", time_scale_name, cp );

  return n;

}  // mbg_snprint_hr_time



/*HDR*/
/**
 * @brief Print date and time from a ::PCPS_TIME_STAMP structure
 *
 * First the HR timestamp passed by parameter @p p_ts is printed.
 *
 * If the parameter @p p_prv_ts is not NULL then it should specify
 * an earlier timestamp, and the elapsed time since @p p_ts
 * is appended.
 *
 * Finally the latency value is printed in microseconds, unless
 * parameter @p no_latency is != 0.
 *
 * @param[in]   p_ts         Pointer to a ::PCPS_TIME_STAMP structure to be evaluated.
 * @param[in]   hns_latency  A latency number in hectonanoseconds, i.e. 100 ns units.
 * @param[in]   p_prv_ts     Pointer to a ::PCPS_TIME_STAMP structure to be evaluated, may be NULL
 * @param[in]   no_latency   A flag indicating if printing the latency should be suppressed.
 * @param[in]   show_raw     Flag indicating if a raw timestamp (hex) is to be printed, too.
 *
 * @see ::mbg_print_hr_time
 */
void mbg_print_hr_timestamp( PCPS_TIME_STAMP *p_ts, int32_t hns_latency, PCPS_TIME_STAMP *p_prv_ts,
                             int no_latency, int show_raw )
{
  char ws[80];

  mbg_snprint_hr_tstamp( ws, sizeof( ws ), p_ts, 0, show_raw );
  printf( "HR time %s", ws );

  if ( p_prv_ts )
    printf( " (%+.1f us)", delta_timestamps( p_ts, p_prv_ts ) );

  if ( !no_latency )
    printf( ", latency: %.1f us", ( (double) hns_latency ) / 10 );

  puts( str_empty );

}  // mbg_print_hr_timestamp



/*HDR*/
/**
 * @brief Print date and time from a ::PCPS_HR_TIME structure
 *
 * First the HR timestamp passed by parameter @p p_ht is printed.
 *
 * If the parameter @p p_prv_ts is not NULL then it should specify
 * an earlier timestamp, and the elapsed time since @p p_ht
 * is appended.
 *
 * Next the latency value is printed in microseconds, unless
 * parameter @p no_latency is != 0.
 *
 * @param[in]   p_ht         Pointer to a ::PCPS_TIME_STAMP structure to be evaluated.
 * @param[in]   hns_latency  A latency number in hectonanoseconds, i.e. 100 ns units.
 * @param[in]   p_prv_ts     Pointer to a ::PCPS_TIME_STAMP structure to be evaluated, may be NULL
 * @param[in]   no_latency   A flag indicating if printing the latency should be suppressed.
 * @param[in]   show_raw     Flag indicating if a raw timestamp (hex) is to be printed, too.
 * @param[in]   verbose      Increase verbosity of the output:<br>
 *                           > 0: append status code<br>
 *                           > 1: append signal value
 *
 * @see ::mbg_print_hr_timestamp
 */
void mbg_print_hr_time( PCPS_HR_TIME *p_ht, int32_t hns_latency, PCPS_TIME_STAMP *p_prv_ts,
                        int no_latency, int show_raw, int verbose )
{
  char ws[80];

  mbg_snprint_hr_time( ws, sizeof( ws ), p_ht, show_raw );
  printf( "HR time %s", ws );

  if ( p_prv_ts )
    printf( " (%+.1f us)", delta_timestamps( &p_ht->tstamp, p_prv_ts ) );

  if ( !no_latency )
    printf( ", latency: %.1f us", ( (double) hns_latency ) / 10 );

  if ( verbose > 0 )
    printf( ", st: 0x%04lX", (ulong) p_ht->status );

  if ( verbose > 1 )
    printf( ", sig: %i", p_ht->signal );

  puts( "" );

}  // mbg_print_hr_time



/*HDR*/
/**
 * @brief Retrieve and print PZF correlation info for a device which supports this
 *
 * @param[in]   dh              Valid ::MBG_DEV_HANDLE handle to a Meinberg device.
 * @param[in]   show_corr_step  A flag indicating if correlation step indicators are to be appended
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
int mbg_show_pzf_corr_info( MBG_DEV_HANDLE dh, int show_corr_step )
{
  CORR_INFO ci;
  char ws[80];
  const char *cp;

  int rc = mbg_get_corr_info( dh, &ci );

  if ( mbg_cond_err_msg( rc, "mbg_get_corr_info" ) )
    return rc;


  if ( ci.status < N_PZF_CORR_STATE )
    cp = pzf_corr_state_name[ci.status];
  else
  {
    snprintf_safe( ws, sizeof( ws ), "unknown status code: 0x%02X", ci.status );
    cp = ws;
  }

  printf( "%s, PZF correlation: %u%%", cp, ci.val );

  if ( show_corr_step )
    if ( ci.corr_dir != ' ' )
      printf( " Shift: %c", ci.corr_dir );

  return MBG_SUCCESS;

}  // mbg_show_pzf_corr_info



/*HDR*/
/**
 * @brief Retrieve a ::MBG_GNSS_MODE_INFO structure from a device
 *
 * @param[in]   dh  Valid ::MBG_DEV_HANDLE handle to a Meinberg device.
 * @param[out]  p   Pointer to a ::MBG_GNSS_MODE_INFO structure to be filled up.
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 */
int mbg_get_gps_gnss_mode_info_chk( MBG_DEV_HANDLE dh, MBG_GNSS_MODE_INFO *p )
{
  int rc = mbg_get_gps_gnss_mode_info( dh, p );
  mbg_cond_err_msg( rc, "mbg_get_gps_gnss_mode_info" );

  return rc;

}  // mbg_get_gps_gnss_mode_info_chk



